/* main.c
 * 14-may-2004/FK Dropping rotsim.
 * 23-sep-2003/FK Version 0.3.1
 *
 * 2002-11-15/FK  Changed the filename from gentest.c to main.c since this
 *                is more in line with the role of this file.
 * 2002-02-22/FK  Removed a lot of dead code.
 *
 * 2001-12-25/FK
 * This is the main program for the organ simulator.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <ctype.h>
#include <math.h>
#include <sys/soundcard.h>
#include <sys/time.h>
#include <unistd.h>
#include <pthread.h>
#include <assert.h>
#include <limits.h>

#include "preamp.h"
#include "scanner.h"
#include "reverb.h"
#include "main.h"
#include "midiIn.h"
#include "config.h"
#include "pgmParser.h"
#include "whirl.h"

/* #define OFFLINE			/* Batch-mode debugging */

#define NOF_CFG_OVERS 32	/* Max command-line config commands */

static char * defaultConfigFile = "default.cfg";

static pthread_t t_midi;	/* MIDI input thread reference */

static char * audioDevice = "/dev/dsp";
static char audioDeviceBuffer[PATH_MAX + 1];
static int  audio_fd;
static int  audio_format;
static int  audio_format_requested = AFMT_S16_LE;
static int  audio_channels;
static int  audio_channels_requested = 2;
static int  audio_splrate;
static int  audio_splrate_requested = 22050;
static int  audio_fragsize;
static int  audio_fragsize_requested = 0x00040009; /* 4 bufs, 2^9=512 bytes */
static int  audio_profile;
static int  audio_profile_requested = APF_CPUINTENS;
/* static int  audio_profile_requested = APF_NORMAL; */

/* 
 * The audio chain is mono until it leaves the rotary spaker FX.
 * We work with 16 bit signed values at 22,050 Hz.
 * That means:
 *   22,050.00 samples per second.
 *       22.05 samples per millisecond.
 *      220.50 samples per 10 ms
 *    2,205.00 samples per 100 ms
 *
 * With buffer sizes a power of two, we get (at 22,050 Hz, 16-bit samples):
 *  Samples      ms    Mono buf.   Stereo  buf.
 *  -------------------------------------------
 *      128    5.81          256            512
 *      256   11.61          512          1,024
 *      512   23.22        1,024          2,048
 *    1,024   46.44        2,048          4,096
 *    2,048   92.88        4,096          8,192
 */

#define BUFFER_SIZE_SAMPLES  128 // 256 // 2048
#define BUFFER_SIZE_BYTES    256 // 512 // 4096
#define STEREO_BUFR_BYTES    512 // 1028 // 8192

struct _drawbarPreset {
  char * key;
  unsigned int setting [B_size];
};

unsigned char bufA [BUFFER_SIZE_BYTES];
unsigned char bufB [BUFFER_SIZE_BYTES];

unsigned char bufC [STEREO_BUFR_BYTES];

double SampleRateD = 22050.0;
int    SampleRateI = 22050;

double vMax = 2.0 * M_PI;
double unitAmplification = 32767.0;
double fragTimeMs;

double DrbAmp [B_size];		/* Drawbar amplification */

static char * defaultProgrammeFile = "default.pgm";


/*
 * -------------- FUNCTIONS ------------------
 */

/*
 *
 */
static void setAudioDevice (char * s) {
  if (s != NULL) {
    strncpy (audioDeviceBuffer, s, PATH_MAX);
    audioDevice = audioDeviceBuffer;
  }
}

/*
 * This routine initializes the audio driver according to the current
 * configuration parameters.
 */
static void initAudio () {
  /* Open the audio hardware. */

  if ((audio_fd = open (audioDevice, O_WRONLY, 0)) == -1) {
    perror (audioDevice);
    exit (1);
  }

  /* Set the requested fragment size. */

  audio_fragsize = audio_fragsize_requested;

  if (ioctl (audio_fd, SNDCTL_DSP_SETFRAGMENT, & audio_fragsize) == -1) {
    perror ("SNDCTL_DSP_SETFRAGMENT");
    exit (1);
  }

  /* Set the requested Application DSP Profile. */

  audio_profile = audio_profile_requested;

#ifdef USE_DSP_PROFILE
  if (ioctl (audio_fd, SNDCTL_DSP_PROFILE, & audio_profile) == -1) {
    perror ("SNDCTL_DSP_PROFILE");
    exit (1);
  }
#endif /* USE_DSP_PROFILE */

  /* Set the requested audio format. */

  audio_format = audio_format_requested;

  if (ioctl (audio_fd, SNDCTL_DSP_SETFMT, & audio_format) == -1) {
    perror ("SNDCTL_DSP_SETFMT");
    exit (1);
  }

  if (audio_format != audio_format_requested) {
    fprintf (stderr, "Audio format not supported by system.\n");
    exit (1);
  }

  /* Set the number of audio channels. */

  audio_channels = audio_channels_requested;

  if (ioctl (audio_fd, SNDCTL_DSP_CHANNELS, & audio_channels) == -1) {
    perror ("SNDCTL_DSP_CHANNELS");
    exit (1);
  }

  if (audio_channels != audio_channels_requested) {
    fprintf (stderr,
	     "%d channels not supported by system.\n",
	     audio_channels_requested);
    exit (1);	     
  }

  /* Select sample rate. */

  audio_splrate = audio_splrate_requested;

  if (ioctl (audio_fd, SNDCTL_DSP_SPEED, & audio_splrate) == -1) {
    perror ("SNDCTL_DSP_SPEED");
    exit (1);
  }

  if (audio_splrate != audio_splrate_requested) {
    fprintf (stderr,
	     "Sampling rate %d Hz not supported by system.\n",
	     audio_splrate_requested);
    exit (1);
  }

  /* Inquire fragment size */

#ifdef COMMENT
  if (ioctl (audio_fd, SNDCTL_DSP_GETBLKSIZE, & audio_fragsize) == -1) {
    perror ("SNDCTL_DSP_GETBLKSIZE");
    exit (1);
  }

  fprintf (stderr, "AUDIO FRAGMENT SIZE : %d\n", audio_fragsize);
#endif /* COMMENT */

}

/*
 *
 */
static void initAll () {
  fragTimeMs = (1000 * BUFFER_SIZE_SAMPLES) / SampleRateI;

  fprintf (stderr, "Scanner : ");
  fflush (stderr);
  initVibrator ();

  fprintf (stderr, "Oscillators : ");
  fflush (stderr);
  initToneGenerator ();

  fprintf (stderr, "Preamp : ");
  fflush (stderr);
  initPreamp ();

  fprintf (stderr, "Reverb : ");
  fflush (stderr);
  initReverb ();

  fprintf (stderr, "Whirl : ");
  fflush (stderr);
  initWhirl ();

  fprintf (stderr, "Audio :\n");
  fflush (stderr);
  initAudio ();

  /* MIDI initialization is done by the MIDI thread when it starts.
   * Probably not a good idea, but there it is.
   */

/*    fprintf (stderr, "MIDI..."); */
/*    fflush (stderr); */
/*    (void) prepareMIDI (); */


  //  fprintf (stderr, " Done\n");
  //  fflush (stderr);
}


/*
 * Pads the mono buffer into both channels of the stereo output buffer.
 */
static unsigned char* padBuffer (unsigned char * inbuf,
				 unsigned char * outbuf,
				 size_t bufferLengthSamples) {
  int i;
  short * xsp = (short *) inbuf;
  short * ysp = (short *) outbuf;

  for (i = 0; i < bufferLengthSamples; i++) {
    *ysp++ = *xsp;
    *ysp++ = *xsp++;
  }

  return outbuf;
}

/*
 * 
 */
int genConfig (ConfigContext * cfg) {
  if (strcasecmp (cfg->name, "main.audiodevice") == 0) {
    setAudioDevice (cfg->value);
  }
}


#ifdef OFFLINE
/*
 * This routine is used when we want to step-debug the DSP. It is called
 * from function offLineAnalysis() below.
 */
static void offLineGenerate (FILE * op, int features) {
  unsigned char * obuf;

  oscGenerateFragment (bufA, BUFFER_SIZE_SAMPLES);

  obuf = vibratoProc (bufA, bufB, BUFFER_SIZE_SAMPLES);

  obuf = preamp (obuf, bufB, BUFFER_SIZE_SAMPLES);

  if (features & 4) {
    whirlProc (obuf, bufC, BUFFER_SIZE_SAMPLES);
  }
  else {
    padBuffer (obuf, bufC, BUFFER_SIZE_SAMPLES);
  }

  fwrite (bufC, sizeof (unsigned char), STEREO_BUFR_BYTES, op);
}

/*
 * This routine is used for step-debugging of the DSP chain.
 */
static void offLineAnalysis () {
  int i;
  unsigned char * obuf;
  unsigned int drwb[9]={8,0, 8,8,0,0, 0,0,0};
  char * fname = "analysis.pcm";
  FILE * op;
  int note = 60;
  int not2 = note + 7;
  int features = 0;

  setMIDINoteShift (0);
  setDrawBars (drwb);
  setSwell (1.0);
  set7bitGain ((unsigned char) 127);
  setClean (FALSE);
  setVibrato (0);
/*   setFastPercussionDecay (0.25); */

  if ((op = fopen (fname, "wb")) != NULL) {

    /* Initialize IIR lines */
    offLineGenerate (op, features);

    /* ATTACK */
    oscNoteOn (note);
    oscNoteOn (not2);

    offLineGenerate (op, features);

    /* HOLD */
    for (i = 0; i < 128; i++) {
      offLineGenerate (op, features);
    }

    /* RELEASE */
    oscNoteOff (note);
    oscNoteOff (not2);

    offLineGenerate (op, features);

    fclose (op);
    printf ("Output captured on file %s\n", fname);
  }
  else {
    perror (fname);
  }
}
#endif /* OFFLINE */

#ifdef STATIC_CHORD
#define MAXNOTES 16
/*
 * This routine is used when we for some reason have no MIDI input.
 * It is crude - a better approach would be to write a MIDI-file reader.
 */
void makeChord (int key) {
  int notes[MAXNOTES];
  int n = 0;
  int i;
  int bass = rand () % 2;
  int b7th = rand () % 2;
  int b12th= rand () % 2;
  int minor= 1 - (rand () % 3);	/* minor, major, sus(pended) */
  if (bass) {
    notes[n++] = 36 + (key % 12);
  }
  if (b7th) {
    notes[n++] = 36 + 7 + (key % 12);
  }
  if (b12th) {
    notes[n++] = 36 + 12 + (key % 12);
  }
  if (!(bass || b7th || b12th) || (rand () % 2)) {
    notes[n++] = key;
    notes[n++] = key + 4 + minor;
    notes[n++] = key + 7;
  }
  if (rand () % 2) notes[n++] = key + 10 + (rand () % 2);
  if (rand () % 2) notes[n++] = key + 12;
  for (i = 0; i < n; i++) {
    oscNoteOn ((unsigned char) notes[i]);
  }
}

#endif /* STATIC_CHORD */

/*
 * Print rudimentary commandline syntax.
 */
static void Usage () {
  printf ("Syntax: [-r][c][-p pgmfile] [property=value ...]\n");
  printf ("Where:\n");
  printf ("-r         Randomize initial preset (whacky but true)\n");
  printf ("-c         Do not read the default configuration file '%s'\n",
	  defaultConfigFile);
  printf ("-p pgmfile Load alternate programme file over default.\n");
  printf ("property=value  Set/Override configuration property.\n");
}

/*
 * Main program.
 */
int main (int argc, char * argv []) {
  int i;
  int k;
  int state = 0;
  int doDefaultConfig = TRUE;
  char * configOverride [NOF_CFG_OVERS];
  size_t configOverEnd = 0;
  unsigned int randomPreset[9];
  unsigned int defaultPreset[9] = {8,6,8};
  unsigned int * presetSelect = defaultPreset;

  char * alternateProgrammeFile = NULL;

  srand ((unsigned int) time (NULL));

  for (i = 1; i < argc; i++) {
    char * av = argv[i];
    if (state == 0) {
      if (!strcmp (av, "-r")) {
	for (k = 0; k < 9; k++) {
	  randomPreset[k] = rand () % 9;
	}
	fprintf (stderr,
		 "%d%d %d%d%d%d %d%d%d\n",
		 randomPreset[0],
		 randomPreset[1],
		 randomPreset[2],
		 randomPreset[3],
		 randomPreset[4],
		 randomPreset[5],
		 randomPreset[6],
		 randomPreset[7],
		 randomPreset[8]);
	presetSelect = randomPreset;
      }
      else if (!strcmp (av, "-c")) {
	doDefaultConfig = FALSE;
      }
      else if (!strcmp (av, "-p")) {
	state = 1;
      }
      else if (strchr (av, '=') != NULL) {
	/* Remember this as a config parameter */
	if (configOverEnd < NOF_CFG_OVERS) {
	  configOverride[configOverEnd++] = av;
	}
	else {
	  fprintf (stderr,
		   "Too many configuration parameters (%d), please consider using a\n",
		   NOF_CFG_OVERS);
	  fprintf (stderr,
		   "configuration file instead of the commandline.\n");
	  return -1;
	}
      }
      else {
	fprintf (stderr, "Unknown argument : '%s'\n", av);
	Usage ();
	return -1;
      }
    }
    else if (state == 1) {
      alternateProgrammeFile = av;
      state = 0;
    }
  }

  /*
   * Commandline arguments are parsed. If we are of a mind to try the
   * default configuration file we do that now.
   */

  if (doDefaultConfig == TRUE) {
    if (access (defaultConfigFile, R_OK) == 0) {
      parseConfigurationFile (defaultConfigFile);
    }
  }

  /*
   * Then apply any configuration parameters collected from the commandline.
   * These must be applied last so that they can override the parameters
   * read from the files (if any).
   */

  for (i = 0; i < configOverEnd; i++) {
    parseConfigurationLine ("commandline argument",
			    0,
			    configOverride[i]);
  }

  /*
   * Having configured the initialization phase we can now actually do it.
   */

  initAll ();

  /*
   * We are initialized and now load the programme file.
   */

  loadProgrammeFile (defaultProgrammeFile);
  if (alternateProgrammeFile != NULL) {
    loadProgrammeFile (alternateProgrammeFile);
  }
  listProgrammes (stderr);
  setDisplayPgmChanges (TRUE);

#ifdef OFFLINE  /* SPECIAL OFF-LINE ANALYSIS */
  offLineAnalysis ();
  close (audio_fd);
  return 0;
#endif /* SPECIAL OFF-LINE ANALYSIS */

  /*
   * With the programmes eager and ready to go, we spawn off the MIDI
   * listener thread. The thread will initialize the MIDI device.
   */
#ifdef NO_MIDI
  k = 0;
#else
  k = pthread_create (& t_midi, NULL, MIDIInReader, NULL);
#endif /* NO_MIDI */

  if (k != 0) {
    fprintf (stderr, "%d : %s\n", k, "pthread_create : MIDIInReader thread");
  }
  else {
    unsigned char * obuf;

    setMIDINoteShift (0);
    setDrawBars (presetSelect);
    setSwell (1.0);
    setClean (TRUE);
    setVibrato (0);
    setRevSelect (WHIRL_SLOW);

#ifdef STATIC_CHORD
    set7bitGain ((unsigned char) 127);
#if 0
    {
      int notes[8];
      int i;
      notes[0] = 60-24;
      notes[1] = 67-24;
      notes[2] = 72-24;
      for (i = 1; i < 5; i++) {
	int clash = FALSE;
	do {
	  int j;
	  notes[i] = 32 + (rand () % 84);
	  for (j = 0; j < i; j++) {
	    if (notes[j] == notes[i]) {
	      clash = TRUE;
	      break;
	    }
	  }
	} while (clash);
      }
      
      for (i = 0; i < 5; i++) {
	oscNoteOn ((unsigned char) notes[i]);
      }
    }
#else
/*     makeChord (60 + (rand () % 24)); */
    oscNoteOn (60);
    oscNoteOn (60);
#endif
    setClean (FALSE);
    setRevSelect (WHIRL_SLOW);
#endif /* STATIC_CHORD */

    while (TRUE) {

      oscGenerateFragment (bufA, BUFFER_SIZE_SAMPLES);


      obuf = vibratoProc (bufA, bufB, BUFFER_SIZE_SAMPLES);
      obuf = preamp (obuf, bufB, BUFFER_SIZE_SAMPLES);

#if 1

      obuf = reverb (obuf, bufB, BUFFER_SIZE_SAMPLES);

      whirlProc (obuf, bufC, BUFFER_SIZE_SAMPLES);

#else
      /* This runs the reverb in parallel to the whirl. */
      whirlProc (obuf, bufC, BUFFER_SIZE_SAMPLES);
      obuf = reverb (obuf, bufB, BUFFER_SIZE_SAMPLES);
      {
	int i;
	unsigned short * sp = (unsigned short *) bufC;
	unsigned short * tp = (unsigned short *) obuf;
	for (i = 0; i < BUFFER_SIZE_SAMPLES; i++) {
	  *sp++ += (*tp);
	  *sp++ += (*tp++);
	}
      }
#endif

#if 0
      padBuffer (obuf, bufC, BUFFER_SIZE_SAMPLES);
#endif

      if (write (audio_fd, bufC, sizeof (bufC)) == -1) {
	perror ("write audio_fd bufC");
	exit (1);
      }
    }
  }

  close (audio_fd);

  return 0;
}
