/*
 * preamp.c --- Valve/tube distortion simulation.
 * 2003-02-25/FK Trying with new anti-imaging filters. Threw out a LOT of
 *               old and futile experiments. They are all there in
 *               preamp_orig.c if you care to have a look.
 *
 * 2002-07-22/FK Has tried several variants of a real-time polynomial:
 *               y=ax+bx+c, varying a and b from input level, output
 *               level and intermediate variants. Some sound good, other
 *               quite ugly.
 *
 * 2002-02-26/FK Found something which is close to the characteristic
 *               'chuffing', or distorted attack; whatever its called.
 *
 * 2002-02-21/FK Finally something that at least resembles an acceptable
 *               overdrive. A compression stage before the lookup table
 *               seemed to be the key. Some pumping effects to tune out.
 *
 * 2002-02-14/FK Mapping across the sine function appears to do almost exactly
 *               that which I tried to do with the spline, only simpler and
 *               better. Implemented a cross-fade between clean sound and
 *               the lookup table. 
 *
 * 2002-02-11/FK First entry, but the file is much older, as all the dead
 *               code can vouch for. Tonight I tried the pure sine function
 *               and it worked rather well, but still not as good as the
 *               Hermite spline. One good thing about the sine solution
 *               however, is that is has two simple but powerful parameters.
 *               I will try the Hermite spline with additional control points
 *               to avoid discontinuities around the origin.
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <limits.h>
#include <assert.h>

#include "main.h"
#include "config.h"

/* #define USE_WAVESHAPER */
/* #define USE_TUBEMODEL1 */
#define USE_TUBEMODEL2

/* #define SINGLE_TRIODE00 */
#define NAIVEPUSHPULL1

/* Polynom a temporary test routine */
/* #undef USE_TUBEMODEL2 */
/* #define USE_POLYNOM */

#ifdef SINGLE_TRIODE00
#define TUBEMODEL2_ALT2		/* Variant of feedback */
#endif

#define MIN_MACRO_N 1
#define MAX_MACRO_N 9
#define DEFAULT_WSHAPEMACRO 6

//#define CLIP_DETECT

#define TBLSZ_SHORTS 65536	/* Size of the waveshape table */

#ifndef TMID
#define TMID 32768	/* Signed position 0 in the waveshape table */
#endif

typedef struct _splineSegment {
  double P1x, P1y;
  double R1x, R1y;
  double P4x, P4y;
  double R4x, R4y;
} SplineSegment;

/*
 * 26-mar-2003/FK
 * Tube model 1 was the first implementation of the cloud model. It
 * uses a push-pull arrangement and takes its output from the tap.
 * Quite nice shaping of the output, except that at higher frequencies
 * the assymetry developed into a sharp peak that appears both atypical
 * and too sharp.
 */

#ifdef USE_TUBEMODEL1 
static float anode0 = 1.0;	/* Electrons ejected on each sample */
static float anode1 = 1.0;
static float leakCtrl0 = 0.02;	/* Electrons 'leaking away' on each sample */
static float leakCtrl1 = 0.02;
static float cloud0 = 0.0;	/* Pressure of electron 'cloud' */
static float cloud1 = 0.0;
static float pressure0 = 0.0;	/* Transfer variable: previous cloud */
static float pressure1 = 0.0;
static float outputGain = 32768.0 * 0.6;
#endif /* USE_TUBEMODEL1 */

/*
 * 29-mar-2003/FK
 * Tube model 2 was the second version of the cloud model. It simulates
 * a single tube (not push-pull) and takes its output from the cloud
 * rather than the tap. This gives true amplification and true clipping
 * although it has no linear range at all. What it does have, however,
 * is less intermodulation at high overdrive, a feature that I recognize
 * as a typical tube amp property.
 */

#ifdef USE_TUBEMODEL2
#ifdef SINGLE_TRIODE00
static float anode     = 0.15;
static float leakCtrl  = 0.01;
static float feedback  = 0.01;
static float cloud     = 0.0;
static float pressure  = 0.0;
#ifdef TUBEMODEL2_ALT2
static float exitBias  = -2.734;
static float inputAttn = 0.16 / 32768.0;
#else
static float exitBias  = -2.0;
static float inputAttn = 0.5 / 32768.0;
#endif
static float outputGain = 32768 * 0.24;
#endif /* SINGLE_TRIODE00 */

#ifdef NAIVEPUSHPULL1
#if 1
static float inputAttn    = 1.0 / 32768.0;
static float anodeA       = 0.1495;
static float leakcontrolA = 0.01;
static float feedbackA    = 0.01;
static float anodeB       = 0.15;
static float leakcontrolB = 0.012;
static float feedbackB    = 0.01;
static float offsetCompensation = -0.1666666666;
static float outputGain = 32767.0 / 3.8;
#else
static float inputAttn    = 0.7 / 32768.0;
static float anodeA       = 0.5;
static float leakcontrolA = 0.01;
static float feedbackA    = 0.11;
static float anodeB       = 0.5;
static float leakcontrolB = 0.01;
static float feedbackB    = 0.1;
static float fcoff0       = -0.5;
static float fcoff1       = -1.0;
static float fxa;
static float fxb;
static float offsetCompensation = 0.0;
static float outputGain = 32767.0 / 1.0;
#endif

static float plateA = 0.0;
static float plpA   = 0.0;
static float plateB = 0.0;
static float plpB   = 0.0;

#endif /* NAIVEPUSHPULL1 */

#endif /* USE_TUBEMODEL2 */

static short ampOut [TBLSZ_SHORTS]; /* This should go */

/* --- Interpolating waveshaper 03-apr-2003/FK */

#define WAVESHAPE_SZ 10000

static float wshape[WAVESHAPE_SZ + 1];

enum {fc0, fc1, vgmin, vgmax, K_, mu, vgbreak, vpk, R_,
      xmin, xmax, np};

static double p[np];

static float outscl, maxf, minf;

/* --- */

/*
 * This table is used in wshapeMacro6.
 */
static double harmonicAmplitude[5] = {
  0.002,			/* f2 */
  0.005,			/* f3 */
  0.010,			/* f4 */
  0.020,			/* f5 */
  0.050				/* f6 */
};

static int graphStep = 8;

static int gain = 128;
static int antiGain = 127;
static int isClean = FALSE;

static int posAmp = 128;
static int negAmp = 128;
static int diff = 2;

static double minGain = 0.0;
static double maxGain = 1.0;
static int gainTable[128];	/* Map from MIDI 7-bit value to gain */
static int gainTableInitialized = FALSE;

static double expressionTable_0[128];
static double expressionTable_1[128];
static double exp_0 = 0.5;
static double exp_1 = 0.5;

static char * transferDumpFname = NULL;
static char   transferDumpFNBuffer [PATH_MAX + 1];

static int wshapeSelect = DEFAULT_WSHAPEMACRO;

static double m3_left  = -0.5;	/* wshapeMacro3() */
static double m3_right =  0.5;	/* wshapeMacro3() */

/* wshapeMacro9 - coefficients for the polynomial */

#define MAXPAR 6
static double par[MAXPAR] = {
   0.0,				/* C */
   0.35,			/* x^1 */
   0.0,				/* x^2 */
  -0.6,				/* x^3 */
   0.0,				/* x^4 */
   1.25				/* x^5 */
};

/*
 * This term scales the output values of the waveshape table.
 * It is a trim device, mainly.
 */
static double tableOutputScale = 0.9;

/*
 * Used in waveshape macro #8 to control the shape of the curve.
 * Configuration key : amp.whape.triode
 */
static double triodeCurve = 0.01;

/* Temporary */
static double saturation = 0.7;
static int ov = 0;
static int nm = 1024;

/* ================================================================ */

/*
 *
 */
static void computeGainTable () {
  double u = minGain * 256.0;
  double r = (maxGain * 256.0) - u;
  int i;

  for (i = 0; i < 128; i++) {
    double w = ((double) i) / 127.0; /* From 0 -- 1.0 */
    double g = minGain + (w * (maxGain - minGain));

    gainTable[i] = (int) rint (u + (w * r));

    expressionTable_0[i] = g * (1.0 - fabs (w - 0.5));
    expressionTable_1[i] = g * (w - 0.5);
  }

  gainTableInitialized = TRUE;
}

/*
 *
 */
void setMinGain (double g) {
  minGain = g;
  computeGainTable ();
}

/*
 *
 */
void setMaxGain (int g) {
  maxGain = g;
  computeGainTable ();
}

/*
 *
 */
void setSwell (double level) {
  double v = (level < 0.0) ? 0.0 : (1.0 < level) ? 1.0 : level;
  gain = (int) (v * 256.0);
}

/*
 * This function is called by the MIDI parser. The function maps a continous
 * controller to the gain. The received value is 0-127.
 */
void set7bitGain (unsigned char b) {

  gain = gainTable[b];

  exp_0 = expressionTable_0[b];
  exp_1 = expressionTable_1[b];

}

/*
 *
 */
void setClean (int useClean) {
  isClean = useClean;
}

/*
 * The overdrive function provides the maximum overdrive curve.
 * The actual curves used in the amplification tables are weighed
 * averages between the value of this curve and the unit amplification.
 * The curve is basically 1/n, flipped over, translated and scaled.
 * The argument, x, must be zero or positive to avoid division with zero.
 */
static double overdrive (double x) {
  static double k = 0.8; //1.0;
  static double d = 0.1;
  static double b = 0.1;
  return k - (d / (x + b));
}

static short ovrMellow (int x, double w) {
  static double UNIT = 32767.5;
  double e;
  double u;
  double r;
  double sign = (x < 0) ? -1.0 : 1.0;
  u = (sign * x) / UNIT;
  e = overdrive (u);
  r = (u - (u * u * w)) + (e * u * w);
  return (short) (sign * r * UNIT);
}

static short ovrSaturated (int x, double w) {
  static double UNIT = 32767.5;
  double e;
  double u;
  double r;
  double sign = (x < 0) ? -1.0 : 1.0;
  u = (sign * x) / UNIT;
  e = overdrive (u);
  r = (u * (1.0 - w)) + (e * w);
  return (short) (sign * r * UNIT);
}

static double Q (double x, double k) {
  return (x - k) / (1.0 - k);
}

static double F (double x, double k) {
  double v = 1.0 - Q(x, k);
  return v * v;
}

static short oversplit (int ix, double w, double k, double g) {
  static double UNIT = 32768.0;
  double sign = (ix < 0) ? -1.0 : 1.0;
  double x = (sign * ix) / UNIT;
  double r = (x < k) ? x * g : w - ((w - (k * g)) * F (x, k));
  return (short) (sign * r * UNIT);
}

static short waveshape1 (int ix, double a, double t, double g) {
  static double UNIT = 32768.0;
  double sign = (ix < 0) ? -1.0 : 1.0;
  double x = (sign * ix) / UNIT;
  double r;
  if (x < a) {
    r = x;
  }
  else if (x < t) {
    double b = (x - a) / (1.0 - a);
    r = a + ((x - a) / (1.0 + (b * b)));
  }
  else {
    r = (a + 1) / 2.0;
  }
  return (short) (sign * r * g * UNIT);
}

/*
 * @param x  Input sample value : -1..1
 * @param s  Saturation level   :  0..1
 */
static double topfold (double x, double s) {
  double sign = (x < 0) ? -1.0 : 1.0;
  double u = sign * x;
  double e = (u - s) / (1.0 - s);

  return sign * ((u < s) ? u : s + ((u - s) / (1.0 + (e * e))));
}

/*
 * This function represents a spline in Hermite form. It has two points, 
 * P1 at (0, y) and P4 at (1, y). In each point, a tangent vector indicates
 * the direction in which the transfer function curve passes through the
 * point. The tangent vectors are R1 (1, y) and R4 (1, y). When the points
 * and their tangent vectors are restricted in this way, the x input value
 * equals the x output value of the spline (x(t)=t), and we can use it
 * as a waveshaper. Usually, a spline's blending functions specifies
 * x, y and z for the unit interval (0--1).
 * You can think of the parameter R1y as the gain near the low input end,
 * and R4y as the gain near the high input end. Similarly, P1y is the output
 * level when the input is zero, and P4y the output level when the input
 * is 1.
 * Be careful with the parameters, since the curve can easily escape the
 * 1 by 1 unit box in which case the result overflows the representation.
 *
 * If you have no idea what the above means and just want to tinker with
 * the settings, keep P1y=0, P4y in {0.5 -- 0.9}, R1y in {0 -- 3} and
 * R4y in {-3 -- 3}.
 * Examples:
 * Symmetric +-: 0.0, 3.0, 0.5, 0.0;   A bit over
 * Assym. +: 0, 3, 0.5, 0; -:0, 2, 5, 0;  More gritty edge
 * Sym. +-:0.0, 6.0, 0.6, 2.0; Quite nasty, very high gain.
 * Sym. +-:0.0, 4, 0, 4; Weird curve, but a smoother sound.
 * ASym. +-:0.0, 3.7|4, 0, 4; Slightly harsher, I think.
 * Sym : 0, -1, 0, -6: Broken transistor amplifier.
 * Sym : 0, 2, 0, -5: Good amplification, almost no dist.
 */
static short HSpline (int ix, double P1y, double R1y, double P4y, double R4y) {
  static double UNIT = 32768.0;
  double sign = (ix < 0) ? -1.0 : 1.0;
  double t = (sign * ix) / UNIT;
  double tSq = t * t;		/* t^2 */
  double tCb = tSq * t;		/* t^3 */
  double r;

  r = P1y * ( 2.0 * tCb - 3.0 * tSq + 1.0)
    + P4y * (-2.0 * tCb + 3.0 * tSq)
    + R1y * (tCb - 2.0 * tSq + t)
    + R4y * (tCb - tSq);

  return (sign * r * UNIT);
}

/*
 *
 */
static double HSplineD (int ix, double P1y, double R1y, double P4y, double R4y)
{
  static double UNIT = 32768.0;
  double sign = (ix < 0) ? -1.0 : 1.0;
  double t = (sign * ix) / UNIT;
  double tSq = t * t;		/* t^2 */
  double tCb = tSq * t;		/* t^3 */
  double r;

  r = P1y * ( 2.0 * tCb - 3.0 * tSq + 1.0)
    + P4y * (-2.0 * tCb + 3.0 * tSq)
    + R1y * (tCb - 2.0 * tSq + t)
    + R4y * (tCb - tSq);

  return (sign * r);
}

/*
 *
 */
static double HSplineSegmentD (double t, SplineSegment * seg, double * x) {
  double tSq = t * t;		/* t^2 */
  double tCb = tSq * t;		/* t^3 */
  double r;
  double s;

  r = seg->P1y * ( 2.0 * tCb - 3.0 * tSq + 1.0)
    + seg->P4y * (-2.0 * tCb + 3.0 * tSq)
    + seg->R1y * (tCb - 2.0 * tSq + t)
    + seg->R4y * (tCb - tSq);

  s = seg->P1x * ( 2.0 * tCb - 3.0 * tSq + 1.0)
    + seg->P4x * (-2.0 * tCb + 3.0 * tSq)
    + seg->R1x * (tCb - 2.0 * tSq + t)
    + seg->R4x * (tCb - tSq);

  if (x != NULL) *x = s;

  return r;
}

/*
 *
 */
static void applyHermiteSpline (SplineSegment segs[], size_t nSegs) {
  static double UNIT = 32768.0;	
  double span;
  int nofSampleValues;
  int firstSampleValue;
  int lastSampleValue;
  int i;
  int j;

  fprintf (stderr, "\n");

  for (i = 0; i < nSegs; i++) {
    /* Compute the mapping range of the i/x component. */
    span = segs[i].P4x - segs[i].P1x;
    /* Fix the first and last sample values in range. */
    firstSampleValue = (int) (UNIT * segs[i].P1x);
    if (firstSampleValue < -32768) {
      fprintf (stderr,
	       "applyHermiteSpline: WARNING:firstSampleValue = %d\n",
	       firstSampleValue);
      firstSampleValue = -32768;
    }

    lastSampleValue = (int) (UNIT * segs[i].P4x);

    if (32767 < lastSampleValue) {
      fprintf (stderr,
	       "applyHermiteSpline: WARNING:lastSampleValue = %d\n",
	       lastSampleValue);
      lastSampleValue = 32767;
    }

    nofSampleValues = lastSampleValue - firstSampleValue;

    fprintf (stderr,
	     "SEGMENT %d: span=%g, spls=%d, 1st=%d lst=%d\n",
	     i,
	     span,
	     nofSampleValues,
	     firstSampleValue,
	     lastSampleValue);


    /* Traverse the sample input values mapped by this segment of spline. */
    for (j = firstSampleValue; j < lastSampleValue; j++) {
      /* Transform onto the segment's range of 0.0 -- 1.0 */
      double u = ((double) (j - firstSampleValue));
      double t = u / (UNIT * span);
      double L;
      double x;
      double up = 1.0;
      double dn = 0.0;
      int k;
      assert (0.0 <= t);
      assert (t <= 1.0);
      /* Compute the y (output) value at this point. */
      /* Unfortunately we must do a binary search for the value of t. */
      do {
	L = HSplineSegmentD (t, &(segs[i]), & x);
	k = (int) (x * UNIT);
	if (j != k) {

	  if (!(dn < up)) {
	    fprintf (stderr,
		     "dn=%g, up=%g, t=%g, x=%g, j=%d, k=%d\n",
		     dn, up, t, x, j, k);
	    assert (0);
	  }

	  if (k < j) {		/* t is too low */
	    dn = t;
	    t = (up + dn) / 2.0;
	  }
	  else {		/* t is too high */
	    up = t;
	    t = (up + dn) / 2.0;
	  }
	}
      } while (j != k);

      if ((L < -1.0) || (1.0 < L)) {
	fprintf (stderr,
		 "applyHermiteSpline:WARNING:level out of range:%g\n",
		 L);
	L = (L < 0.0) ? -1.0 : 1.0;
      }

      ampOut[32768 + j] = (short) (32767.0 * L);

      if ((32768 + j) == 19660) {
	fprintf (stderr,
		 "L=%g at t=%g, j=%d, seg=%d, 1st=%d, lst=%d\n",
		 L, t, j+32768, i, firstSampleValue, lastSampleValue);
      }

    }
  }

  {
    int k;
    double L;
    k = nSegs - 1;
    j = (int) rint (UNIT * segs[k].P4x);
    L = rint (UNIT * segs[k].P4y);
    j = (j < -32768) ? -32768 : ((32767 < j) ? 32767 : j);
    L = (L < -1.0) ? -1.0 : ((1.0 < L) ? 1.0 : L);
    ampOut[32768 + j] = (short) (32767.0 * L);
  }

}

/*
 *
 */
static int dumpXlate (int i) {
  double v;
  v = (((double) i) / 65536.0) * M_PI * 2.0;
  return ampOut[32768 + (int) (32767.0 * sin (v))];
}

/*
 * Dump a graphical rendering of the ampOut array of the given file.
 */
static void dumpAmpOut (char *fname, char * title) {
  static double UNIT = 32768.0;
  FILE * fp;
  int i;
  int j;

  if ((fp = fopen (fname, "w")) != NULL) {
    fprintf (fp, "%%!PS\n");
    fprintf (fp, "/pt {lineto} def\n");
    fprintf (fp, "/Helvetica findfont 18 scalefont setfont\n");
    if (title != NULL) {
      fprintf (fp, "50 50 moveto (%s) show\n", title);
    }
    fprintf (fp, "gsave\n");
    fprintf (fp, "100 200 translate\n");
    fprintf (fp, "400.0 65536.0 div dup scale\n");
    fprintf (fp, "32767 0 moveto 32767 65537 lineto\n");
    fprintf (fp, "0 32767 moveto 65536 32767 lineto\n");
    fprintf (fp, "0 0 moveto\n");
    fprintf (fp, "0 65536 lineto\n");
    fprintf (fp, "65536 dup lineto\n");
    fprintf (fp, "65536 0 lineto\n");
    fprintf (fp, "closepath stroke\n");

    /* Paint the canvas from the table. */
    fprintf (fp, "%d %d moveto\n", 0, ampOut[0] + 32768);
    for (i = 1; i < TBLSZ_SHORTS; i += graphStep) {
      fprintf (fp, "%d %d pt\n", i, ampOut[i] + 32768);
    }
    fprintf (fp, "stroke\n");

    for (i = 0; i < TBLSZ_SHORTS; i += 8096) {
      fprintf (fp, "%d %d moveto ", i, 0);
      fprintf (fp, "%d %d lineto ", i, ampOut[i] + 32768);
      fprintf (fp, "%d %d lineto\n", 65536, ampOut[i] + 32768);
    }
    fprintf (fp, "stroke\n");
    fprintf (fp, "grestore\n");

    fprintf (fp, "100 400 translate\n");
    fprintf (fp, "200.0 65536.0 div dup scale\n");
    fprintf (fp, "32767 0 moveto 32767 65537 lineto\n");
    fprintf (fp, "0 32767 moveto 65536 32767 lineto\n");
    fprintf (fp, "0 0 moveto\n");
    fprintf (fp, "0 65536 lineto\n");
    fprintf (fp, "65536 dup lineto\n");
    fprintf (fp, "65536 0 lineto\n");
    fprintf (fp, "closepath stroke\n");

    fprintf (fp, "%d %d moveto\n", 0, dumpXlate (0) + 32768);
    for (i = 1; i < TBLSZ_SHORTS; i += graphStep) {
      fprintf (fp, "%d %d pt\n", i, dumpXlate (i) + 32768);
    }
    fprintf (fp, "stroke\n");
    fprintf (fp, "grestore\n");

    fclose (fp);
  }
  else {
    perror (fname);
  }
}

/*
 *
 */
int ampConfig (ConfigContext * cfg) {
  int ack = 0;
  if (strcasecmp (cfg->name, "amp.gain.min") == 0) {
    double g;
    ack++;
    if (sscanf (cfg->value, "%lf", &g) == 1) {
      setMinGain (g);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "amp.gain.max") == 0) {
    double g;
    ack++;
    if (sscanf (cfg->value, "%lf", &g) == 1) {
      setMaxGain (g);
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strncasecmp (cfg->name, "amp.gain.b.", 11) == 0) {
    int i;
    double g;
    ack++;
    if (sscanf (cfg->name, "amp.gain.b.%d", &i) == 1) {
      if ((0 <= i) && (i <= 127)) {
	if (sscanf (cfg->value, "%lf", &g) == 1) {
	  gainTable[i] = (int) rint (g * 256.0);
	  gainTableInitialized = TRUE;
	}
	else {
	  configDoubleUnparsable (cfg);
	}
      }
      else {
	configIntOutOfRange (cfg, 0, 127);
      }
    }
  }
  else if (strcasecmp (cfg->name, "amp.wshape.psdump") == 0) {
    ack++;
    strncpy (transferDumpFNBuffer, cfg->value, PATH_MAX);
    transferDumpFname = transferDumpFNBuffer;
  }
  else if (strcasecmp (cfg->name, "amp.wshape.macro") == 0) {
    int m;
    ack++;
    if (sscanf (cfg->value, "%d", &m) == 1) {
      if ((MIN_MACRO_N <= m) && (m <= MAX_MACRO_N)) {
	wshapeSelect = m;
      }
      else {
	configIntOutOfRange (cfg, MIN_MACRO_N, MAX_MACRO_N);
      }
    }
    else {
      configIntUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "amp.wshape.sinrg.left") == 0) {
    double d;
    ack++;
    if (sscanf (cfg->value, "%lf", &d) == 1) {
      m3_left = d;
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "amp.wshape.sinrg.right") == 0) {
    double d;
    ack++;
    if (sscanf (cfg->value, "%lf", &d) == 1) {
      m3_right = d;
    }
    else {
      configDoubleUnparsable (cfg);
    }
  }
  else if (strcasecmp (cfg->name, "amp.wshape.triode") == 0) {
    ack++;
    setConfigDouble (&triodeCurve, cfg);
  }
  else if (strcasecmp (cfg->name, "amp.wshape.graphstep") == 0) {
    ack++;
    setConfigInt (&graphStep, cfg);
  }
  else if (strcasecmp (cfg->name, "amp.wshape.f2") == 0) {
    ack++;
    setConfigDouble (&(harmonicAmplitude[0]), cfg);
  }
  else if (strcasecmp (cfg->name, "amp.wshape.f3") == 0) {
    ack++;
    setConfigDouble (&(harmonicAmplitude[1]), cfg);
  }
  else if (strcasecmp (cfg->name, "amp.wshape.f4") == 0) {
    ack++;
    setConfigDouble (&(harmonicAmplitude[2]), cfg);
  }
  else if (strcasecmp (cfg->name, "amp.wshape.f5") == 0) {
    ack++;
    setConfigDouble (&(harmonicAmplitude[3]), cfg);
  }
  else if (strcasecmp (cfg->name, "amp.wshape.f6") == 0) {
    ack++;
    setConfigDouble (&(harmonicAmplitude[4]), cfg);
  }
  else if (strncasecmp (cfg->name, "amp.wshape.par.x", 16) == 0) {
    int index = cfg->name[16] - '0';
    if ((0 <= index) && (index < MAXPAR)) {
      ack++;
      setConfigDouble (&(par[index]), cfg);
    }
    else {
      configIntOutOfRange (cfg, 0, MAXPAR);
    }
  }

#ifdef COMMENT
  else if (strncasecmp (cfg->name, "amp.tube.anode", 14) == 0) {
    int c = cfg->name[15];
    double v;
    setConfigDouble (&v, cfg);
    if (c == '\0') {
#ifdef USE_TUBEMODEL1
      anode0 = anode1 = v;
#endif
#ifdef USE_TUBEMODEL2
      anode = v;
#endif
      ack++;
    }
    else if (c == '-') {
#ifdef USE_TUBEMODEL1
      anode0 = v;
#endif
      ack++;
    }
    else if (c == '+') {
#ifdef USE_TUBEMODEL1
      anode1 = v;
#endif
      ack++;
    }
  }
#endif /* COMMENT */

#ifdef COMMENT
  else if (strncasecmp (cfg->name, "amp.tube.leak", 13) == 0) {
    int c = cfg->name[14];
    double v;
    setConfigDouble (&v, cfg);
    if (c == '\0') {
#ifdef USE_TUBEMODEL1
      leakCtrl0 = leakCtrl1 = v;
#endif
#ifdef USE_TUBEMODEL2
      leakCtrl = v;
#endif
      ack++;
    }
    else if (c == '-') {
#ifdef USE_TUBEMODEL1
      leakCtrl0 = v;
#endif
      ack++;
    }
    else if (c == '+') {
#ifdef USE_TUBEMODEL1
      leakCtrl1 = v;
#endif
      ack++;
    }
  }
#endif /* COMMENT */

#ifdef COMMENT
#ifdef USE_TUBEMODEL2
  else if (strcasecmp (cfg->name, "amp.tube.feedback") == 0) {
    double v;
    setConfigDouble (&v, cfg);
    feedback = (float) v;
    ack++;
  }
#endif /* USE_TUBEMODEL2 */
#endif /* COMMENT */

  return ack;
}

/*
 *
 */
static void wshapeMacro1 () {
  SplineSegment segs[4];
  double m = 0.1;
  double n = 0.0;
  double k = 0.5;

  segs[0].P1x = -1.0;
  segs[0].P1y = -0.5;
  segs[0].R1x =  1.1 * m;
  segs[0].R1y =  0.2 * m;

  segs[0].P4x = segs[1].P1x = -0.1 * n;
  segs[0].P4y = segs[1].P1y = -0.3 * n;
  segs[0].R4x = segs[1].R1x =  1.0 * k;
  segs[0].R4y = segs[1].R1y =  3.0 * k;

  segs[1].P4x = segs[2].P1x =  0.1 * n;
  segs[1].P4y = segs[2].P1y =  0.3 * n;
  segs[1].R4x = segs[2].R1x =  1.0 * k;
  segs[1].R4y = segs[2].R1y =  3.0 * k;

  segs[2].P4x = 1.0;
  segs[2].P4y = 0.5;
  segs[2].R4x = 1.1 * m;
  segs[2].R4y = 0.2 * m;

  applyHermiteSpline (segs, 3);
}

/*
 *
 */
static void wshapeMacro2 () {
  SplineSegment segs[5];
  segs[0].P1x = -1.0;
  segs[0].P1y = -1.0;
  segs[0].R1x =  0.50;
  segs[0].R1y =  0.10;

  segs[0].P4x = segs[1].P1x = -0.5;
  segs[0].P4y = segs[1].P1y = -0.8;
  segs[0].R4x = segs[1].R1x =  0.3;
  segs[0].R4y = segs[1].R1y =  0.7;

  segs[1].P4x = segs[2].P1x = -0.2;
  segs[1].P4y = segs[2].P1y = -0.1;
  segs[1].R4x = segs[2].R1x =  0.4;
  segs[1].R4y = segs[2].R1y =  0.2;

  segs[2].P4x = segs[3].P1x =  0.2;
  segs[2].P4y = segs[3].P1y =  0.1;
  segs[2].R4x = segs[3].R1x =  0.3;
  segs[2].R4y = segs[3].R1y =  0.7;

  segs[3].P4x = segs[4].P1x =  0.5;
  segs[3].P4y = segs[4].P1y =  0.8;
  segs[3].R4x = segs[4].R1x =  0.0;
  segs[3].R4y = segs[4].R1y =  0.2;

  segs[4].P4x = 1.0;
  segs[4].P4y = 1.0;
  segs[4].R4x = 0.50;
  segs[4].R4y = 0.0;

  applyHermiteSpline (segs, 5);
}

/*
 * In this attempt the input is mapped onto a fragment of a sine curve.
 * The fragment begins in x = a and ends in x = b. A complete cycle
 * is thus represented by a = -pi/2 and b = pi/2. 
 */
static void wshapeMacro3 () {
  int i;
  double a = M_PI * m3_left;
  double b = M_PI * m3_right;

  for (i = -32768; i < 32768; i++) {
    double v = ((a + b) / 2.0) + ((((double) i) / 32767.0) * ((b - a) / 2.0));
    ampOut[32768 + i] = (short) (32767.0 * sin (v));
  }
}

/*
 *
 */
static void wshapeMacro4 () {
  int i;
  for (i = -32768; i < 32768; i++) {
    /* This one is really good and its been in for a long time. */
    if (i < 0) {
      ampOut[32768 + i] = HSpline (i, 0.0, 2.0, 0.0, -2.0);
    }
    else {
      ampOut[32768 + i] = HSpline (i, 0.0, 2.0, 0.7, -2.0);
    }
  }
}

/*
 *
 */
static void wshapeMacro5 () {
  int i;
  for (i = -32768; i < 32768; i++) {
    /* Experimental settings */
    /* The min value for a is -2PI, the max for b is 2PI. */
    static double a = M_PI * -0.8;
    static double b = M_PI *  0.8;
    double ab_2 = (a + b) / 2.0;
    double ba_2 = (b - a) / 2.0;
    static double UNIT = 32767.0;

    double u = ((double) i) / UNIT;

    double splLvl = 0.0;	/* Spline shaper level */

    double sinLvlP = 1.0;	/* Positive sine shaper level  */
    double sinLvlN = 1.0;	/* Negative sine shaper level */
    double sinwP = 0.30;	/* Positive sine weight */
    double sinwN = 0.15;	/* Negative sine weight */

    double tpfLvl = 0.0;	/* Topfold mix level */
    double tpfLvP = 0.5;	/* Topfold positive threshold level */
    double tpfLvN = 0.3;	/* Topfold negative threshold level */
    double tpfPrG = 2.0;	/* Topfold pre-gain */

    /*
     * This expression maps the range of i (-32768..32767) onto the
     * range a..b (as defined above). The output value v is the argument
     * to sin(). 
     */
    // v = ((a + b) / 2.0) + ((((double) i) / 32767.0) * ((b - a) / 2.0));
    double v = ab_2 + (u * ba_2);

    if (i < 0) {		/* NEGATIVE SIDE */
      ampOut[32768 + i] =  (short)
	(UNIT *
	 ((splLvl * HSplineD (i, 0.0, 1.0, 1.0, 1.0)) +
	  (sinLvlN * (((1.0 - sinwN) * u) + (sinwN * sin (v)))) +
	  (tpfLvl * topfold (u * tpfPrG, tpfLvN))));
    }
    else {			/* POSITIVE SIDE */
      ampOut[32768 + i] = (short)
	(UNIT *
	 ((splLvl * HSplineD (i, 0.0, 1.0, 1.0, 1.0)) +
	  (sinLvlP * (((1.0 - sinwP) * u) + (sinwP * sin (v)))) +
	  (tpfLvl * topfold (u * tpfPrG, tpfLvP))));
    }

  }  
}

/*
 *
 */
static void wshapeMacro6 () {
  int i;
  double U = 32767.0;
  double a = M_PI * -0.5;
  double b = M_PI *  0.5;
  double ab_2 = (a + b) / 2.0;
  double ba_2 = (b - a) / 2.0;
  double f1;
  double f2 = harmonicAmplitude[0];
  double f3 = harmonicAmplitude[1];
  double f4 = harmonicAmplitude[2];
  double f5 = harmonicAmplitude[3];
  double f6 = harmonicAmplitude[4];

  double M = f2 + f3 + f4 + f5 + f6;
  f1 = 1.0 - M;

  for (i = -32768; i < 32768; i++) {
    double u = ((double) i) / U;
    double v = ab_2 + (u * ba_2);

    ampOut[32768 + i] =
      (short) (U * ((f1 * sin (v * 1.0)) +
		    (f2 * cos (v * 2.0)) +
		    (f3 * sin (v * 3.0)) +
		    (f4 * cos (v * 4.0)) +
		    (f5 * sin (v * 5.0)) +
		    (f6 * cos (v * 6.0))));
  }

}

/*
 * Same as 6 but the harmonics increase towards the upper input.
 */
static void wshapeMacro7 () {
  int i;
  double U = 32768.0;
  double a = M_PI * -0.5;
  double b = M_PI *  0.5;
  double ab_2 = (a + b) / 2.0;
  double ba_2 = (b - a) / 2.0;

  for (i = -32768; i < 32768; i++) {
    double u = ((double) i) / U;
    double v = ab_2 + (u * ba_2);
    double w = (u < 0.0) ? -u : u;

    double f2 = w * harmonicAmplitude[0];
    double f3 = w * harmonicAmplitude[1];
    double f4 = w * harmonicAmplitude[2];
    double f5 = w * harmonicAmplitude[3];
    double f6 = w * harmonicAmplitude[4];

    double M = f2 + f3 + f4 + f5 + f6;
    double f1 = 1.0 - M;

    ampOut[32768 + i] =
      (short) (U * tableOutputScale *
	       ((f1 * sin (v * 1.0)) +
		(f2 * cos (v * 2.0)) +
		(f3 * sin (v * 3.0)) +
		(f4 * cos (v * 4.0)) +
		(f5 * sin (v * 5.0)) +
		(f6 * cos (v * 6.0))));
  }

}

/*
 * This is based on the 'classic' triode model.
 */
static void wshapeMacro8 () {
  int i;
  double norm = 32768.0;
  double K = 8.0;
  double u = 2.0;		/* Actually mu */
  double Vgk = -58.0;		/* Volts between gate and kathode (input) */
  double Vpk = 130.0;		/* Volts between plate and kathode */
  /*  double r = 0.01;		/* Curve shaper: tweak this */
  double r = triodeCurve;	/* Curve shaper: tweak this */
  double Ip;			/* Output */
  double offset = pow (K * (u * Vgk + Vpk), 1.5); /* y = f(0) */

  for (i = 0; i < 32768; i++) {
    double y;
    double x = ((double) i) / norm;
    double v = Vgk * pow (r, x);
    Ip = pow (K * (u * v + Vpk), 1.5);
    y = Ip - offset;
    ampOut[32768 + i] = (short) ( y * tableOutputScale);
    ampOut[32767 - i] = (short) (-y * tableOutputScale);
  }
  
}

/*
 * Fifth degree polynomial y = ax^5+bx^4+cx+dx+ex+f
 * Net lore has it that this cannot produce harmonics above the degree
 * of the polynomial, so there is a known limit to the oversampling.
 * 25-feb-2003/FK True, but it does not sound quite, eh, like IT.
 */
static void wshapeMacro9 () {
  int i;
  double norm = 32768.0;

  for (i = 0; i < 65536; i++) {
    double x = ((double) (i - 32768)) / norm;
    double x2 = x  * x;
    double x3 = x2 * x;
    double x4 = x2 * x2;
    double x5 = x2 * x3;
    double y =
      (par[5] * x5) +
      (par[4] * x4) +
      (par[3] * x3) +
      (par[2] * x2) +
      (par[1] * x ) +
      par[0];
    ampOut[i] = (short) (y * norm);
  }
}

/*
 * loadWaveShaper -- This routine initializes the floating point, inter-
 * polating waveshaper. 03-apr-2003/FK
 */

static void loadWaveShaper () {
  int i;
  double xMin  = p[xmin];
  double xMax  = p[xmax];
  double VgMin = p[vgmin];
  double VgMax = p[vgmax];
  double VgScl = (xMax - xMin) / (VgMax - VgMin);
  double K     = p[K_] * 0.000001;
  double u     = p[mu];
  double VgBreak = p[vgbreak];
  double Vpk   = p[vpk];
  double r     = p[R_];
  int    initp = 0;

  double VgBp = VgMin + (VgBreak * (VgMax - VgMin));
  double IpMax = pow (K * ((u * VgBp) + Vpk), 1.5);

  for (i = 0; i < WAVESHAPE_SZ; i++) {
    double y;
    double x = ((((double) i / (double) WAVESHAPE_SZ)) * 2.0) - 1.0;
    double Vg = (VgScl * (x - xMin)) + VgMin;
    double Ip = pow (K * ((u * Vg) + Vpk), 1.5);

#if 1
    double xx = (double) i / (double) WAVESHAPE_SZ;
    double v = 1.0 - (r / pow (r, xx));
    Ip = (v * Ip) + ((1.0 - v) * IpMax);
#endif

    wshape[i] = Ip;

    if (!initp) {
      minf = maxf = Ip;
      initp = 1;
    }
    else {
      if (Ip < minf) minf = Ip;
      if (maxf < Ip) maxf = Ip;
    }
  }

  outscl = 2.0 / (maxf - minf);	/* Output scaling */

  wshape[WAVESHAPE_SZ] = wshape[WAVESHAPE_SZ - 1]; /* Overrun protect. */

  /*  printf ("\noutscl=%g, minf=%g maxf=%g\n", outscl, minf, maxf); */
}

/*
 * This routine initializes this module.
 */
void initPreamp () {
  char title [80];

  if (gainTableInitialized == FALSE) {
    computeGainTable ();
  }

  switch (wshapeSelect) {
  case 1: wshapeMacro1 (); break;
  case 2: wshapeMacro2 (); break;
  case 3: wshapeMacro3 (); break;
  case 4: wshapeMacro4 (); break;
  case 5: wshapeMacro5 (); break;
  case 6: wshapeMacro6 (); break;
  case 7: wshapeMacro7 (); break;
  case 8: wshapeMacro8 (); break;
  case 9: wshapeMacro9 (); break;
  }

  if (transferDumpFname != NULL) {
    sprintf (title, "Waveshape Macro #%d", wshapeSelect);
    dumpAmpOut (transferDumpFname, title);
  }

  /* TEMPORARY */
  ov = (int) (saturation * 1024.0);
  nm = 1024 - ov;

  /* --- Interpolating waveshaper --- */

  p[fc0]     =   1.0;
  p[fc1]     =   0.0;
  p[vgmin]   =  -2.0;
  p[vgmax]   =   0.25;
  p[K_]     = 1970.0;
  p[mu]      =  79.0;
  p[vgbreak] =   1.0;
  p[vpk]     = 160.0;
  p[R_]      =   0.01;
  p[xmin]    =  -1.0;
  p[xmax]    =   1.0;
  loadWaveShaper ();
}

/* ========================================================================
 * 17 feb 2003/FK
 * Now we try our new-won filter design knowledge. This time we go for
 * a six-stage biquad cascade and six times oversampling with float type.
 */
#ifdef COMMENT

#define BQSTAGES 6		/* The nof biquad stages (sections) */

#define AWSIZE  (3 * BQSTAGES)	/* The nof a[] weigths needed */
#define ZSIZE   (2 * BQSTAGES)	/* The nof z[] state vars needed */
#define VSIZE   BQSTAGES	/* The nof intermediate values needed */

#define TIMESOVER   6
#define TIMESOVER_f 6.0

unsigned char * XXX_preamp (unsigned char * inBuf,
			unsigned char * outBuf,
			size_t bufLengthSamples) {
  static float pedal = 0;
  static float x_1 = 0;		/* Previous input */
  static float preNormalization = 500.0 /* 32768.0 */ ;
  static float restorationGain  = 32767.0 * 250.0;
  
  int i;

  short * xsp = (short *) inBuf;
  short * ysp = (short *) outBuf;

  float xs[TIMESOVER];		/* Interpolated samples */

  static float a[AWSIZE] = {
    /* gain, a1, a2 */
    0.002627, 1.840820, -0.851328, /* section 0 */
    0.010981, 1.817990, -0.861919, /* section 1 */
    0.025346, 1.780600, -0.881985, /* section 2 */
    0.041871, 1.742220, -0.909707, /* section 3 */
    0.056381, 1.717580, -0.943104, /* section 4 */
    0.065381, 1.718860, -0.980389  /* section 5 */
  };

  static float z[ZSIZE];	/* Delay state */

  if (isClean) {
    for (i = 0; i < bufLengthSamples; i++) {
      *ysp++ = ((*xsp++) * gain) >> 8;
    }
    return outBuf;
  }

  for (i = 0; i < bufLengthSamples; i++) {
    int j;
    float x0 = (float) (((*xsp++) * gain) >> 8); /* Obtain next input sample */
    float xd = (x0 - x_1) / (float) TIMESOVER; /* Delta between oversamples */
    float x = x_1;
    float r;

    x_1 = x0;			/* Update previous sample value */

    /* Interpolate samples */
    for (j = 0; j < TIMESOVER; j++, x += xd) {
      int k = TMID + (int) x;
      k &= 0xFFFF;
      xs[j] = ((float) ampOut[k]) * preNormalization;
    }

    /* Filter */
    for (j = 0; j < TIMESOVER; j++) {
      int k;			/* stage index */
      float * ap = a;
      float * zp = z;
      float v;
      r = xs[j];
      for (k = 0; k < BQSTAGES; k++, ap += 3, zp += 2) {
	v = (ap[0] * r) + (ap[1] * zp[0]) + (ap[2] * zp[1]);
	r = v + zp[0] + zp[0] + zp[1];
      }
    }

    *ysp++ = (short) (r * restorationGain);
  }

  return outBuf;
}
#endif /* COMMENT */

/* ========================================================================
 * 25-feb-2003/FK Well, here we go again; a new attempt back at 3x over-
 * sampling, a FIR filter and a fifth-grade polynom.
 */

#define OVERDLYSZ 2048
#define KERNELSZ  33
#ifdef TIMESOVER
#undef TIMESOVER
#endif
#define TIMESOVER 3

#ifdef USE_WAVESHAPER
unsigned char * preamp (unsigned char * inBuf,
			unsigned char * outBuf,
			size_t bufLengthSamples)
{
  /* Past input samples */

  static float x_1 = 0.0;	/* Previous 22 KHz input sample */
  static float dly[OVERDLYSZ];	/* Previous 66 KHz input samples */
  static int L = 0;		/* Position in dly[] */

  /* Weights for the anti-imaging filter. */

  static float w[KERNELSZ] = {
    1.41996e-19,
    6.17861e-06,
    0.000215766,
    0.000829305,
    0.00163691,
    0.00173642,
    -0.000302395,
    -0.00551453,
    -0.0130394,
    -0.0189322,
    -0.0165021,
    0.00117266,
    0.0372746,
    0.0878833,
    0.14154,
    0.182759,
    0.198242,
    0.182759,
    0.14154,
    0.0878833,
    0.0372746,
    0.00117266,
    -0.0165021,
    -0.0189322,
    -0.0130394,
    -0.00551453,
    -0.000302395,
    0.00173642,
    0.00163691,
    0.000829305,
    0.000215766,
    6.17861e-06,
    1.41996e-19,
  };
  
  /* Dynamic variables */

  int i;
  short * xsp = (short *) inBuf;
  short * ysp = (short *) outBuf;
  int lastL;

  /* There may be a small click when switching clean/overdrive here. */

  if (isClean) {
    for (i = 0; i < bufLengthSamples; i++) {
      *ysp++ = ((*xsp++) * gain) >> 8;
    }
    return outBuf;
  }

  /* Remember the position in the delay line. */

  lastL = L;

  for (i = 0; i < bufLengthSamples; i++) {
    int j;
    float x0 = (float) (((*xsp++) * gain) >> 8); /* Next input sample */
    float xd = (x0 - x_1) / (float) TIMESOVER; /* Interpolation rate */
    float x  = x_1;
    
    x_1 = x0;

    /* Interpolate samples, shape them and put them in the high-rate delay */

    for (j = 0; j < TIMESOVER; j++, x += xd) {
      int k = TMID + (int) x;
#if 0
      k &= 0xFFFF;
#else
      if (k < 0) k = 0; else if (65535 < k) k = 65535;
#endif
      L = (L + 1) & (OVERDLYSZ - 1); /* Pre-increment of delay line */
      dly[L] = (float) ampOut[k];
    }
  }

  /* The dly[] buffer now contains the oversampled signal */

  L = lastL;			/* Restore delay index */

  for (i = 0; i < bufLengthSamples; i++) {
    int j;
    int k;			/* Convolution index into delay line */
    float sum = 0.0;

    L = (L + TIMESOVER) & (OVERDLYSZ - 1); /* Pre-increment the delay index */

    if ((KERNELSZ - 1) <= L) {	/* Common case; conv. run is included */
      for (j = 0, k = L; j < KERNELSZ; j++, k--) {
	sum += dly[k] * w[j];
      }
    }
    else {			/* Special case; conv. run wraps buffer */
      for (k = L, j = 0; 0 <= k; k--, j++) { /* Low end of dly[] buffer */
	sum += dly[k] * w[j];
      }
      for (k = OVERDLYSZ - 1; j < KERNELSZ; j++, k--) {	/* High end of dly[] */
	sum += dly[k] * w[j];
      }
    }

    *ysp++ = (short) sum;
  }

  return outBuf;
}
#endif /* USE_WAVESHAPER */

#ifdef USE_TUBEMODEL1
unsigned char * preamp (unsigned char * inBuf,
			unsigned char * outBuf,
			size_t bufLengthSamples)
{
  /* Past input samples */

  static float x_1 = 0.0;	/* Previous 22 KHz input sample */
  static float dly[OVERDLYSZ];	/* Previous 66 KHz input samples */
  static int L = 0;		/* Position in dly[] */

  /* Weights for the anti-imaging filter. */

  static float w[KERNELSZ] = {
    1.41996e-19,
    6.17861e-06,
    0.000215766,
    0.000829305,
    0.00163691,
    0.00173642,
    -0.000302395,
    -0.00551453,
    -0.0130394,
    -0.0189322,
    -0.0165021,
    0.00117266,
    0.0372746,
    0.0878833,
    0.14154,
    0.182759,
    0.198242,
    0.182759,
    0.14154,
    0.0878833,
    0.0372746,
    0.00117266,
    -0.0165021,
    -0.0189322,
    -0.0130394,
    -0.00551453,
    -0.000302395,
    0.00173642,
    0.00163691,
    0.000829305,
    0.000215766,
    6.17861e-06,
    1.41996e-19,
  };
  
  /* Dynamic variables */

  int i;
  short * xsp = (short *) inBuf;
  short * ysp = (short *) outBuf;
  int lastL;

  /* There may be a small click when switching clean/overdrive here. */

  if (isClean) {
    for (i = 0; i < bufLengthSamples; i++) {
      *ysp++ = ((*xsp++) * gain) >> 8;
    }
    return outBuf;
  }

  /* Remember the position in the delay line. */

  lastL = L;

  for (i = 0; i < bufLengthSamples; i++) {
    int j;
    float x0 = (float) (((*xsp++) * gain) >> 8); /* Next input sample */
    float xd = (x0 - x_1) / (float) TIMESOVER; /* Interpolation rate */
    float x  = x_1;
    
    x_1 = x0;

    /* Interpolate samples, shape them and put them in the high-rate delay */

    for (j = 0; j < TIMESOVER; j++, x += xd) {
      float pp;			/* Post-pressure */
      float tap;
      float leak;
      float tx = x / 32768.0;	/* Normalize input */

      L = (L + 1) & (OVERDLYSZ - 1); /* Pre-increment of delay line */

      cloud0 = anode0 + pressure0;
      cloud1 = anode1 + pressure1;

      if (x < 0.0) {		/* Negative half of push-pull */
	tap = (-tx) * cloud0;
	dly[L] = -tap;
	pp = cloud0 - tap;
	leak = leakCtrl0 * pp * pp * pp;
	pressure0 = pp - leak;
      }
      else {			/* Positive half of push-pull */
	tap = tx * cloud1;
	dly[L] = tap;
	pp = cloud1 - tap;
	leak = leakCtrl1 * pp * pp * pp;
	pressure1 = pp - leak;
      }

    }
  }

  /* The dly[] buffer now contains the oversampled signal */

  L = lastL;			/* Restore delay index */

  for (i = 0; i < bufLengthSamples; i++) {
    int j;
    int k;			/* Convolution index into delay line */
    float sum = 0.0;

    L = (L + TIMESOVER) & (OVERDLYSZ - 1); /* Pre-increment the delay index */

    if ((KERNELSZ - 1) <= L) {	/* Common case; conv. run is included */
      for (j = 0, k = L; j < KERNELSZ; j++, k--) {
	sum += dly[k] * w[j];
      }
    }
    else {			/* Special case; conv. run wraps buffer */
      for (k = L, j = 0; 0 <= k; k--, j++) { /* Low end of dly[] buffer */
	sum += dly[k] * w[j];
      }
      for (k = OVERDLYSZ - 1; j < KERNELSZ; j++, k--) {	/* High end of dly[] */
	sum += dly[k] * w[j];
      }
    }

    *ysp++ = (short) (sum * exitCompensation);
  }

  return outBuf;
}
#endif /* USE_TUBEMODEL1 */

/* ************************************************************************ */

#ifdef USE_POLYNOM
unsigned char * preamp (unsigned char * inBuf,
			unsigned char * outBuf,
			size_t bufLengthSamples)
{
  /* Past input samples */

  static float x_1 = 0.0;	/* Previous 22 KHz input sample */
  static float dly[OVERDLYSZ];	/* Previous 66 KHz input samples */
  static int L = 0;		/* Position in dly[] */

  /* Weights for the 3X anti-imaging filter. */

  static float w[KERNELSZ] = {

    /* This is a lowpass Blackman sinc kernel, generated by
       fdes -o 3 -kn 33 -w 1 */

    -0.00000000000000000023926614, /* [0] */
    0.00007470000934983863846717, /* [1] */
    0.00028824444347884663631537, /* [2] */
    0.00042731864367296951579683, /* [3] */
    -0.00000000000000000043198477, /* [4] */
    -0.00161581887797293110436980, /* [5] */
    -0.00474718289940297046708428, /* [6] */
    -0.00879654153211008663082460, /* [7] */
    -0.01172384080969100274283168, /* [8] */
    -0.01009364252958080271360686, /* [9] */
    0.00000000000000000360670865, /* [10] */
    0.02125138054931337283637305, /* [11] */
    0.05334715770066042472841161, /* [12] */
    0.09198610203671282736781478, /* [13] */
    0.12944614129839987493042486, /* [14] */
    0.15676502115540202897392419, /* [15] */
    0.16678192162353513983141795, /* [16] */
    0.15676502115540205672949980, /* [17] */
    0.12944614129839987493042486, /* [18] */
    0.09198610203671282736781478, /* [19] */
    0.05334715770066043166730552, /* [20] */
    0.02125138054931337283637305, /* [21] */
    0.00000000000000000360670865, /* [22] */
    -0.01009364252958080965250076, /* [23] */
    -0.01172384080969100968172558, /* [24] */
    -0.00879654153211008663082460, /* [25] */
    -0.00474718289940296873236081, /* [26] */
    -0.00161581887797293283909328, /* [27] */
    -0.00000000000000000043198477, /* [28] */
    0.00042731864367296957000694, /* [29] */
    0.00028824444347884766630744, /* [30] */
    0.00007470000934983920767331, /* [31] */
    -0.00000000000000000023926614  /* [32] */

  };
  
  /* Dynamic variables */

  int i;
  short * xsp = (short *) inBuf;
  short * ysp = (short *) outBuf;
  int lastL;

  /* There may be a small click when switching clean/overdrive here. */

  if (isClean) {
    for (i = 0; i < bufLengthSamples; i++) {
      *ysp++ = ((*xsp++) * gain) >> 8;
    }
    return outBuf;
  }

  /* Remember the position in the delay line. */

  lastL = L;

  for (i = 0; i < bufLengthSamples; i++) {

    int j;
    float x0 = (float) (((*xsp++) * gain) >> 8); /* Next input sample */
    float xd = (x0 - x_1) / (float) TIMESOVER; /* Interpolation rate */
    float x  = x_1;
    
    x_1 = x0;

    /* Interpolate samples, shape them and put them in the high-rate delay */

    for (j = 0; j < TIMESOVER; j++, x += xd) {

      /* NaivePushPull3 : sounds like crap 22-apr-03/FK */

      static float uga = 0.45;
      static float bfa = 0.0025;
      static float ca  = 50; /* 2900; */
      static float ugb = 0.55;
      static float bfb = 0.0025;
      static float cb  = 50; /* 3100; */
      static float bias = -1;
      static float gain = 1.0 / 32768.0;
      
      static float postEa = 0;
      static float postEb = 0;
      static float xnorm = 1.0 / 1000.0;

      float xx = (x * gain) + bias;
      float E = 0.5 * (ca + postEa);
      float preB = E * bfa;
      float B = preB * preB * preB;
      float ea = (E - preB) * xx * uga;
      float eb;
      postEa = E - ea;

      xx = (-x * gain) + bias;
      E = 0.5 * (cb + postEb);
      preB = E * bfb;
      B = preB * preB * preB;
      eb = (E - preB) * xx * ugb;
      postEb = E - eb;

      L = (L + 1) & (OVERDLYSZ - 1); /* Pre-increment of delay line */

      dly[L] = xnorm * (ea - eb);

    }
  }

  /* The dly[] buffer now contains the oversampled signal */

  L = lastL;			/* Restore delay index */

  for (i = 0; i < bufLengthSamples; i++) {

    int j;
    int k;			/* Convolution index into delay line */
    float sum = 0.0;

    L = (L + TIMESOVER) & (OVERDLYSZ - 1); /* Pre-increment the delay index */

    if ((KERNELSZ - 1) <= L) {	/* Common case; conv. run is included */
      for (j = 0, k = L; j < KERNELSZ; j++, k--) {
	sum += dly[k] * w[j];
      }
    }
    else {			/* Special case; conv. run wraps buffer */
      for (k = L, j = 0; 0 <= k; k--, j++) { /* Low end of dly[] buffer */
	sum += dly[k] * w[j];
      }
      for (k = OVERDLYSZ - 1; j < KERNELSZ; j++, k--) {	/* High end of dly[] */
	sum += dly[k] * w[j];
      }
    }

    *ysp++ = (short) (sum * 32767.0);
  }

  return outBuf;
}
#endif /* USE_POLYNOM */


#ifdef COMMENT
#ifdef USE_TUBEMODEL2
unsigned char * preamp (unsigned char * inBuf,
			unsigned char * outBuf,
			size_t bufLengthSamples)
{
  /* Past input samples */

  static float x_1 = 0.0;	/* Previous 22 KHz input sample */
  static float dly[OVERDLYSZ];	/* Previous 66 KHz input samples */
  static int L = 0;		/* Position in dly[] */

  /* Weights for the 3X anti-imaging filter. */

  static float w[KERNELSZ] = {

#if 0
    /* Early output from fdes. */
    1.41996e-19,
    6.17861e-06,
    0.000215766,
    0.000829305,
    0.00163691,
    0.00173642,
    -0.000302395,
    -0.00551453,
    -0.0130394,
    -0.0189322,
    -0.0165021,
    0.00117266,
    0.0372746,
    0.0878833,
    0.14154,
    0.182759,
    0.198242,
    0.182759,
    0.14154,
    0.0878833,
    0.0372746,
    0.00117266,
    -0.0165021,
    -0.0189322,
    -0.0130394,
    -0.00551453,
    -0.000302395,
    0.00173642,
    0.00163691,
    0.000829305,
    0.000215766,
    6.17861e-06,
    1.41996e-19,

#else
    /* This is a lowpass Blackman sinc kernel, generated by
       fdes -o 3 -kn 33 -w 1 */
    -0.00000000000000000023926614, /* [0] */
    0.00007470000934983863846717, /* [1] */
    0.00028824444347884663631537, /* [2] */
    0.00042731864367296951579683, /* [3] */
    -0.00000000000000000043198477, /* [4] */
    -0.00161581887797293110436980, /* [5] */
    -0.00474718289940297046708428, /* [6] */
    -0.00879654153211008663082460, /* [7] */
    -0.01172384080969100274283168, /* [8] */
    -0.01009364252958080271360686, /* [9] */
    0.00000000000000000360670865, /* [10] */
    0.02125138054931337283637305, /* [11] */
    0.05334715770066042472841161, /* [12] */
    0.09198610203671282736781478, /* [13] */
    0.12944614129839987493042486, /* [14] */
    0.15676502115540202897392419, /* [15] */
    0.16678192162353513983141795, /* [16] */
    0.15676502115540205672949980, /* [17] */
    0.12944614129839987493042486, /* [18] */
    0.09198610203671282736781478, /* [19] */
    0.05334715770066043166730552, /* [20] */
    0.02125138054931337283637305, /* [21] */
    0.00000000000000000360670865, /* [22] */
    -0.01009364252958080965250076, /* [23] */
    -0.01172384080969100968172558, /* [24] */
    -0.00879654153211008663082460, /* [25] */
    -0.00474718289940296873236081, /* [26] */
    -0.00161581887797293283909328, /* [27] */
    -0.00000000000000000043198477, /* [28] */
    0.00042731864367296957000694, /* [29] */
    0.00028824444347884766630744, /* [30] */
    0.00007470000934983920767331, /* [31] */
    -0.00000000000000000023926614  /* [32] */

#endif

  };
  
  /* Dynamic variables */

  int i;
  short * xsp = (short *) inBuf;
  short * ysp = (short *) outBuf;
  int lastL;

  /* There may be a small click when switching clean/overdrive here. */

  if (isClean) {
    for (i = 0; i < bufLengthSamples; i++) {
      *ysp++ = ((*xsp++) * gain) >> 8;
    }
    return outBuf;
  }

  /* Remember the position in the delay line. */

  lastL = L;

  for (i = 0; i < bufLengthSamples; i++) {
    int j;
    float x0 = (float) (((*xsp++) * gain) >> 8); /* Next input sample */
    float xd = (x0 - x_1) / (float) TIMESOVER; /* Interpolation rate */
    float x  = x_1;
    
    x_1 = x0;

    /* Interpolate samples, shape them and put them in the high-rate delay */

    for (j = 0; j < TIMESOVER; j++, x += xd) {

#ifdef SINGLE_TRIODE00
      float pp;			/* Post-pressure */
      float tap;
      float leak;
#endif /* SINGLE_TRIODE00 */

#ifdef NAIVEPUSHPULL1
      float tapA;
      float ptpA;
      float leakA;
      float tapB;
      float ptpB;
      float leakB;
#endif /* NAIVEPUSHPULL1 */

      float tx = x * inputAttn;	/* Scale input */

      L = (L + 1) & (OVERDLYSZ - 1); /* Pre-increment of delay line */

#ifdef SINGLE_TRIODE00
      cloud = anode + pressure;

#ifdef TUBEMODEL2_ALT2
      tap = cloud * (tx - (feedback * pressure));
#else
      tap = tx * (cloud - (feedback * pressure));
#endif
      pp = cloud - tap;
      leak = leakCtrl * pp * pp * pp;
      pressure = pp - leak;

      dly[L] = cloud + exitBias; /* -2 -- +2 after assignment */
#endif /* SINGLE_TRIODE00 */

#ifdef NAIVEPUSHPULL1
      plateA = anodeA + plpA;
      tapA = plateA * (tx - (feedbackA * plpA));
      ptpA = plateA - tapA;
      leakA = leakcontrolA * ptpA * ptpA * ptpA;
      plpA = ptpA - leakA;

      plateB = anodeB + plpB;
      tapB = plateB * ((-tx) - (feedbackB * plpB));
      ptpB = plateB - tapB;
      leakB = leakcontrolB * ptpB * ptpB * ptpB;
      plpB = ptpB - leakB;

      dly[L] = ((plateA - plateB) + offsetCompensation);
#endif /* NAIVEPUSHPULL1 */

    }
  }

  /* The dly[] buffer now contains the oversampled signal */

  L = lastL;			/* Restore delay index */

  for (i = 0; i < bufLengthSamples; i++) {
    int j;
    int k;			/* Convolution index into delay line */
    float sum = 0.0;

    L = (L + TIMESOVER) & (OVERDLYSZ - 1); /* Pre-increment the delay index */

    if ((KERNELSZ - 1) <= L) {	/* Common case; conv. run is included */
      for (j = 0, k = L; j < KERNELSZ; j++, k--) {
	sum += dly[k] * w[j];
      }
    }
    else {			/* Special case; conv. run wraps buffer */
      for (k = L, j = 0; 0 <= k; k--, j++) { /* Low end of dly[] buffer */
	sum += dly[k] * w[j];
      }
      for (k = OVERDLYSZ - 1; j < KERNELSZ; j++, k--) {	/* High end of dly[] */
	sum += dly[k] * w[j];
      }
    }

    *ysp++ = (short) (sum * outputGain);
  }

  return outBuf;
}
#endif /* USE_TUBEMODEL2 */
#endif /* COMMENT */

/****************************************************************
 * This is wshapePushPull1
 * (Although a single triode for the moment)
 ****************************************************************/

#ifdef USE_TUBEMODEL2
unsigned char * preamp (unsigned char * inBuf,
			unsigned char * outBuf,
			size_t bufLengthSamples)
{
  /* Past input samples */

  static float x_1 = 0.0;	/* Previous 22 KHz input sample */
  static float dly[OVERDLYSZ];	/* Previous 66 KHz input samples */
  static int L = 0;		/* Position in dly[] */

  /* Weights for the 3X anti-imaging filter. */

  static float w[KERNELSZ] = {

    /* This is a lowpass Blackman sinc kernel, generated by
       fdes -o 3 -kn 33 -w 1 */

    -0.00000000000000000023926614, /* [0] */
    0.00007470000934983863846717, /* [1] */
    0.00028824444347884663631537, /* [2] */
    0.00042731864367296951579683, /* [3] */
    -0.00000000000000000043198477, /* [4] */
    -0.00161581887797293110436980, /* [5] */
    -0.00474718289940297046708428, /* [6] */
    -0.00879654153211008663082460, /* [7] */
    -0.01172384080969100274283168, /* [8] */
    -0.01009364252958080271360686, /* [9] */
    0.00000000000000000360670865, /* [10] */
    0.02125138054931337283637305, /* [11] */
    0.05334715770066042472841161, /* [12] */
    0.09198610203671282736781478, /* [13] */
    0.12944614129839987493042486, /* [14] */
    0.15676502115540202897392419, /* [15] */
    0.16678192162353513983141795, /* [16] */
    0.15676502115540205672949980, /* [17] */
    0.12944614129839987493042486, /* [18] */
    0.09198610203671282736781478, /* [19] */
    0.05334715770066043166730552, /* [20] */
    0.02125138054931337283637305, /* [21] */
    0.00000000000000000360670865, /* [22] */
    -0.01009364252958080965250076, /* [23] */
    -0.01172384080969100968172558, /* [24] */
    -0.00879654153211008663082460, /* [25] */
    -0.00474718289940296873236081, /* [26] */
    -0.00161581887797293283909328, /* [27] */
    -0.00000000000000000043198477, /* [28] */
    0.00042731864367296957000694, /* [29] */
    0.00028824444347884766630744, /* [30] */
    0.00007470000934983920767331, /* [31] */
    -0.00000000000000000023926614  /* [32] */

  };
  
  static float outGain = 32767.0 * 0.707;
  static float inpGain = 2.0 / 32768.0;

  /* Dynamic variables */

  int i;
  short * xsp = (short *) inBuf;
  short * ysp = (short *) outBuf;
  int lastL;

  float FC0 = p[fc0];
  float FC1 = p[fc1];

  /* There may be a small click when switching clean/overdrive here. */

  if (isClean) {
    for (i = 0; i < bufLengthSamples; i++) {
      *ysp++ = ((*xsp++) * gain) >> 8;
    }
    return outBuf;
  }

  /* Remember the position in the delay line. */

  lastL = L;

#define HF_BIAS /* EXPERIMENTAL : 01-sep-2003/FK */
#ifdef HF_BIAS
 {
   static short z = 0;
   short * p = (short *) inBuf;
   for (i = 0; i < bufLengthSamples; i++) {
     short m = *p;
     *p++ = (m + m + m - z) >> 2;
     z = m;
   }
 }
#endif /* HF_BIAS */

  for (i = 0; i < bufLengthSamples; i++) {
    int j;
    float x0 = (float) (((*xsp++) * gain) >> 8); /* Next input sample */
    float xd = (x0 - x_1) / (float) TIMESOVER; /* Interpolation rate */
    float x  = x_1;
    
    x_1 = x0;

    /* Interpolate samples, shape them and put them in the high-rate delay */

    for (j = 0; j < TIMESOVER; j++, x += xd) {
      static float xx_1;
      static float bias = 0.24;
      int i;
      float q;
      float a;
      float b;
      float y;
      float xx_0;
      float xi;
      float tx = x * inpGain;	/* Scale input to -1 ... +1 */

      xx_0 = (FC0 * (tx + bias)) + (FC1 * xx_1);

      xi = ((1.0 + xx_0) * (float) (WAVESHAPE_SZ / 2));

      if (xi < 0.0) {
	xi = 0.0;
      }
      else if ((float) WAVESHAPE_SZ <= xi) {
	xi = (float) (WAVESHAPE_SZ - 1);
      }

      i = (int) xi;
      q = xi - (float) i;
      a = wshape[ i ];
      b = wshape[i+1];

      y = a + (q * (b - a));

      xx_1 = (outscl * (y - minf)) - 1.0;

      L = (L + 1) & (OVERDLYSZ - 1); /* Pre-increment of delay line */
      dly[L] = xx_1;
    }
  }

  /* The dly[] buffer now contains the oversampled signal */

  L = lastL;			/* Restore delay index */

  for (i = 0; i < bufLengthSamples; i++) {
    int j;
    int k;			/* Convolution index into delay line */
    float sum = 0.0;

    L = (L + TIMESOVER) & (OVERDLYSZ - 1); /* Pre-increment the delay index */

    if ((KERNELSZ - 1) <= L) {	/* Common case; conv. run is included */
      for (j = 0, k = L; j < KERNELSZ; j++, k--) {
	sum += dly[k] * w[j];
      }
    }
    else {			/* Special case; conv. run wraps buffer */
      for (k = L, j = 0; 0 <= k; k--, j++) { /* Low end of dly[] buffer */
	sum += dly[k] * w[j];
      }
      for (k = OVERDLYSZ - 1; j < KERNELSZ; j++, k--) {	/* High end of dly[] */
	sum += dly[k] * w[j];
      }
    }

    *ysp++ = (short) (sum * outGain);
  }

#ifdef HF_BIAS
 {
   static short z = 0;
   short * p = (short *) outBuf;
   for (i = 0; i < bufLengthSamples; i++) {
     short m = *p;
     *p++ = (m + z) >> 1;
     z = m;
   }
 }
#endif /* HF_BIAS */

  return outBuf;
}
#endif /* USE_TUBEMODEL2 */
