/* Copyright (C) 2003, Fredrik Kilander */
/*
 * program.c
 * 14-may-2004/FK Replacing rotsim module with whirl.
 * 10-may-2003/FK New syntax and parser in a separate file.
 * 2001-12-28/FK
 * 
 * Manager for program change events.
 */

#include <stdio.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <ctype.h>
#include <string.h>
#include <math.h>
#include <sys/soundcard.h>
#include <sys/time.h>
#include <unistd.h>
#include <pthread.h>
#include <assert.h>

#include "preamp.h"
#include "scanner.h"
#include "main.h"
#include "midiIn.h"
#include "reverb.h"
#include "whirl.h"

#define SET_TRUE 1
#define SET_NONE 0
#define SET_FALSE -1

#define MESSAGEBUFFERSIZE 256

#define MAXPROGS 128
#define FILE_BUFFER_SIZE 2048
#define NAMESZ 22

#define NFLAGS 1		/* The nof flag fields in Programme struct */



/* Flag bits used in the first field */

#define FL_INUSE  0x0001	/* Record is in use */

#define FL_DRAWBR 0x0002	/* Set drawbars */

#define FL_ATKENV 0x0004	/* Attack envelope */
#define FL_ATKCKL 0x0008	/* Attack Click level */
#define FL_ATKCKD 0x0010	/* Attack click duration */

#define FL_RLSENV 0x0020	/* Release envelope */
#define FL_RLSCKL 0x0040	/* Release level */
#define FL_RLSCKD 0x0080	/* Release duration */

#define FL_SCANNR 0x0100	/* Vibrato scanner */

#define FL_PRCENA 0x0200	/* Percussion on/off */
#define FL_PRCVOL 0x0400	/* Percussion soft/normal */
#define FL_PRCSPD 0x0800	/* Percussion slow/fast */
#define FL_PRCHRM 0x1000	/* Percussion 2nd/3rd */

#define FL_OVRSEL 0x2000	/* Overdrive on/off */

#define FL_ROTENA 0x4000	/* Rotary on/off */
#define FL_ROTSPS 0x8000	/* Rotary speed select */

#define FL_RVBMIX 0x00010000	/* Reverb on/off */

#define FL_DRWRND 0x00020000	/* Randomize drawbars */

typedef struct _programme {
  char name [NAMESZ];
  unsigned int flags[NFLAGS];
  unsigned int drawbars[9];
  short        keyAttackEnvelope;
  float        keyAttackClickLevel;
  float        keyAttackClickDuration;
  short        keyReleaseEnvelope;
  float        keyReleaseClickLevel;
  float        keyReleaseClickDuration;
  short        scanner;
  short        percussionEnabled;
  short        percussionVolume;
  short        percussionSpeed;
  short        percussionHarmonic;
  short        overdriveSelect;
  short        rotaryEnabled;
  short        rotarySpeedSelect;
  float        reverbMix;
} Programme;

static Programme defaultProgramme = {
  "DEFAULT",
  {FL_INUSE|FL_DRAWBR},
  {0,0, 8,8,0,0, 8,0,0}
};


static Programme programmes[MAXPROGS];

static int displayPgmChanges = FALSE;

/* Property codes */

enum propertyId {
  pr_Name,
  pr_Drawbars,
  pr_KeyAttackEnvelope,
  pr_KeyAttackClickLevel,
  pr_KeyAttackClickDuration,
  pr_KeyReleaseEnvelope,
  pr_KeyReleaseClickLevel,
  pr_KeyReleaseClickDuration,
  pr_Scanner,
  pr_PercussionEnabled,
  pr_PercussionVolume,
  pr_PercussionSpeed,
  pr_PercussionHarmonic,
  pr_OverdriveSelect,
  pr_RotaryEnabled,
  pr_RotarySpeedSelect,
  pr_ReverbMix,
  pr_void
};

typedef struct _symbolmap {
  char * propertyName;
  int property;
} SymbolMap;

static SymbolMap propertySymbols [] = {
  {"name",      pr_Name},
  {"drawbars",  pr_Drawbars},
  {"attackenv", pr_KeyAttackEnvelope},
  {"attacklvl", pr_KeyAttackClickLevel},
  {"attackdur", pr_KeyAttackClickDuration},
  {"vibrato",   pr_Scanner},
  {"perc",      pr_PercussionEnabled},
  {"percvol",   pr_PercussionVolume},
  {"percspeed", pr_PercussionSpeed},
  {"percharm",  pr_PercussionHarmonic},
  {"overdrive", pr_OverdriveSelect},
  {"rotary",    pr_RotaryEnabled},
  {"rotaryspeed", pr_RotarySpeedSelect},
  {"reverbmix", pr_ReverbMix},
  {NULL, pr_void}
};

/*
 *
 */
static int getPropertyIndex (char * sym) {
  int i;
  for (i = 0; propertySymbols[i].propertyName != NULL; i++) {
    if (!strcasecmp (propertySymbols[i].propertyName, sym)) {
      return propertySymbols[i].property;
    }
  }
  return -1;
}

/*
 * Controls if the name of a new program is displayed or not.
 * doDisplay   TRUE = display changes, FALSE = do not display changes.
 */
void setDisplayPgmChanges (int doDisplay) {
  displayPgmChanges = doDisplay;
}

/* ======================================================================== */

/*
 * Prints a message followed by the given filename and linenumber.
 * Returns the given return code, so that it may be called in a return
 * statement from within a parsing function.
 */
static int stateMessage (char * fileName,
			 int lineNumber,
			 char * msg,
			 int code) {
  printf ("%s in file %s on line %d\n", msg, fileName, lineNumber);
  return code;
}

/*
 * Parses a drawbar registration.
 */
static int parseDrawbarRegistration (char * drw,
				     unsigned int bar[],
				     int    lineNumber,
				     char * fileName) {

  char msg[MESSAGEBUFFERSIZE];
  int bus = 0;
  char * t = drw;

  while (bus < 9) {
    if (*t == '\0') {
      sprintf (msg, "Drawbar registration incomplete '%s'", drw);
      return stateMessage (fileName, lineNumber, msg, -1);
    }
    if ((isspace (*t)) || (*t == '-') || (*t == '_')) {
      t++;
      continue;
    }
    if (('0' <= *t) && (*t <= '8')) {
      bar[bus] = *t - '0';
      t++;
      bus++;
      continue;
    }
    else {
      sprintf (msg, "Illegal char in drawbar registration '%c'", *t);
      return stateMessage (fileName, lineNumber, msg, -1);
    }
  }

  return 0;
}



/*
 * Return TRUE if the supplied string can be interpreted as an enabling arg.
 */
static int isAffirmative (char * value) {
  int n;
  if (!strcasecmp (value, "on")) return TRUE;
  if (!strcasecmp (value, "yes")) return TRUE;
  if (!strcasecmp (value, "true")) return TRUE;
  if (!strcasecmp (value, "enabled")) return TRUE;
  if (sscanf (value, "%d", &n) == 1) {
    if (n != 0) return TRUE;
  }
  return FALSE;
}

/*
 * Return TRUE if the supplied string can be interpreted as a disabling arg.
 */
static int isNegatory (char * value) {
  int n;
  if (!strcasecmp (value, "off")) return TRUE;
  if (!strcasecmp (value, "no")) return TRUE;
  if (!strcasecmp (value, "none")) return TRUE;
  if (!strcasecmp (value, "false")) return TRUE;
  if (!strcasecmp (value, "disabled")) return TRUE;
  if (sscanf (value, "%d", &n) == 1) {
    if (n == 0) return TRUE;
  }
  return FALSE;
}

/*
 * This function is called from scanner/syntax parser in file pgmParser.c.
 * Return: 0 OK, non-zero error.
 */
int bindToProgram (char * fileName,
		   int    lineNumber,
		   int    pgmnr,
		   char * sym,
		   char * val)
{
  static int previousPgmNr = -1;
  int i;
  int prop;
  char msg[MESSAGEBUFFERSIZE];
  float fv;
  Programme * PGM;

  /*
  printf ("%s:%d:pgm=%d '%s'='%s'\n", fileName, lineNumber, pgmnr, sym, val);
  */


  /* Check the program number */

  if ((pgmnr < 0) || (MAXPROGS <= pgmnr)) {
    sprintf (msg, "Program number %d out of range", pgmnr);
    return stateMessage (fileName, lineNumber, msg, -1);
  }

  PGM = &(programmes[pgmnr]);

  /* If this is a new program number, clear the property flags */

  if (pgmnr != previousPgmNr) {
    PGM->flags[0] = 0;
    previousPgmNr = pgmnr;
  }

  /* Scan for a matching property symbol */

  prop = getPropertyIndex (sym);

  if (prop < 0) {
    sprintf (msg, "Unrecognized property '%s'", sym);
    return stateMessage (fileName, lineNumber, msg, -1);
  }

  switch (prop) {

  case pr_Name:
    strncpy (PGM->name, val, NAMESZ);
    PGM->name[NAMESZ-1] = '\0';
    PGM->flags[0] |= FL_INUSE;
    break;

  case pr_Drawbars:
    if (!strcasecmp (val, "random")) {
      PGM->flags[0] |= (FL_INUSE | FL_DRAWBR | FL_DRWRND);
    }
    else if (!parseDrawbarRegistration (val,
					PGM->drawbars,
					lineNumber,
					fileName)) {
      PGM->flags[0] |= (FL_INUSE|FL_DRAWBR);
    }
    else {
      return -1;
    }
    break;

  case pr_Scanner:
    if (!strcasecmp (val, "v1")) {
      PGM->scanner = VIB1;	/* in scanner.h */
      PGM->flags[0] |= (FL_INUSE|FL_SCANNR);
    }
    else if (!strcasecmp (val, "v2")) {
      PGM->scanner = VIB2;	/* in scanner.h */
      PGM->flags[0] |= (FL_INUSE|FL_SCANNR);
    }
    else if (!strcasecmp (val, "v3")) {
      PGM->scanner = VIB3;	/* in scanner.h */
      PGM->flags[0] |= (FL_INUSE|FL_SCANNR);
    }
    else if (!strcasecmp (val, "c1")) {
      PGM->scanner = CHO1;	/* in scanner.h */
      PGM->flags[0] |= (FL_INUSE|FL_SCANNR);
    }
    else if (!strcasecmp (val, "c2")) {
      PGM->scanner = CHO2;	/* in scanner.h */
      PGM->flags[0] |= (FL_INUSE|FL_SCANNR);
    }
    else if (!strcasecmp (val, "c3")) {
      PGM->scanner = CHO3;	/* in scanner.h */
      PGM->flags[0] |= (FL_INUSE|FL_SCANNR);
    }
    else if (isNegatory (val)) {
      PGM->scanner = 0;
      PGM->flags[0] |= (FL_INUSE|FL_SCANNR);
    }
    else {
      sprintf (msg, "Unrecognized vibrato value '%s'", val);
      return stateMessage (fileName, lineNumber, msg, -1);
    }
    break;

  case pr_PercussionEnabled:
    if (isAffirmative (val)) {
      PGM->percussionEnabled = TRUE;
      PGM->flags[0] |= (FL_INUSE|FL_PRCENA);
    }
    else if (isNegatory (val)) {
      PGM->percussionEnabled = FALSE;
      PGM->flags[0] |= (FL_INUSE|FL_PRCENA);
    }
    else {
      sprintf (msg, "Unrecognized percussion enabled value '%s'", val);
      return stateMessage (fileName, lineNumber, msg, -1);
    }
    break;

  case pr_PercussionVolume:
    if (!strcasecmp (val, "normal") ||
	!strcasecmp (val, "high")   ||
	!strcasecmp (val, "hi")) {
      PGM->percussionVolume = FALSE;
      PGM->flags[0] |= (FL_INUSE|FL_PRCVOL);
    }
    else if (!strcasecmp (val, "soft") ||
	     !strcasecmp (val, "low")  ||
	     !strcasecmp (val, "lo")) {
      PGM->percussionVolume = TRUE;
      PGM->flags[0] |= (FL_INUSE|FL_PRCVOL);
    }
    else {
      sprintf (msg, "Unrecognized percussion volume argument '%s'", val);
      return stateMessage (fileName, lineNumber, msg, -1);
    }
    break;

  case pr_PercussionSpeed:
    if (!strcasecmp (val, "fast") ||
	!strcasecmp (val, "high") ||
	!strcasecmp (val, "hi")) {
      PGM->percussionSpeed = TRUE;
      PGM->flags[0] |= (FL_INUSE|FL_PRCSPD);
    }
    else if (!strcasecmp (val, "slow") ||
	     !strcasecmp (val, "low")  ||
	     !strcasecmp (val, "lo")) {
      PGM->percussionSpeed = FALSE;
      PGM->flags[0] |= (FL_INUSE|FL_PRCSPD);
    }
    else {
      sprintf (msg, "Unrecognized percussion speed argument '%s'", val);
      return stateMessage (fileName, lineNumber, msg, -1);
    }
    break;

  case pr_PercussionHarmonic:
    if (!strcasecmp (val, "second") ||
	!strcasecmp (val, "2nd")    ||
	!strcasecmp (val, "low")    ||
	!strcasecmp (val, "lo")) {
      PGM->percussionHarmonic = TRUE;
      PGM->flags[0] |= (FL_INUSE|FL_PRCHRM);
    }
    else if (!strcasecmp (val, "third") ||
	     !strcasecmp (val, "3rd")   ||
	     !strcasecmp (val, "high")  ||
	     !strcasecmp (val, "hi")) {
      PGM->percussionHarmonic = FALSE;
      PGM->flags[0] |= (FL_INUSE|FL_PRCHRM);
    }
    else {
      sprintf (msg, "Unrecognized percussion harmonic option '%s'", val);
      return stateMessage (fileName, lineNumber, msg, -1);
    }
    break;

  case pr_OverdriveSelect:
    if (isNegatory (val)) {
      PGM->overdriveSelect = TRUE;
      PGM->flags[0] |= (FL_INUSE|FL_OVRSEL);
    }
    else if (isAffirmative (val)) {
      PGM->overdriveSelect = FALSE;
      PGM->flags[0] |= (FL_INUSE|FL_OVRSEL);
    }
    else {
      sprintf (msg, "Unrecognized overdrive select argument '%s'", val);
      return stateMessage (fileName, lineNumber, msg, -1);
    }
    break;

  case pr_RotarySpeedSelect:
    if (!strcasecmp (val, "tremolo") ||
	!strcasecmp (val, "fast")    ||
	!strcasecmp (val, "high")    ||
	!strcasecmp (val, "hi")) {
      PGM->rotarySpeedSelect = WHIRL_FAST;
      PGM->flags[0] |= (FL_INUSE|FL_ROTSPS);
    }
    else if (!strcasecmp (val, "chorale") ||
	     !strcasecmp (val, "slow")    ||
	     !strcasecmp (val, "low")     ||
	     !strcasecmp (val, "lo")) {
      PGM->rotarySpeedSelect = WHIRL_SLOW;
      PGM->flags[0] |= (FL_INUSE|FL_ROTSPS);
    }
    else if (!strcasecmp (val, "stop")  ||
	     !strcasecmp (val, "zero")  ||
	     !strcasecmp (val, "break") ||
	     !strcasecmp (val, "stopped")) {
      PGM->rotarySpeedSelect = WHIRL_STOP;
      PGM->flags[0] |= (FL_INUSE|FL_ROTSPS);
    }
    else {
      sprintf (msg, "Unrecognized rotary speed argument '%s'", val);
      return stateMessage (fileName, lineNumber, msg, -1);
    }
    break;

  case pr_ReverbMix:
    PGM->flags[0] |= (FL_INUSE|FL_RVBMIX);
    if (sscanf (val, "%f", &fv) == 0) {
      sprintf (msg, "Unrecognized reverb mix value : '%s'", val);
      return stateMessage (fileName, lineNumber, msg, -1);
    }
    else if ((fv < 0.0) || (1.0 < fv)) {
      sprintf (msg, "Reverb mix value out of range : %f", fv);
      return stateMessage (fileName, lineNumber, msg, -1);
    }
    else {
      PGM->reverbMix = fv;
    }
    break;

  } /* switch property */

  return 0;
}


/*
 * This is the routine called by the MIDI parser when it detects
 * a Program Change message.
 */
void installProgram (unsigned char p) {
  if ((0 < p) && (p < MAXPROGS)) {

    Programme * PGM = &(programmes[p]);
    unsigned int flags0 = PGM->flags[0];

    if (flags0 & FL_INUSE) {

      if (flags0 & FL_DRWRND) {
	int i;
	for (i = 0; i < 9; i++) {
	  PGM->drawbars[i] = rand () % 9;
	}
	sprintf (PGM->name,
		 "%c%c%c %c%c%c%c %c%c",
		 '0' + PGM->drawbars[0],
		 '0' + PGM->drawbars[1],
		 '0' + PGM->drawbars[2],
		 '0' + PGM->drawbars[3],
		 '0' + PGM->drawbars[4],
		 '0' + PGM->drawbars[5],
		 '0' + PGM->drawbars[6],
		 '0' + PGM->drawbars[7],
		 '0' + PGM->drawbars[8]);
      }

      if (displayPgmChanges) {
	fprintf (stderr, "%22c\r%s\r", ' ', PGM->name);
      }

      if (flags0 & FL_DRAWBR) {
	setDrawBars (PGM->drawbars);
      }

      /* Key attack click */

      /* Key release click */

      if (flags0 & FL_SCANNR) {
	setVibrato (PGM->scanner);
      }

      if (flags0 & FL_PRCENA) {
	setPercussionEnabled (PGM->percussionEnabled);
      }

      if (flags0 & FL_PRCVOL) {
	setPercussionVolume (PGM->percussionVolume);
      }

      if (flags0 & FL_PRCSPD) {
	setPercussionFast (PGM->percussionSpeed);
      }

      if (flags0 & FL_PRCHRM) {
	setPercussionFirst (PGM->percussionHarmonic);
      }

      if (flags0 & FL_OVRSEL) {
	setClean (PGM->overdriveSelect);
      }

      if (flags0 & FL_ROTENA) {
	/* Rotary enabled */
      }

      if (flags0 & FL_ROTSPS) {
	setRevSelect ((int) (PGM->rotarySpeedSelect));
      }

      if (flags0 & FL_RVBMIX) {
	setReverbMix (PGM->reverbMix);
      }

    }
  }
}

#define MAXROWS 18
#define MAXCOLS  4

/*
 * Displays the number and names of the loaded programmes in a multicolumn
 * list on the given output stream.
 */
void listProgrammes (FILE * fp) {
  int matrix [MAXROWS][MAXCOLS];
  int row;
  int col;
  int i;
  int mxUse = 0;
  int mxLimit = MAXROWS * MAXCOLS;

  for (row = 0; row < MAXROWS; row++) {
    for (col = 0; col < MAXCOLS; col++) {
      matrix [row][col] = -1;
    }
  }

  for (i = row = col = 0; i < MAXPROGS; i++) {
    if (programmes[i].flags[0] & FL_INUSE) {
      if (mxUse < mxLimit) {
	matrix[row][col] = i;
	mxUse++;
	row++;
	if (MAXROWS <= row) {
	  row = 0;
	  col++;
	}
      }
    }
  }

  for (row = 0; row < MAXROWS; row++) {

    for (col = 0; col < MAXCOLS; col++) {
      int x = matrix[row][col];
      if (-1 < x) {
	fprintf (fp, "%3d:%-15.15s", x, programmes[x].name);
      }
      else {
	fprintf (fp, "%19s", " ");
      }
      if (col < 3) {
	fprintf (fp, " ");
      }
      else {
	fprintf (fp, "\n");
      }
    }
  }
}
