/* 2001-11-30/FK
 * scanner.c
 * Vibrato scanner simulator.
 * 04-apr-2004/FK This must be the oldest piece of code that remains.
 *                Added MIDI controller support and edited some comments.
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <assert.h>

#include "main.h"
#include "radium.h"
#include "midiIn.h"
#include "scanner.h"

/*
 * The increment is encoded using a 16-bit fixed-point mantissa.
 * This means that an increment of 1.0 is represented by 2^16, or (1<<16).
 * This quantity is added to inPos using ordinary arithmetic.
 *
 * When writing at inPos, the position is divided into (h.l).
 * The integer part, h, is the first sample to write to. It receives
 * x(1.0 - l) + y(h) of the input sample. That is, we add the fraction to
 * the value already present. The next sample, receives x*l + y(h+1).
 * In order to prevent repeated echoes, the buffer is cleared by the
 * read operation.
 *
 * We need to translate these values into fixed-point integer aritmetic.
 * Since the fraction is 16-bit, they are represented by l/(2^16).
 * Thus we do (x * l) >> 16 to compute (x * l).
 * Having that, we simply do  x - xl to arrive at the value to add to x(h).
 *
 * In order to avoid a buffer overrun at (h+1), we must use a temporary
 * variable for the address of (h+1) and wrap it back to 0 if it equals
 * the length of the buffer.
 *
 * Increments less than 1.0 will be perceived as a rise in pitch, because
 * they are the ones first seen by the reader, and the energy will be
 * compressed into fewer samples than in the source. Conversely, increments
 * above 1.0 will be perceived as a drop in pitch, because the information
 * tries to escape from the reader, and the energy is distributed across
 * more samples than in the source.
 *
 * Increments must be less than 2.0, or some samples will be skipped which
 * will cause nasty clicks or noise, depending on the output DACs.
 */

#define INCTBL_SIZE 2048
#define INCTBL_MASK 0x07ffffff	/* Fixed point */

#define BUF_SIZE_SAMPLES 1024
#define BUF_SIZE_BYTES   2048
#define BUF_MASK_SAMPLES 0x000003FF
#define BUF_MASK_POSN    0x03FFFFFF /* Fixed-point mask */

static unsigned int incr1Table [INCTBL_SIZE]; /* V1/C1 */
static unsigned int incr2Table [INCTBL_SIZE]; /* V2/C2 */
static unsigned int incr3Table [INCTBL_SIZE]; /* V3/C3 */

static unsigned int * incrTable = incr3Table;

static unsigned int offset1Table [INCTBL_SIZE];
static unsigned int offset2Table [INCTBL_SIZE];
static unsigned int offset3Table [INCTBL_SIZE];

static unsigned int * offsetTable = offset3Table;

static unsigned int stator = 0;
static unsigned int statorIncrement = 0;

static unsigned int inPos  = 0;
static unsigned int outPos = BUF_MASK_SAMPLES / 2;

static short vibBuffer [BUF_SIZE_BYTES];

static double vib1ModAmp = 0.005;
static double vib2ModAmp = 0.010;
static double vib3ModAmp = 0.015;

static double vib1OffAmp = 1.0;
static double vib2OffAmp = 2.5;
static double vib3OffAmp = 5.0;

static double vibFqHertz = 8.0;

static int mixedBuffers = FALSE;
static int effectEnabled = FALSE;

/*
 *
 */
void setScannerFrequency (double Hertz) {
  vibFqHertz = Hertz;
  statorIncrement =
    (unsigned int) (((vibFqHertz * INCTBL_SIZE) / SampleRateD) * 65536.0);
}

/*
 *
 */
void setVibrato (int select) {
  incrTable = incr1Table;

  switch (select & 3) {

  case 0:			/* disable */
    effectEnabled = FALSE;
    break;

  case 1:
    effectEnabled = TRUE;
    incrTable = incr1Table;
    offsetTable = offset1Table;
    break;

  case 2:
    effectEnabled = TRUE;
    incrTable = incr2Table;
    offsetTable = offset2Table;
    break;

  case 3:
    effectEnabled = TRUE;
    incrTable = incr3Table;
    offsetTable = offset3Table;
    break;
  }

  mixedBuffers = select & 0x80;
}

/*
 * This routine was meant to be called from the MIDI parser in response to
 * data entry increment and data entry decrement controllers.
 */
void setScannerAdvance (int forward) {
  static int controlPosition = 0;
  controlPosition = (controlPosition + ((forward) ? 1 : 7)) & 0x07;

  switch (controlPosition) {
  case 0:
    setVibrato (0);
    //    fprintf (stderr, " OFF\r"); fflush (stderr);
    break;
  case 1:
    setVibrato (VIB1);
    //    fprintf (stderr, "VIB1\r"); fflush (stderr);
    break;
  case 2:
    setVibrato (VIB2);
    //    fprintf (stderr, "VIB2\r"); fflush (stderr);
    break;
  case 3:
    setVibrato (VIB3);
    //    fprintf (stderr, "VIB3\r"); fflush (stderr);
    break;
  case 4:
    setVibrato (0);
    //    fprintf (stderr, " OFF\r"); fflush (stderr);
    break;
  case 5:
    setVibrato (CHO1);
    //    fprintf (stderr, "CHO1\r"); fflush (stderr);
    break;
  case 6:
    setVibrato (CHO2);
    //    fprintf (stderr, "CHO2\r"); fflush (stderr);
    break;
  case 7:
    setVibrato (CHO3);
    //    fprintf (stderr, "CHO3\r"); fflush (stderr);
    break;
  }
}

static void setVibratoFromMIDI (unsigned char u) {
  switch (u / 19) {
  case 0:
    setVibrato (0);
    break;
  case 1:
    setVibrato (VIB1);
    break;
  case 2:
    setVibrato (CHO1);
    break;
  case 3:
    setVibrato (VIB2);
    break;
  case 4:
    setVibrato (CHO2);
    break;
  case 5:
    setVibrato (VIB3);
    break;
  case 6:
    setVibrato (CHO3);
    break;
  }
}

/*
 *
 */
static double angularOffset (int i, int N) {
  return sin ((2.0 * M_PI * i) / N);
}

/*
 *
 */
static void initIncrementTables () {
  int i;
  double S = 65536.0;
  double voff1;
  double voff2;
  double voff3;
  double sin_3qtr_circle = sin (1.5 * M_PI);

  /*
   * Since we implement pitch shift by varying the writer's speed, we
   * need to know in each instant how much to increment the fixed point
   * position of the writer. The incrTable[] contains the fixed point
   * update amount, for each point of the circle the stator makes.
   * Thus, the code here maps the circle onto the INCTBL_SIZE slots
   * in the incrTable[] array.
   */

  for (i = 0; i < INCTBL_SIZE; i++) {
    double v = sin ((2.0 * M_PI * i) / INCTBL_SIZE);

    incr1Table[i] = (unsigned int) ((1.0 + v * vib1ModAmp) * 65536.0);
    incr2Table[i] = (unsigned int) ((1.0 + v * vib2ModAmp) * 65536.0);
    incr3Table[i] = (unsigned int) ((1.0 + v * vib3ModAmp) * 65536.0);
  }

  voff1 = vib1OffAmp;
  voff2 = vib2OffAmp;
  voff3 = vib3OffAmp;

  for (i = 0; i < INCTBL_SIZE; i++) {
    double v = sin ((2.0 * M_PI * i) / INCTBL_SIZE);
    offset1Table [i] = (unsigned int) ((1.0 + voff1 + (v * vib1OffAmp)) * S);
    offset2Table [i] = (unsigned int) ((1.0 + voff2 + (v * vib2OffAmp)) * S);
    offset3Table [i] = (unsigned int) ((1.0 + voff3 + (v * vib3OffAmp)) * S);
  }

#ifdef DEBUG
  {
    FILE * fp;
    char * debugfile = "scanner.log";
    if ((fp = fopen (debugfile, "w")) != NULL) {

      fprintf (fp,
	       "statorIncrement = 0x%08x %g\n",
	       statorIncrement,
	       ((double) statorIncrement) / S);

      fprintf (fp, "voff1 = %g,  voff2 = %g, voff3 = %g\n",
	       voff1, voff2, voff3);

      fprintf (fp, "offset1Table\n");
      for (i = 0; i < INCTBL_SIZE; i++) {
	fprintf (fp,
		 "%4d 0x%08x %g\n",
		 i,
		 offset1Table[i],
		 ((double) offset1Table[i]) / S);
      }
      fclose (fp);
    }
    else {
      perror (debugfile);
      exit (1);
    }
  }
#endif /* DEBUG */
}

/*
 *
 */
void initVibrator () {
  setScannerFrequency (vibFqHertz);
  initIncrementTables ();
  setVibrato (0);
  setMIDIControllerFunction (RD_P1_K13, setVibratoFromMIDI);
}

/*
 *
 */
unsigned char * vibratoProc (unsigned char * inbuffer,
			     unsigned char * outbuffer,
			     size_t bufferLengthSamples) {

  int i;
  short * xp = (short *) inbuffer;
  short * yp = (short *) outbuffer;

  if (!effectEnabled) return inbuffer;

  for (i = 0; i < bufferLengthSamples; i++) {
    short x = *xp++;
    unsigned int j =
      ((outPos << 16) + offsetTable[stator >> 16]) & BUF_MASK_POSN;
    int h = j >> 16;
    int k = (h + 1) & BUF_MASK_SAMPLES;
    int f = j & 0xFFFF;
    int g = (f * (int) x) >> 16;

    vibBuffer[h] += x - g;
    vibBuffer[k] += g;

    if (mixedBuffers) {
      *yp++ = (x + vibBuffer[outPos]) >> 1;
    }
    else {
      *yp++ = vibBuffer[outPos];
    }

    vibBuffer[outPos] = 0;
    outPos = (outPos + 1) & BUF_MASK_SAMPLES;
    stator = (stator + statorIncrement) & INCTBL_MASK;
  }

#ifdef COMMENT
  for (i = 0; i < bufferLengthSamples; i++) {
    short x = *xp++;
    int h = (int) (inPos >> 16);
    int l = (int) (inPos & 0xFFFF);
    int k = (h + 1) & BUF_MASK_SAMPLES;
    int xl = (l * (int) x) >> 16;

    vibBuffer[h] += x - xl;
    vibBuffer[k] += xl;

    if (mixedBuffers) {
      *yp++ = (x + vibBuffer[outPos]) >> 1;
    }
    else {
      *yp++ = vibBuffer[outPos];
    }

    vibBuffer[outPos] = 0;
    outPos = (outPos + 1) & BUF_MASK_SAMPLES;

    inPos = (inPos + incrTable[(stator >> 16)]) & BUF_MASK_POSN;
    stator = (stator + statorIncrement) & INCTBL_MASK;
  }
#endif /* COMMENT */

  return outbuffer;
}

