/* tonegen.c
 * 04-apr-2004/FK Added MIDI controllers for the drawbars.
 *
 * 23-sep-2003/FK Version 0.3.1 to experiment with stuff.
 *                Upped the precision by a magnitude to get rid of looping
 *                noise in the higher frequencies.
 *
 * 22-sep-2003/FK Fixed a small bug in the click envelope initializer.
 *                Fixed a small bug in the initialization routine.
 *                Added routine dumpOscToText() as a debugging aid.
 *
 * 20-Aug-2003/FK Added additional harmonics to wavetable loading, for
 *                experimental purposes. However, the residual harmonics
 *                in the real instruments are likely to be way low, possibly
 *                below the threshold. Aged capacitors, however, is another
 *                matter.
 *
 * 30-May-2003/FK Removed fixed-point percussion code. Cleanup, commenting.
 *
 * 25-May-2003/FK Redid percussion with a float multiplier that decays expo-
 *                nentially, like it should (or so I think). The quantization
 *                noise disappeared, at last.
 *
 * 07-may-2003/FK Added a single bit to the acx field in the Connection struct
 *                to distinguish between the added and active lists. That
 *                would appear to take care of the crashing bug, although I
 *                did suffer a hung note while testing it. Strange, but not
 *                conclusive.
 *
 * 06-may-2003/FK New way of writing dates. And I've begun looking for a way
 *                to fix another crashing bug: when the same connection is
 *                first closed and then opened in the same call to generate-
 *                Fragment(), the open code thinks the connection is active
 *                and trashes the active connection list, because the con-
 *                nection is still on the added list. This rare condition
 *                (open and close in the same call) happens more often than
 *                you would assume; rapid playing triggers it 5 percent of
 *                the time because fingers bounce off keys that just bounces
 *                down and the up again.
 *
 * 2002-11-26/FK Reinstated the key compression with a fix for the release
 *               click and compensated for mixing at unity. Seems to work
 *               better. Percussion is still noisy (egde noise from steps).
 *
 * 2002-11-23/FK Radical update of the oscillator structure and fragment
 *               generator. In the previous scheme we modelled for each
 *               oscillator the buses it was connected to. In the current
 *               scheme we keep track of the connections and see which
 *               are active. The hung note bug is fixed. The routine
 *               oscGenerateFragment() became simpler and more efficient.
 *               New initialization routines for tapering has been added.
 *               Routine dumpKey() added to dump keyOsc and the connections
 *               it references to text files.
 *
 * 2002-11-15/FK Changed the name from osctest.c to tonegen.c.
 *               Began to change to a single connection event queue.
 *               Removed code conditioned on ENV_COMPUTED.
 *               Made code dependent on ENV_SHAPED unconditioned and removed
 *               the symbol definition.
 *               Removed dead code in COMMENT ifdefs.
 *               Commented out old keyclick code. Setting function remains
 *               but is currently void.
 *
 * 2002-07-02/FK Found the source of the key release click. It was the
 *               key compression feature that was no longer compatible
 *               with the rewritten fragment generator. The reason for
 *               that was in turn that the when a release buffer
 *               is rendered, the key down count is too low, which gives
 *               it the wrong amplification. Percussion was also subject
 *               to the same kind of timing problem. because it was being
 *               reset prior to rendering the release fragment.
 *
 * 2002-06-29/FK Crash bugs found and fixed. Still a bit of gritty sound.
 *               Percussion sounds weird on key release.
 *
 * 2002-06-27/FK First version compiled. Crashed.
 *
 * 2002-06-25/FK Began work on a major rewrite of the fragment generator.
 *
 * 2002-02-23/FK Modified the on-key click to give 9 distinct clicks
 *               rather than one. Slightly better, although I wonder
 *               if one should simulate contact bounce too.
 *
 * 2002-01-21/FK Major rewrite of the wave buffer compiler. The code now
 *               performs a constrained search for the best number of waves
 *               to compile for each frequency in the given tuning. This
 *               pushes the loop artefacts down both in frequency and
 *               amplitude and it sounds better without them.
 *
 *               Also rewired the generator/key contact grid model to
 *               support three different variants of tonegenerator.
 *
 * 2002-01-19/FK Minor adjustments and adding comments.
 *
 * 2002-01-18/FK Fixed a bug in the tone generator initialization. Higher
 *               notes copied the lowest note buffers after they had been
 *               attenuated. As a result, the higher notes had buffers that
 *               were attenuated twice.
 *
 * 2002-01-14/FK Further adjustments to key compression. It must now help
 *               to balance the normalisation of the output value in the
 *               tone generator.
 *
 * 2002-01-06/FK Changed the keycompression to count the number of depressed
 *               keys, rather than just look at the count of active
 *               oscillators. Truly, I do not know what causes this
 *               effect, but my best guess is a decreased signal level as
 *               more and more contacts close with the busbars. The assumption
 *               is of course that just as the line from the TG provides the
 *               bus with signal, its connection with earth also drops the
 *               voltage in the bus slightly, and many parallell connections
 *               only makes the situation worse.
 *
 * 2002-01-05/FK Fixed a bug in the osc-key assignment. 16' and 8' gave the
 *               the same generators.
 *
 *               Added KEYCOMPRESSION as a conditional compilation option.
 *               This simulates a volume drop when several keys on the
 *               manual are depressed simultaneously. Currently it is
 *               dependent on the number of oscillators which probably is
 *               not correct, since the effect very likely is a consequence of
 *               the number of connections to the buses.
 *
 * 2002-01-04/FK Tone generator upgraded to hold individual wave buffers.
 *               We do this in order to support varying volumes in the
 *               different generators. In the process we learned something
 *               interesting. Version 1 of the tonegenerator had 12 sample
 *               buffers, one for each note. Higher octave oscillators reused
 *               those buffers, but had position increments of 2, 4, 8 etc,
 *               thus playing it back at a higher rate. However, since the
 *               samples were the same all oscillators had the same volume.
 *               In order to compensate for this and give each oscillator
 *               its own output level, version 2 modified the oscillator
 *               setup code to give each oscillator its own buffer of three
 *               cycles, scaled in amplitude. This turned out to sound bad,
 *               because of looping clicks that gave rise to odd harmonics.
 *               To get rid of the unwanted artefacts version 3 made sure
 *               that each buffer had at least 512 samples. However, this
 *               only resulted in a click noise that had the same frequency
 *               in the oscillators were it occurred. Version 4, finally,
 *               went back to the version 1 solution, but rather than reusing
 *               the actual buffers in the lowest octave, version 4 made
 *               copies of them and scaled the amplitude of the copy. The
 *               clicks disappeared (almost).
 *
 *               Now, it may appear wasteful to cross a buffer at twice
 *               the original playback speed, because half or more of the
 *               samples will not be ever used. But, that is only true if
 *               the buffer length is even. If the buffer length is odd,
 *               the playback point will start on samples:
 *                               0, 1, 2, ... n-1, 0, 1, ...
 *               where n is the position increment (1, 2, 4, 8, ...).
 *               Thus, all oscillators with odd length buffers make full
 *               use of their buffers at all increments.
 * 
 *               It is of course natural to consider ways in which ALL
 *               buffers could be utilized in this way. One way would be
 *               to force all buffers to be odd, but that is not good
 *               for certain frequencies.
 *               
 *
 * 2001-11-30/FK Oscillator test using a sample buffer for the basic
 *               sinusoid and 9 individually updated read positions.
 */
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <limits.h>
#include <math.h>
#include <assert.h>
#include "main.h"
#include "tonegen.h"
#include "radium.h"
#include "midiIn.h"

/* These are assertion support macros. */
/* In range? : A <= V < B  */
#define inRng(A,V,B) (((A) <= (V)) && ((V) < (B)))
/* Is a B a valid bus number? */
#define isBus(B) ((0 <= (B)) && ((B) < 9))
/* Is O a valid oscillator number? */
#define isOsc(O) ((0 <= (O)) && ((O) < 128))

#ifndef BUFFER_SIZE_SAMPLES
#define BUFFER_SIZE_SAMPLES 128
#endif

#define KEYCOMPRESSION

#ifdef KEYCOMPRESSION
#define MAXKEYS 128
static int keyCompTable[MAXKEYS];
static int keyDownCount = 0;
#define KEYCOMP_ZERO_LEVEL 48
#endif /* KEYCOMPRESSION */

/*
 * The size of the connection message queues.
 */
#define CNXQSZ 1024
static unsigned short   cnxQueue [CNXQSZ];     /* Connect events */

static unsigned short * cnxqWriter = cnxQueue; /* Write pointer */
static unsigned short * cnxqReader = cnxQueue; /* Read pointer */
static unsigned short * cnxqEnd = &(cnxQueue[CNXQSZ]);

/* 2002-11-15: Switching to a single queue for serialization reasons. */
/* static unsigned short   dscQueue [CNXQSZ];     /* Disconnect events */
/* static unsigned short * dscqWriter = dscQueue; /* Write pointer */
/* static unsigned short * dscqReader = dscQueue; /* Read pointer */
/* static unsigned short * dscqEnd = &(dscQueue[CNXQSZ]); */

/*
 * 2002-11-21: Layout changed again. This time the events generated by
 * noteOn/noteOff only contains indexes into the connection table. In that
 * table are records that describe the connection and its state.
 * 2002-11-15: New layout defined.
 * The connection queue element layout is:
 *
 * 15 14 13 12 11 10  9  8  7  6  5  4  3  2  1  0
 *  B +--------------------A---------------------+
 *
 * A: 0-14, 15 bits; 0-32767    The connection index.
 * B: 15,    1 bit;  0-1        0 = contact opened, 1 = contact closed.
 */

/* Event field access macros */
#define CX_CLOSE 0x8000
#define CX_OPEN  0x0000
#define CX_CNXMSK (~CX_CLOSE)

/* Get event : true if contact closed, false if contact opened */
#define CX_GETEVT(E) ((E) & CX_CLOSE)

/* Get connection index. */
#define CX_GETCNX(E) ((E) & CX_CNXMSK)

/* Event construction macro; C = CX_CLOSE or CX_OPEN, X = connection index. */
#define CX_SETEVT(C,X) ((C) | (X))
#define CX_INIEVT(X) ((X) & CX_CNXMSK)

/*
 * This is the instruction format for the core generator.
 */

typedef struct _coreins {
  short   opr;			/* Instruction */
  short * src;			/* Pointer to source buffer */
  int   * tgt;			/* Pointer to bus accumulation buffer */
  short * env;			/* Pointer to envelope array */
  int     cnt;			/* Sample count */
} CoreIns;

/*
 * Fixed-point arithmetic - mantissa of 10 bits.
 * ENV_BASE is the scaling applied to the attack and release envelope curves
 * for key-click.
 * ENV_NORM is number of bits of right-shift needed to normalize the result:
 * y[i] = (x[i] * e[j] * ENV_BASE) >> ENV_NORM
 */

#define ENV_NORM 10
#define ENV_BASE (1 << ENV_NORM)

/* Core instruction codes (opr field in struct _coreins). */

#define CR_CPY    0		/* Copy instruction */
#define CR_ADD    1		/* Add instruction */
#define CR_CPYENV 2		/* Copy via envelope instruction */
#define CR_ADDENV 3		/* Add via envelope instruction */

#define CR_PGMMAX 256		/* Max length of a core program */

/*
 * 2002-11-17/FK: It may be prudent to remember that we do not check
 * for overflow. 256/18=14 keys. Worst case scenario should be
 * 128 * 9 * 2 = 2304. Does that length hurt performance? Probably not
 * as much as an overrun will.
 */

static CoreIns   corePgm[CR_PGMMAX];
static CoreIns * coreWriter;
static CoreIns * coreReader;
static CoreIns * coreEnd = &(corePgm[CR_PGMMAX-1]); /* For assert only */

/* Attack and relase buffer envelopes for 9 buses. */

static short attackEnv[9][BUFFER_SIZE_SAMPLES];
static short releaseEnv[9][BUFFER_SIZE_SAMPLES];


static int envAttackModel  = ENV_CLICK;
static int envReleaseModel = ENV_LINEAR;
static int envAttackClickLevel  = ENV_BASE / 2;
static int envReleaseClickLevel = ENV_BASE / 4;
/* Minimum random length (in samples) of attack key click noise burst. */
static int envAtkClkMinLength =  8;
/* Maximum random length (in samples) of attack key click noise burst. */
static int envAtkClkMaxLength = 40;

/*
 * There is one oscillator struct for each frequency.
 * The waves array has a pointer to each pre-tapered sample buffer.
 * Reading always occurs in parallel over the buffers.
 */

struct _oscillator {

  short * * waves;		/* Pointer to array of pointers */
  float * taper;		/* Taper levels in fractions of unit gain */
  size_t  numWaves;		/* Nof elements in waves[] */

  size_t  lengthSamples;	/* Length of wave */
  size_t  lengthBufferBytes;	/* Size of each buffer */
  double  frequency;		/* The frequency (Hertz) */
  double  attenuation;		/* Signal level (0.0 -- 1.0) */
  int     pos;			/* Read position */

  int     srcPosition;		/* Cached source position */
  int     srcCount0;		/* Fragment size or less */
  unsigned int bitI;		/* This oscillator's reference set index */
  unsigned int bitM;		/* This oscillator's reference set bitmask */

};

/*
 * A matrix of these structs is used in the initialization stage.
 * The matrix is indexed by MIDI note nr and bus number. Each element
 * thus models the connection between a tonegenerator and a bus.
 * Whole or part of the matrix can be loaded during configuration, while
 * unitialized elements are always set to default values.
 * We load or compute MIDI->oscillator, foldback and tapering.
 * The information is then used to create the wave buffers and
 * runtime version of the keyOsc and connections arrays.
 */
typedef struct _tmbassembly {	/* Tone generator, Manual and Bus assembly */
  float          taper;		/* Taper value in dB */
  unsigned char  oscNr;		/* Oscillator number */
  unsigned char  inited;	/* 0 = Element not set, 1 = set */
  unsigned char  waveSlot;	/* Index into oscillator's wavetable */
} TMBAssembly, * TMBAssemblyPtr;

/*
 * Matrix of tmb assembly structs. There is one row for each MIDI note
 * and one column for each bus.
 */
static TMBAssembly TMBAS[128][9];

/*
 * Connection descriptor structure. This structure is the runtime description
 * of the connection between a resistance wire and a bus.
 */
typedef struct _connection {
  unsigned char osc;		/* Oscillator that provides the wire */
  unsigned char tpx;		/* Taper index in that oscillator */
  unsigned char bus;		/* Bus number */
  unsigned short acx;		/* Position in list of active connections */
} Connection, * ConnectionPtr;

/*
 * Two manuals with five octaves plus one top c is the number of keys (K)
 * K = (5 * 12 + 1) * 2 = 122
 * Pedals are P = 2 * 12 + 1 = 25
 * C = 9K + 2P = 1148
 */
#define MAX_CONNECTIONS 1148
static Connection connections [MAX_CONNECTIONS];
static size_t connections_Length = 0;

static int activeCnx [MAX_CONNECTIONS];
static size_t actCnxEnd = 0;

static int addedCnx [MAX_CONNECTIONS];
static size_t addedCnxEnd = 0;

/*
 * This flag is used in the Connection.acx. When set, the field refers to the
 * addedCnx[] list. When cleared, the field refers to activeCnx field.
 */
#define CNX_ADDED      0x8000	/* 07-may-2003 */
#define CNX_ADDED_MASK 0x7FFF	/* 07-may-2003 */

/*
 * Our master tuning frequency.
 */
static double tuning = 440.0;

/*
 * Not all of these are used.
 */
static struct _oscillator oscillators [128];

/*
 * Vector of active keys, used to correctly manage
 * sounding and non-sounding keys.
 */
static unsigned int activeKeys [128];

/*
 * Each bus is assigned one bit to help us keep track of which generator
 * is feeding which bus. When a key in the organ model is depressed, it
 * connects nine oscillators to each of the nine buses. The core instruction
 * compiler and interpreter use these bits to keep track of buses that need
 * not be written or read when the next sample fragment is assembled.
 */
static unsigned int oscBusFlags [B_size] = {
  0x0001,
  0x0002,

  0x0004,
  0x0008,
  0x0010,
  0x0020,

  0x0040,
  0x0080,
  0x0100
};

/*
 * The keyOsc matrix defines, for each (MIDI) key, the oscillator
 * structs that gets connected to the 9 buses by that particular key.
 * 2002-11-21/FK The width is retained, but the format is simpler: the
 * close/open bit and then the connection index.
 * 2002-11-18/FK We widen this so that we can precompile all the
 * lower 15 bits (the ABC fields).
 */

static unsigned short keyOsc [128][B_size];

/*
 * The drawBars array define the settings of the drawbars.
 * In the organ there are only 9 possible values (0, 1, ... , 8),
 *
 *     busOutput = (sum * drawBars[i]) / 8;
 *
 * This model assumes linearity of attenuation, and could of course
 * be modified to include further values, to model dirt, wear or other
 * factors that cause a deviation from the ideal value.
 */

static unsigned int drawBars [B_size];

/*
 * Here is where we place the sum of the oscillators connected
 * to each bus. We keep these int so that we may accumulate some
 * overflow. Maybe not needed, really.
 */

static int busSum [B_size][BUFFER_SIZE_SAMPLES];

/*
 * True when percussion is enabled. With percussion enabled, the output
 * from the 1' bus is muted, to simulate its operation as the percussion
 * trigger.
 */
static int percEnabled = FALSE;

/*
 * When percussion is enabled, and percFirst is TRUE, the percussion tone
 * is taken from the first harmonic (4'), otherwise from the second
 * harmonic (2 2/3').
 */
static int percFirst = TRUE;


static int percIsSoft;		/* Runtime toggle */
static int percIsFast;		/* Runtime toggle */

static double percFastDecaySeconds = 1.0; /* Config parameter */
static double percSlowDecaySeconds = 4.0; /* Config parameter */


/* Alternate percussion 25-May-2003/FK */

/*
 * Runtime: The instantaneous gain applied to the percussion signal.
 */
static float percEnvGain;

/*
 * Runtime: The starting value of percEnvGain when all keys are released.
 */
static float percEnvGainReset;

/*
 * Runtime: The percEnvGain value is updated by being multiplied by this
 * variable between each sample. The value is close to, but less than 1.
 */
static float percEnvGainDecay;

/*
 * Config: scaling factor used to raise the percussion gain reset level
 * to audible levels. Can be viewed as overall 'volume' for percussion.
 */
static float percEnvScaling = 26.0;

/*
 * Runtime: Normal volume reset value, copied to percEnvGainReset when
 * normal percussion is selected.
 */
static float percEnvGainResetNorm     = 1.0; 

/*
 * Runtime: Soft volume reset value, copied to percEnvGainReset when
 * soft percussion is selected.
 */
static float percEnvGainResetSoft     = 0.707107;

/*
 * Runtime: Fast and Normal decay multiplier, copied to percEnvGainDecay
 * when that combination is selected.
 */
static float percEnvGainDecayFastNorm = 0.9995; /* Runtime select */

/*
 * Runtime: Fast and Soft decay multiplier, copied to percEnvGainDecay
 * when that combination is selected.
 */
static float percEnvGainDecayFastSoft = 0.9995; /* Runtime select */

/*
 * Runtime: Slow and Normal decay multiplier, copied to percEnvGainDecay
 * when that combination is selected.
 */
static float percEnvGainDecaySlowNorm = 0.9999; /* Runtime select */

/*
 * Runtime: Slow and Soft decay multiplier, copied to percEnvGainDecay
 * when that combination is selected.
 */
static float percEnvGainDecaySlowSoft = 0.9999; /* Runtime select */

/*
 * This variable determines the model simulated by the tonegenerator
 * and the keyboard wiring.
 */
static int tgVariant = TG_91FB12;

/*
 * This variable determines the precision with which the wave buffers
 * for the tonegenerator are sized and created.
 */
static double tgPrecision = 0.01;

/*
 * The name (if any) of the PostScript dump file.
 */
static char * tgDumpFname = NULL;
static char   tgDumpFNBuffer [PATH_MAX + 1];

/*
 * Equalisation macro selection.
 */
#define EQ_SPLINE 0
#define EQ_PEAK24 1		/* Legacy */
#define EQ_PEAK46 2		/* Legacy */

static int eqMacro = EQ_SPLINE;
static double eqvCeiling = 1.0;	/* Normalizing manual osc eq. */
static double eqvAtt [128];	/* Values from config file */
static char   eqvSet [128];	/* Value is set by a config command */

/*
 * EQ_SPLINE parameters.
 */
static double eqP1y =  1.0;
static double eqR1y =  0.0;
static double eqP4y =  0.1;
static double eqR4y = -1.0;

/*
 * Oscillator reference gain. It must be less than one in order to
 * allow some dynamic range for the percussion.
 */
static double oscRefGain = 0.7079457843841379; /* - 3dB */

/*
 * Amplitudes of tonewheel harmonics
 */
#define MAX_PARTIALS 12
static double wheelHarmonics [MAX_PARTIALS] = { 1.0 };


/*
 * ================================================================
 */

/*
 * This function converts from a dB value to a fraction of unit gain.
 * Both values describe the relation between two levels.
 */

double dBToGain (double dB) {
  return pow (10.0, (dB / 20.0));
}

/*
 * This routine sets the tonegenerator model. The call must be made
 * before calling initToneGenerator() to have effect and is thus the
 * target for startup configuration values.
 */
void setToneGeneratorModel (int variant) {
  switch (variant) {
  case TG_91FB00:
  case TG_82FB09:
  case TG_91FB12:
    tgVariant = variant;
    break;
  }
}

/*
 * This routine sets the tonegenerator's wave precision. The call must
 * be made before calling initToneGenerator() to have effect. It is the
 * target of startup configuration values.
 */
void setWavePrecision (double precision) {
  if (0.0 < precision) {
    tgPrecision = precision;
  }
}

/*
 * Sets the tuning.
 */
void setTuning (double refA_Hz) {
  if ((220.0 <= refA_Hz) && (refA_Hz <= 880.0)) {
    tuning = refA_Hz;
  }
}

/*
 * This function models the attenuation of the tone generators.
 * Note that the tone generator parameters expect the first/lowest
 * generator to be numbered 1.
 * The function basically takes the value 1.0 and subtracts the value of the
 * function  -x^2 * w   in the range v..u. Thus, w is a scale function and
 * should be in the range 0..1. 
 */
static double damperCurve (int thisTG, /* Number of requested TG */
			   int firstTG,	/* Number of first TG in range */
			   int lastTG, /* Number of last TG in range */
			   double w, /* Weight */
			   double v, /* Lower x of -x^2 */
			   double u) /* Upper x of -x^2 */
{
  double x = ((double) (thisTG - firstTG)) / ((double) (lastTG - firstTG));
  double z = (x * (u - v)) - u;
  return 1.0 - w * z * z;
}

/*
 * Implements a built-in oscillator equalization curve.
 * Applies a constrained spline (p1x=0, p4x=0) to the oscillator's
 * attenuation.
 */
static int apply_CH_Spline (int nofOscillators,
			    double p1y,
			    double r1y,
			    double p4y,
			    double r4y) {
  int i;
  double k = nofOscillators - 1;
  for (i = 1; i <= nofOscillators; i++) {
    double t = ((double) (i - 1)) / k;
    double tSq = t * t;
    double tCb = tSq * t;
    double r;
    double a;

    r = p1y * ( 2.0 * tCb - 3.0 * tSq + 1.0)
      + p4y * (-2.0 * tCb + 3.0 * tSq)
      + r1y * ( tCb - 2.0 * tSq + t)
      + r4y * ( tCb - tSq);

    a = (r < 0.0) ? 0.0 : (1.0 < r) ? 1.0 : r;
    oscillators[i].attenuation = a;
  }

  return 0;
}

/*
 * Implements a built-in oscillator equalization curve.
 */
static int applyOscEQ_peak24 (int nofOscillators) {
  int i;

  for (i = 1; i <= 43; i++) {
    oscillators[i].attenuation = damperCurve (i,  1, 43, 0.2, -0.8,  1.0);
  }

  for (i = 44; i <= 48; i++) {
    oscillators[i].attenuation = damperCurve (i, 44, 48, 1.6, -0.4, -0.3);
  }

  for (i = 49; i <= nofOscillators; i++) {
    oscillators[i].attenuation =
      damperCurve (i, 49, nofOscillators, 0.9, -1.0, -0.7);
  }

  return 0;
}

/*
 * Implements a built-in oscillator equalization curve.
 */
static int applyOscEQ_peak46 (int nofOscillators) {
  int i;

  for (i = 1; i <= 43; i++) {
    oscillators[i].attenuation = damperCurve (i,  1, 43,  0.3, 0.4,  1.0);
  }

  for (i = 44; i <= 48; i++) {
    oscillators[i].attenuation = damperCurve (i, 44, 48,  0.1, -0.4, 0.4);
  }

  for (i = 49; i <= nofOscillators; i++) {
    oscillators[i].attenuation =
      damperCurve (i, 49, nofOscillators, 0.8, -1.0, -0.3);
  }

  return 0;
}

/*
 * This function returns the number of samples required to produce a
 * waveform of the given frequency below the provided precision.
 * The function will not attempt solutions above the maximum number
 * of samples. If no solution is found the best solution is returned.
 * The error/precision value is the positive distance between the
 * ideal number of samples and the nearest integer.
 * As for the maximum number of samples to try, higher frequencies fit
 * more waves into the same memory, but also make longer leaps around
 * the unit circle, making it harder to land close to the starting point.
 * Practical experiments seems to indicate that there is no gain in
 * making the maxSamples parameter dependent on the frequency.
 *
 * @param Hz         The frequency of the wave.
 * @param precision  The absolute value error threshold. Figures in the
 *                   range 0.1 - 0.01 may be adequate. Lower thresholds
 *                   will result in longer (more memory) solutions.
 * @param minSamples The minimum number of samples to use.
 * @param maxSamples The maximum number of samples to use.
 *
 * @return  The number of samples to allocate for the wave.
 */

static size_t fitWave (double Hz,
		       double precision,
		       int minSamples,
		       int maxSamples) {
  double minErr = 99999.9;
  double minSpn = 0.0;
  int    i;
  int minWaves;
  int maxWaves;
  double w;
  minWaves = ceil  ((Hz * (double) minSamples) / SampleRateD);
  maxWaves = floor ((Hz * (double) maxSamples) / SampleRateD);

  assert (minWaves <= maxWaves);

  for (i = minWaves; i <= maxWaves; i++) {
    double nws = (SampleRateD * i) / Hz; /* Compute ideal nof samples */
    double spn = rint (nws);	/* Round to a discrete nof samples  */
    double err = fabs (nws - spn); /* Compute mismatch */
    if (err < minErr) {		/* Remember best so far */
      minErr = err;
      minSpn = spn;
    }
    if (err < precision) break;	/* If ok, stop searching. */
  }

  assert (0.0 < minSpn);

  return (size_t) minSpn;
}

/*
 * This routine writes the sample buffer for a simulated tone wheel.
 * In addition to the sine wave of the fundamental frequency, the routine
 * also allows the specification of the first and second chromatic harmonics.
 * There is no specification of phase.
 * This is an attempt to simulate the effect of transformer distortion and
 * generally make the output more 'warm'. If that really is the shape
 * generated by the real tonewheels remains to be seen (or heard). Anyway,
 * it is my current best guess as to the direction to go. Remember, the
 * chromatic harmonics are different from the harmonics added by the
 * drawbar system. The tonewheels are tuned to the tempered scale, and
 * thus will 'beat' very subtly against the chromatics. 
 *
 * @param buf           Pointer to wave buffer
 * @param sampleLength  The number of 16-bit samples in the buffer
 * @param ap            Array of partial amplitudes
 * @param apLen         Nof elements in ap[]
 * @param f3a           Amplitude of second harmonic (Hz = 3 * fundamental).
 * @param attenuation   Final volume of wave (0.0 -- 1.0).
 * @param f1Hz          Frequency of the fundamental.
 *
 * Please note that the amplitudes of the fundamental and harmonic frequencies
 * are normalised so that the volume of the composite curve is 1.0. This
 * means that if you supply  f1a=1.0, f2a=0.1 and f3a=0.05 the actual
 * proportions will be:
 *
 *         f1a = 1.00 / 1.15 = 0.8695652173913044
 *         f2a = 0.10 / 1.15 = 0.08695652173913045
 *         f3a = 0.05 / 1.15 = 0.04347826086956522
 *
 * Be aware of this, or make sure that the arguments sum to 1.
 */

static void writeSamples (unsigned char * buf,
			  size_t sampleLength,
			  double ap [],
			  size_t apLen,
			  double attenuation,
			  double f1Hz)
{
  static double fullCircle = 2.0 * M_PI;
  double apl[MAX_PARTIALS];
  double plHz[MAX_PARTIALS];
  double aplSum;
  double U;
  short * yp = (short *) buf;
  int i;

  for (i = 0, aplSum = 0.0; i < MAX_PARTIALS; i++) {
    /* Select absolute amplitude */
    apl[i] = (i < apLen) ? ap[i] : 0.0;
    /* Accumulate normalization base */
    aplSum += fabs (apl[i]);
    /* Compute harmonic frequency */
    plHz[i] = f1Hz * ((double) (i + 1));
    /* Prevent aliasing; mute just below the Nyquist rate */
    if ((SampleRateD * 0.5) <= plHz[i]) {
      apl[i] = 0.0;
    }
  }

  /* Normalise amplitudes */

  U = (attenuation * 32767.0) / aplSum;

  for (i = 0; i < sampleLength; i++) {

    int j;
    double s = 0.0;

    for (j = 0; j < MAX_PARTIALS; j++) {
      s +=
	apl[j] * sin (drem ((plHz[j] * fullCircle * (double) i) / SampleRateD,
			    fullCircle));
    }

    /* 24-sep-2003/FK
     * Noise-shaping in an attempt to diffuse the quantization artifacts.
     * It did not work of course, but may add some analogue credibility
     * so it can be in for the moment.
     */

#if 1
      *yp = (rand () < (RAND_MAX >> 1)) ? 1 : 0;
      *yp++ += (short) (U * s);      
#else
      *yp++ = (short) (U * s);
#endif

  } /* for */
}

#ifdef COMMENT /* 20-aug-2003 backup */
static void writeSamples (unsigned char * buf,
			  size_t sampleLength,
			  double f1a,
			  double f2a,
			  double f3a,
			  double attenuation,
			  double f1Hz)
{
  static double fullCircle = 2.0 * M_PI;
  double f2Hz = f1Hz * 2.0;
  double f3Hz = f1Hz * 3.0;
  double U = (attenuation * 32767.0) / (f1a + f2a + f3a);
  short * yp = (short *) buf;
  int i;

  /*
   * Prevent aliasing from overenthusiastic tuning or too high
   * harmonics.
   */

  if ((SampleRateD * 0.5) <= f1Hz) f1a = 0.0;
  if ((SampleRateD * 0.5) <= f2Hz) f2a = 0.0;
  if ((SampleRateD * 0.5) <= f3Hz) f3a = 0.0;

  for (i = 0; i < sampleLength; i++) {

    double v1 = (i * f1Hz * fullCircle) / SampleRateD;
    double v2 = (i * f2Hz * fullCircle) / SampleRateD;
    double v3 = (i * f3Hz * fullCircle) / SampleRateD;

    double a1 = drem (v1, fullCircle);
    double a2 = drem (v2, fullCircle);
    double a3 = drem (v3, fullCircle);

    *yp++ = (short) (U * ((sin (a1) * f1a) +
			  (sin (a2) * f2a) +
			  (sin (a3) * f3a)));
  } /* for */
}
#endif /* COMMENT */

/*
 * This routine initializes the oscillators.
 *
 * @param variant  Selects one of tree modelled tonegenerators:
 *                 0 : 91 generators, lowest generator is C-2
 *                 1 : 82 generators, lowest generator is A-2
 *                 2 : 91 generators, lowest generator is C-2, generators
 *                     1--12 have distinct 2f and 3f harmonics.
 *
 * @param precision  The loop precision value. See function fitWave().
 */
static void initOscillators (int variant, double precision) {
  int i;
  int j;
  double baseTuning;
  int nofOscillators;
  int tuningOsc = 10;
  int complexWheels = FALSE;
  struct _oscillator * osp;
  struct _oscillator * xsp;

  switch (variant) {

  case 0:
    nofOscillators = 91;
    baseTuning     = tuning / 8.0;
    tuningOsc      = 10;
    complexWheels  = FALSE;
    break;

  case 1:
    nofOscillators = 82;
    baseTuning     = tuning / 8.0;
    tuningOsc      = 1;
    complexWheels  = FALSE;
    break;

  case 2:
    nofOscillators = 91;
    baseTuning     = tuning / 8.0;
    tuningOsc      = 10;
    complexWheels  = TRUE;
    break;

  default:
    assert (0);
  } /* switch variant */

  /*
   * Apply equalisation curve. This sets the attenuation field in the
   * oscillator struct.
   */

  switch (eqMacro) {
  case EQ_SPLINE:
    apply_CH_Spline (nofOscillators, eqP1y, eqR1y, eqP4y, eqR4y);
    break;
  case EQ_PEAK24:
    applyOscEQ_peak24 (nofOscillators);
    break;
  case EQ_PEAK46:
    applyOscEQ_peak46 (nofOscillators);
    break;
  }

  for (i = 1; i <= nofOscillators; i++) {
    int j;
    size_t wszs;		/* Wave size samples */
    size_t wszb;		/* Wave size bytes */
    size_t bufBlkBytes;		/* Nof bytes needed for all buffers */
    short * bufBlk;		/* Pointer to all buffer block */
    double t;

    osp = & (oscillators[i]);

    if (eqvSet[i] != 0) {
      osp->attenuation = eqvAtt[i];
    }

    osp->pos       =  0;
    osp->bitI      = i / 32;

    assert (osp->bitI < 4);

    osp->bitM      = 1 << (i % 32);

    t = (double) (i - tuningOsc);
    osp->frequency = baseTuning * pow (2.0, t / 12.0);

    /*
     * The oscGenerateFragment() routine assumes that samples are at least
     * BUFFER_SIZE_SAMPLES long, so we must make sure that they are.
     * If the loop fits in n samples, it certainly will fit in 2n samples.
     */
    wszs = fitWave (osp->frequency,
		    precision,
		    3 * BUFFER_SIZE_SAMPLES, /* Was x1 */
		    2048);

    /* Compute the number of bytes needed for exactly one wave buffer. */

    wszb = wszs * 2;

    /* Allocate an array of pointers to buffers */
    osp->waves = (short * *) malloc (osp->numWaves * sizeof (short *));
    if (osp->waves == NULL) {
      fprintf (stderr,
	       "FATAL:Memory allocation failed in initOscillators, Offending request:\n");
      fprintf (stderr,
	       "Wave index table for osc=%d, elems=%d\n",
	       i,
	       osp->numWaves);
      exit (1);
    }

    /* Multiply wszb by osc->numWaves to find total buffer size needed */

    bufBlkBytes = wszb * osp->numWaves;

    /* Allocate memory for all buffers */

    bufBlk = (short *) malloc (bufBlkBytes);

    if (bufBlk == NULL) {
      fprintf (stderr,
	       "FATAL:Memory allocation failed in initOscillators. Offending request:\n");
      fprintf (stderr,
	       "OSCILLATOR=%d %lf Hz, %d bytes\n",
	       i,
	       osp->frequency,
	       bufBlkBytes);
      exit (1);
    }

    /*
     * Annotate the number of samples. Since all tapering buffers are
     * of the same length we only need one instance of this information.
     */

    osp->lengthSamples = wszs;

    if (complexWheels && ((1 <= i) && (i <= 12))) {
/*       f1a = 0.80; */
/*       f2a = 0.12; */
/*       f3a = 0.08; */
    }
    else {
/*       f1a = 1.00; */
/*       f2a = 0.00;//0.07; */
/*       f3a = 0.05;//0.02; */
    }

    /* Initialize each buffer, multiplying attenuation with taper. */

    for (j = 0; j < osp->numWaves; j++) {
      osp->waves[j] = bufBlk + (j * wszs);
      writeSamples ((unsigned char *) (osp->waves[j]),
		    osp->lengthSamples,
		    wheelHarmonics,
		    (size_t) MAX_PARTIALS,
		    osp->attenuation * (double) osp->taper[j],
		    osp->frequency);
    }

  }
}

/* ================================================================
   setTMBDefaultMap
   This routine is called from initTMBAssembly to fill in the TMBAS
   matrix. The routine determines the default tapering to apply and
   then updates the TMBAS matrix via an access function that in turn
   determines if the cell is already initialized or needs the default.
   @param midiNote  The MIDI note number.
   @param busNr     The number of the bus.
   @param oscNr     The number of the oscillator mapped by foldback
                    to this particular oscillator-bus connection.
   ================================================================ */

/* These units are in dB */

#define taperMinusThree -10.0
#define taperMinusTwo    -7.0
#define taperMinusOne    -3.5
#define taperReference    0.0
#define taperPlusOne      3.5
#define taperPlusTwo      7.0

static void setTMBDefaultMap (int midiNote, int busNr, int oscNr) {
  float tapering = taperReference;

  switch (busNr) {

  case 0:			/* 16 */
    if (midiNote < 48) {	/* C-1 */
      tapering = taperMinusThree;
    }
    else if (midiNote < 53) {	/* F-1 */
      tapering = taperMinusTwo;
    }
    else if (midiNote < 60) {	/* C0 */
      tapering = taperMinusOne;
    }
    else if (midiNote < 72) {	/* C1 */
      tapering = taperReference;
    }
    else if (midiNote < 84) {	/* C2 */
      tapering = taperPlusOne;
    }
    else {
      tapering = taperPlusTwo;
    }
    break;

  case 1:			/* 5 1/3 */
    if (midiNote < 51) {	/* Eb-1 */
      tapering = taperMinusOne;
    }
    else if (midiNote < 74) {	/* D#1 */
      tapering = taperReference;
    }
    else if (midiNote < 86) {	/* D#2 */
      tapering = taperPlusOne;
    }
    else {
      tapering = taperPlusTwo;
    }
    break;

  case 2:			/* 8 */
    if (midiNote < 53) {	/* F-1 */
      tapering = taperMinusTwo;
    }
    else if (midiNote < 58) {	/* A#-1 */
      tapering = taperMinusOne;
    }
    else if (midiNote < 73) {	/* C#1 */
      tapering = taperReference;
    }
    else if (midiNote < 85) {	/* C2 */
      tapering = taperPlusOne;
    }
    else {
      tapering = taperPlusTwo;
    }
    break;

  case 3:			/* 4 */
    if (midiNote < 53) {	/* F-1 */
      tapering = taperMinusOne;
    }
    else if (midiNote < 75) {	/* C0 */
      tapering = taperReference;
    }
    else {
      tapering = taperMinusOne;
    }
    break;

  case 4:			/* 2 2/3 */
    if (midiNote < 50) {
      tapering = taperPlusTwo;
    }
    else if (midiNote < 56) {
      tapering = taperPlusOne;
    }
    else if (midiNote < 76) {
      tapering = taperReference;
    }
    else if (midiNote < 86) {
      tapering = taperMinusOne;
    }
    else {
      tapering = taperMinusTwo;
    }
    break;

  case 5:			/* 2 */
    if (midiNote < 48) {
      tapering = taperPlusTwo;
    }
    else if (midiNote < 51) {
      tapering = taperPlusOne;
    }
    else if (midiNote < 77) {
      tapering = taperReference;
    }
    else if (midiNote < 90) {
      tapering = taperMinusOne;
    }
    else {
      tapering = taperMinusTwo;
    }
    break;

  case 6:			/* 1 3/5 */
    if (midiNote < 50) {
      tapering = taperPlusOne;
    }
    else if (midiNote < 78) {
      tapering = taperReference;
    }
    else if (midiNote < 86) {
      tapering = taperMinusOne;
    }
    else {
      tapering = taperMinusTwo;
    }
    break;

  case 7:			/* 1 1/3 */
    if (midiNote < 79) {
      tapering = taperReference;
    }
    else if (midiNote < 84) {
      tapering = taperMinusOne;
    }
    else {
      tapering = taperMinusTwo;
    }
    break;

  case 8:			/* 1 */
    if (midiNote < 79) {
      tapering = taperReference;
    }
    else {
      tapering = taperMinusTwo;
    }
    break;
  }

  {
    TMBAssemblyPtr T = &(TMBAS[midiNote][busNr]);
    if (!T->inited) {
      T->taper  = tapering;
      T->oscNr  = oscNr;
      T->inited = 1;
    }
  }
  
}

/*
 * This routine initializes the mapping from a manual (keyboard) key
 * and the tonegenerators it connects to the buses. The variant
 * parameter should be the same as used in the call to initOscillators().
 *
 * @param variant  The keyboard mapping variant selected:
 *                 0 : 91 oscillators, 61 keys, no lower foldback.
 *                 1 : 82 oscillators, 61 keys, foldback from A in the lowest
 *                     octave.
 *                 2 : 91 oscillators, 61 keys, foldback from C in the next
 *                     to lowest octave.
 *                 All variants use foldback at the upper part of the manual.
 */
static void initTMBAssembly (int variant) {
  int noteShift;
  int firstOscillator;
  int lastOscillator;
  int bus;
  int note;
  int tg;			/* Tone generator (ie oscillator) */

  int offset[9] = {
    0,				/* Sub 16 ' */
    12+7,			/* Sub harmonic 5 1/3 ' */
    12,				/* Fundamental 8 ' */
    12+12,			/* Octave 4 ' */
    12+12+7,			/* Harmonic 2 2/3 ' */
    12+12+12,			/* Octave 2 ' */
    12+12+12+4,			/* Harmonic 1 3/5 ' */
    12+12+12+7,			/* Harmonic 1 1/3 ' */
    12+12+12+12			/* Octave 1 ' */
  };

  switch (variant) {

  case 0:
    noteShift  =  0;
    firstOscillator =  1;
    lastOscillator  = 91;
    break;

  case 1:
    noteShift  = -9;
    firstOscillator =  1;
    lastOscillator  = 82;
    break;

  case 2:
    noteShift = 0;
    firstOscillator = 13;
    lastOscillator =  91;
    break;

  default:
    assert (0);
    break;
  }

  /*
   * The following code is not immediately transparent. Anyway, this is
   * what happens:
   *
   * The outer loop iterates over the MIDI notes that we recognise.
   * Their values are determined by the variables firstMIDINote and
   * lastMIDINote respectively, and they express valid MIDI note numbers.
   *
   * The inner loop iterates over the 9 oscillators that each key, or in
   * this case MIDI note, will connect to the corresponding output bus.
   *
   * One way of looking at it, is that for each of the 9 contacts closed
   * by the key (MIDI note) we find the correct tonegenerator. The first
   * assignment to tg thus maps the MIDI note to the ideal tonegenerator
   * number by including noteShift and the offset implied by the bus.
   * The noteShift is needed to model the 84 wheel generator in which
   * generator 1 has the frequency of an A. The 91 wheel generator begins
   * with a C (and indeed, this mapping assumes that firstMIDINote is a
   * C on the keyboard).
   *
   * The two while-loops take care of foldback, very simply by bumping
   * the ideal tonegenerator number up or down an octave until it falls
   * within the range of existing generators.
   *
   * Finally, the TMBAS table is filled in so that we can compile the
   * connections array and the keyOsc table.
   */

  for (note = firstMIDINote; note <= lastMIDINote; note++) {
    for (bus = 0; bus < 9; bus++) {
      tg = (1 + note - firstMIDINote) + noteShift + offset[bus];
      while (tg < firstOscillator) tg += 12;
      while (lastOscillator < tg ) tg -= 12;
      setTMBDefaultMap (note, bus, tg);
    }
  }

}

/* ================================================================
   initFromTMBMatrix
   This routine examines the TMBAS matrix and compiles the number
   of waves needed for correctly tapering each oscillator. In reality,
   the taper is created by connecting a resistance wire from the
   oscillator to the particular bus key. We could simulate that wire
   by using a multiplier in the fragment generator, but using pre-
   multiplied wave buffers trades memory for speed.
   ================================================================ */
#define MAX_TAPERS 8
static void initFromTMBMatrix () {
  int i;
  int o;			/* Oscillator index */
  int n;			/* MIDI note index */
  int b;			/* Bus index */
  float tapers[128][MAX_TAPERS]; /* List of unique taper lvls in each osc. */
  size_t nofWaves[128];		/* One counter per oscillator */
  int maxTaperdBInited = FALSE;
  float maxTaperdB = 0.0;	/* Maximum taper dB value found. */
  double scale;			/* Scale factor for taper gains. */

  /* Initialize the counters */

  for (o = 0; o < 128; o++) {
    nofWaves[o] = 0;
  }

  /* First pass, find unique tapers for each oscillator and annotate TMBAS. */

  for (n = 0; n < 128; n++) {
    for (b = 0; b < 9; b++) {
      if (TMBAS[n][b].inited) {
	int   osc = TMBAS[n][b].oscNr;
	float tpr = TMBAS[n][b].taper;
	int append = TRUE;

	if ((maxTaperdB < tpr) || (!maxTaperdBInited)) {
	  maxTaperdB = tpr;
	  maxTaperdBInited = TRUE;
	}

	for (i = 0; i < nofWaves[osc]; i++) { /* Check if taper seen. */
	  if (tapers[osc][i] == tpr) {
	    TMBAS[n][b].waveSlot = i; /* Remember which slot */
	    append = FALSE;
	    break;
	  }
	}

	if (append) {		/* Not seen, append to list for this osc. */
	  if (nofWaves[osc] < MAX_TAPERS) {
	    tapers[osc][nofWaves[osc]] = tpr;
	    TMBAS[n][b].waveSlot = nofWaves[osc]; /* Remember which slot */
	    nofWaves[osc] += 1;
	  }
	  else {
	    fprintf (stderr,
		     "WARNING:More than %d unique tapers for oscillator %d.\n",
		     MAX_TAPERS,
		     osc);
	  }
	}
      }	/* if inited TMBAS element */
    } /* for bus b */
  } /* for MIDI notes n */

  /* Compute the scaling factor needed to keep the tapering levels between
     the oscillator reference gain and zero. */

  scale = oscRefGain / dBToGain ((double) maxTaperdB);

  /* Install taper information in each oscillator. */

  for (o = 0; o < 128; o++) {
    if (0 < nofWaves[o]) {
      size_t elems = nofWaves[o];
      float * flp = (float *) malloc (sizeof (float) * elems);
      if (flp == NULL) {
	fprintf (stderr,
		 "FATAL:Memory allocation failed for oscillator %d in %s\n",
		 o,
		 "initFromTMBMatrix");
	exit (1);
      }
      else {
	for (i = 0; i < elems; i++) {
	  flp[i] = (float) (scale * dBToGain ((double) tapers[o][i]));
	}
	oscillators[o].taper = flp;
	oscillators[o].numWaves = elems;
      }
    }
  }

  /*
   * At this point we have annotated each oscillator struct with the number
   * of wave buffers needed and the gain for each buffer. We have also
   * annotated the TMBAS matrix with the wave buffer's index in the
   * referenced oscillator. Thus we have all the information we need to
   * write the connections[] and keyOsc[][] tables.
   * For each initialized combination of MIDI note and bus, we allocate
   * an element in the connections table. That element holds the runtime
   * description of the connection. The index to that element is what
   * goes into the keyOsc table.
   */

  for (n = 0; n < 128; n++) {
    for (b = 0; b < 9; b++) {
      TMBAssemblyPtr TAP = &(TMBAS[n][b]);
      if (TAP->inited) {
	ConnectionPtr CXP;
	if (connections_Length < MAX_CONNECTIONS) {
	  CXP = &(connections[connections_Length]);
	  CXP->osc = TAP->oscNr;
	  CXP->bus = b;
	  CXP->tpx = TAP->waveSlot;
	  CXP->acx = 0;
	  keyOsc[n][b] = CX_INIEVT(connections_Length);
	  connections_Length += 1;
	}
	else {
	  fprintf (stderr,
		   "FATAL:Too many connections (%d).\n",
		   connections_Length);
	  exit (1);
	}
      }
    }
  }
}

/*
 * This routine sets percussion on or off.
 *
 * @param isEnabled  If true, percussion is enabled. If false, percussion
 *                   is disabled.
 */
void setPercussionEnabled (int isEnabled) {
  percEnabled = isEnabled;
}

/*
 * This routine sets the percCounterReset variable from the current
 * combination of the control flags percIsFast and percIsSoft.
 */
static void setPercussionResets () {
  if (percIsFast) {

    /* Alternate 25-May-2003 */
    percEnvGainDecay =
      percIsSoft ? percEnvGainDecayFastSoft : percEnvGainDecayFastNorm;

  }
  else {			/* Slow */

    /* Alternate 25-May-2003 */
    percEnvGainDecay =
      percIsSoft ? percEnvGainDecaySlowSoft : percEnvGainDecaySlowNorm;

  }
}

/*
 * Selects fast or slow percussion.
 * @param isFast  If true, selects fast percussion decay. If false,
 *                selects slow percussion decay.
 */
void setPercussionFast (int isFast) {
  percIsFast = isFast;
  setPercussionResets ();
}

/*
 * Selects soft or normal percussion volume.
 * @param isSoft  If true, selects soft percussion. If false, selects
 *                normal percussion.
 */
void setPercussionVolume (int isSoft) {
  percIsSoft = isSoft;

  /* Alternate 25-May-2003 */
  percEnvGainReset =
    percEnvScaling * (isSoft ? percEnvGainResetSoft : percEnvGainResetNorm);

  setPercussionResets ();
}

/*
 * Selects first or second percussion harmonic.
 * 
 */
void setPercussionFirst (int isFirst) {
  percFirst = isFirst;
}

/*
 * Computes the constant value with which the percussion gain is multiplied
 * between each output sample. The gain (and the percussion signal) then
 * suffers an exponential decay similar to that of a capacitor.
 *
 * @param ig  Initial gain (e.g. 1.0), must be non-zero positive.
 * @param tg  Target gain (e.g. 0.001 = -60 dB), must be non-zero positive.
 * @param spls Time expressed as samples, or equivalently, the number of
 *             multiplies that will be applied to the ig
 */
double getPercDecayConst_spl (double ig, double tg, double spls) {
  return exp (log (tg / ig) / spls);
}

/*
 * Computes the constant value with which the percussion gain is multiplied
 * between each output sample. The gain (and the percussion signal) then
 * suffers an exponential decay similar to that of a capacitor.
 *
 * @param ig  Initial gain (e.g. 1.0), must be non-zero positive.
 * @param tg  Target gain (e.g. 0.001 = -60 dB), must be non-zero positive.
 * @param seconds Time expressed as seconds
 */
double getPercDecayConst_sec (double ig, double tg, double seconds) {
  return getPercDecayConst_spl (ig, tg, SampleRateD * seconds);
}


/*
 * This routine is called each time one of the percussion parameters
 * has been updated. It recomputes the reset values for the percussion
 * amplification and the percussion amplification decrement counter.
 */
static void computePercResets () {

  /* Compute the percussion reset values. */

  /* Alternate 25-May-2003 */
  percEnvGainDecayFastNorm = getPercDecayConst_sec (percEnvGainResetNorm,
						    dBToGain (-60.0),
						    percFastDecaySeconds);

  percEnvGainDecayFastSoft = getPercDecayConst_sec (percEnvGainResetSoft,
						    dBToGain (-60.0),
						    percFastDecaySeconds);

  percEnvGainDecaySlowNorm = getPercDecayConst_sec (percEnvGainResetNorm,
						    dBToGain (-60.0),
						    percSlowDecaySeconds);

  percEnvGainDecaySlowSoft = getPercDecayConst_sec (percEnvGainResetSoft,
						    dBToGain (-60.0),
						    percSlowDecaySeconds);

  /* Deploy the computed reset values. */

  setPercussionResets ();
}


/*
 * This routine sets the fast percussion decay time.
 * @param seconds  The percussion decay time.
 */
void setFastPercussionDecay (double seconds) {
  percFastDecaySeconds = seconds;
  if (percFastDecaySeconds <= 0.0) {
    percFastDecaySeconds = 0.1;
  }
  computePercResets ();
}

/*
 * This routine sets the slow percussion decay time.
 * @param seconds  The percussion decay time.
 */
void setSlowPercussionDecay (double seconds) {
  percSlowDecaySeconds = seconds;
  if (percSlowDecaySeconds <= 0.0) {
    percSlowDecaySeconds = 0.1;
  }
  computePercResets ();
}

/*
 * Sets the percussion starting gain of the envelope for normal volume.
 * The expected level is 1.0 with soft volume less than that.
 * @param g  The starting gain.
 */
void setNormalPercussionGain (double g) {
  percEnvGainResetNorm = (float) g;
}

/*
 * Sets the percussion starting gain of the envelope for soft volume.
 * The expected level is less than 1.0.
 * @param g  The starting gain.
 */
void setSoftPercussionGain (double g) {
  percEnvGainResetSoft = (float) g;
}

/*
 * Sets the percussion gain scaling factor.
 * @param s  The scaling factor.
 */
void setPercussionGainScaling (double s) {
  percEnvScaling = (float) s;
}

/*
 * Sets the on-key key click (like you need it: this program version clicks
 * obnoxiously all the time!). [HEY, NO LONGER! 2002-07-02/FK]
 * @param v   The click sample value (max allowed 12000).
 */
void setKeyClick (int v) {

/*    if (v < 0) { */
/*      keyOnClickReset = 0; */
/*    } */
/*    else if (12000 < v) { */
/*      keyOnClickReset = 12000; */
/*    } */
/*    else { */
/*      keyOnClickReset = v; */
/*    } */

}

void setEnvAttackModel (int model) {
  if ((0 <= model) && (model < ENV_CLICKMODELS)) {
    envAttackModel = model;
  }
}

void setEnvReleaseModel (int model) {
  if ((0 <= model) && (model < ENV_CLICKMODELS)) {
    envReleaseModel = model;
  }
}

static void setEnvClickLevel (int * L, double u) {
  int level = (int) (u * (double) ENV_BASE);
  if (level == 0) {
    level = 1;
  }
  if ((0 < level) && (level <= ENV_BASE)) {
    *L = level;
  }
}

/*
 * This parameter actually simulates the amount of noise created when the
 * closing/opening contact surfaces move against each other. Setting it
 * to zero will still click, due to the timing differences in the separate
 * envelopes (they close at different times). However, a higher value will
 * simulate more oxidization of the contacts, as in a worn instrument.
 */
void setEnvAttackClickLevel (double u) {
  setEnvClickLevel (&envAttackClickLevel, u);
}

static void setEnvAtkClkLength (int * p, double u) {
  if (p != NULL) {
    if ((0.0 <= u) && (u <= 1.0)) {
      *p = (int) (((double) BUFFER_SIZE_SAMPLES) * u);
    }
  }
}

/*
 * Sets the minimum duration of a keyclick noise burst. The unit is a fraction
 * (0.0 -- 1.0) of the adjustable range.
 */
void setEnvAtkClkMinLength (double u) {
  setEnvAtkClkLength (&envAtkClkMinLength, u);
}

/*
 * Sets the maximum duration of a keyclick noise burst.
 */
void setEnvAtkClkMaxLength (double u) {
  setEnvAtkClkLength (&envAtkClkMaxLength, u);
}

void setEnvReleaseClickLevel (double u) {
  setEnvClickLevel (&envReleaseClickLevel, u);
}

#ifdef KEYCOMPRESSION

/*
 * This routine initializes the key compression table. It is in all
 * simplicity a volume control on the output side of the tone generator,
 * controlled by the number of depressed keys.
 * The output is calculated as   y = (x * T[n]) / 48, so a value of 48
 * is equivalent to unit amplification, 24 is 0.5 etc. The reason for not
 * using a power of 2 is in the tone generator. When it is time to apply
 * the key compression, we have a value of x that is approximately 6.5 bits
 * too high. So what we do is to increase the unit amplification to 48 (which
 * is worth ~5.5 bits) and then shift down by 9.
 * I know, it does not cancel out, the result is still to high. But that is
 * where the key compression table enters handy. The more oscillators that
 * contribute, on each bus, the more we lower the volume. So its all a matter
 * of keeping the output below that nasssty digital clipping.
 * I really need to think about this...
 *
 * The first eight positions are estimated to be not so obvious.
 */
static void initKeyCompTable () {
  int i;

 /* The first two entries, 0 and 1, should be equal! */

  keyCompTable[0] = keyCompTable[1] = KEYCOMP_ZERO_LEVEL;/* 48 */
  keyCompTable[2] = 42;		/* -1.1598 dB */
  keyCompTable[3] = 38;		/* -2.0291 dB */
  keyCompTable[4] = 36;		/* -2.4987 dB */
  keyCompTable[5] = 34;		/* -2.9952 dB */
  keyCompTable[6] = 32;		/* -3.5218 dB */
  keyCompTable[7] = 30;		/* -4.0823 dB */
  keyCompTable[8] = 28;		/* -4.6815 dB */
  keyCompTable[9] = 27;		/* -4.9975 dB */
  keyCompTable[10] = 27;	/* -4.9975 dB */

  for (i = 11; i < MAXKEYS; i++) {
    keyCompTable[i] = 27 - (i / 16);
  }

}

#endif /* KEYCOMPRESSION */

/*
 * Dump the tone generator array to PostScript.
 */
static int dumpOscPS (char * fname, char * title) {
  int i;
  FILE * fp;

  if ((fp = fopen (fname, "w")) == NULL) {
    perror (fname);
    return -1;
  }

  fprintf (fp, "%%!PS\n");
  fprintf (fp, "/pt {lineto} def\n");
  fprintf (fp, "/Helvetica findfont 18 scalefont setfont\n");

  if (title != NULL) {
    fprintf (fp, "50 50 moveto (%s) show\n", title);
  }

  fprintf (fp, "gsave\n");
  fprintf (fp, "/Helvetica findfont 10 scalefont setfont\n");
  fprintf (fp, "50 150 translate\n");
  fprintf (fp, "-10 0 moveto 500 0 lineto stroke\n");
  fprintf (fp, "0 -10 moveto 0 250 lineto stroke\n");
  
  for (i = 1; i <= 91; i++) {
    int x = (i * 500) / 91;
    int y = (int) (250.0 * oscillators[i].attenuation);
    fprintf (fp, "%d %d moveto %d %d lineto\n", x, 0, x, y);
    if ((i % 12) == 1) {
      fprintf (fp, "stroke %d %d moveto (%d) show\n", x, -12, i);
    }
  }

  fprintf (fp, "stroke\n");

  fprintf (fp, "grestore\n");
  fprintf (fp, "showpage\n");

  fclose (fp);

  return 0;
}

/*
 * Dumps the keyOsc[MIDInote][bus] to a text file.
 * @param fname  The filename to write to.
 * @param xpand  If FALSE, write connection number in the keyOsc entry.
 *               If TRUE, lookup the connection entry and expand that.
 */

int dumpKeyOsc (char * fname, int xpand) {
  int n;
  int b;
  FILE * fp;

  if ((fp = fopen (fname, "w")) == NULL) {
    perror (fname);
    return -1;
  }

  fprintf (fp,
	   "Connections from the keyOsc table. Row: MIDI note, Column: %s\n",
	   xpand ? "cnx" : "bus");
  if (xpand) {
    fprintf (fp, "osc:tap:bus\n");
  }

  fprintf (fp, "Note");
  for (b = 0; b < 9; b++) {
    fprintf (fp, "%4d", b);
  }
  fprintf (fp, "\n");

  for (n = 0; n < 128; n++) {
    fprintf (fp, "%3d:", n);
    for (b = 0; b < 9; b++) {
      if (xpand) {
	ConnectionPtr CXP = &(connections[keyOsc[n][b]]);
	fprintf (fp, "%4d:%d:%d", CXP->osc, CXP->tpx, CXP->bus);
      }
      else {
	fprintf (fp, "%4d", keyOsc[n][b]);
      }
    }
    fprintf (fp, "\n");
  }

  fclose (fp);
  return 0;
}

/* ================================================================
 * 22-sep-2003/FK
 * Dump the oscillator array for diagnostics.
 */
static int dumpOscToText (char * fname) {
  FILE * fp;
  int i;
  int j;
  size_t bufferSamples = 0;
  int leastOsc = 0;
  int leastWav = 0;
  double leastTaper = 1.0;

  if ((fp = fopen (fname, "w")) == NULL) {
    perror (fname);
    return -1;
  }

  fprintf (fp, "Oscillator dump\n");
  fprintf (fp, "[%3s]:%10s:%5s:%6s:%5s\n",
	   "OSC",
	   "Frequency",
	   "Sampl",
	   "Bytes",
	   "Gain");
  for (i = 0; i < 128; i++) {
    fprintf (fp, "[%3d]:%7.2lf Hz:%5d:%6d:%5.2lf\n",
	     i,
	     oscillators[i].frequency,
	     oscillators[i].lengthSamples,
	     oscillators[i].lengthBufferBytes,
	     oscillators[i].attenuation);
    fprintf (fp, "   ");
    for (j = 0; j < oscillators[i].numWaves; j++) {
      fprintf (fp, "[%d:%7.2f]", j, oscillators[i].taper[j]);
    }
    fprintf (fp, "\n");
    /* This summary count is probably just an approximation */
    if (0 < oscillators[i].numWaves) {
      bufferSamples += oscillators[i].lengthSamples * oscillators[i].numWaves;
    }
    else {
      bufferSamples += oscillators[i].lengthSamples;
    }
    /* Find last weakest signal */
    for (j = 0; j < oscillators[i].numWaves; j++) {
      if (oscillators[i].taper[j] <= leastTaper) {
	leastOsc = i;
	leastWav = j;
	leastTaper = oscillators[i].taper[j];
      }
    }
  }

  fprintf (fp, "TOTAL MEMORY: %d samples, %d bytes\n",
	   bufferSamples,
	   bufferSamples * sizeof (short));

  leastOsc = 91;
  leastWav = 2;

  if (0 < leastOsc) {
    short * sp = oscillators[leastOsc].waves[leastWav];
    int minv =  32767;
    int maxv = -32767;
    int range;
    fprintf (fp, "Sample dump of generator %d, buffer %d\n",
	     leastOsc,
	     leastWav);
    for (i = 0; i < oscillators[leastOsc].lengthSamples; i++) {
      if (sp[i] < minv) minv = sp[i];
      if (maxv < sp[i]) maxv = sp[i];
    }
    range = maxv - minv;
    for (i = 0; i < oscillators[leastOsc].lengthSamples; i++) {
      fprintf (fp, "%6d: %6d ", i, sp[i]);
      fprintf (fp, "%*c\n", (30 * (sp[i] - minv)) / range, '*');
    }
  }

  fclose (fp);
  return 0;
}

/*
 * This routine configures this module.
 */
int oscConfig (ConfigContext * cfg) {
  int ack = 0;
  double d;
  if ((ack = getConfigParameter_d ("osc.tuning", cfg, &d)) == 1) {
    setTuning (d);
  }
  else if ((ack = getConfigParameter_d ("osc.x-precision", cfg, &d)) == 1) {
    setWavePrecision (d);
  }
  else if (!strcasecmp (cfg->name, "osc.psdump")) {
    ack++;
    strncpy (tgDumpFNBuffer, cfg->value, PATH_MAX);
    tgDumpFname = tgDumpFNBuffer;
  }
  else if (ack = getConfigParameter_d ("osc.perc.fast",
				       cfg,
				       &percFastDecaySeconds)) {
    ;
  }
  else if (ack = getConfigParameter_d ("osc.perc.slow",
				       cfg,
				       &percSlowDecaySeconds)) {
    ;
  }
  else if ((ack = getConfigParameter_d ("osc.perc.normal", cfg, &d)) == 1) {
    setNormalPercussionGain (d);
  }
  else if ((ack = getConfigParameter_d ("osc.perc.soft", cfg, &d)) == 1) {
    setSoftPercussionGain (d);
  }
  else if ((ack = getConfigParameter_d ("osc.perc.gain", cfg, &d)) == 1) {
    setPercussionGainScaling (d);
  }
  else if (!strcasecmp (cfg->name, "osc.eq.macro")) {
    ack++;
    if (!strcasecmp (cfg->value, "chspline")) {
      eqMacro = EQ_SPLINE;
    } else if (!strcasecmp (cfg->value, "peak24")) {
      eqMacro = EQ_PEAK24;
    } else if (!strcasecmp (cfg->value, "peak46")) {
      eqMacro = EQ_PEAK46;
    } else {
      fprintf (stderr,
	       "%s:line %d:%s expected chspline, peak24 or peak46:%s\n",
	       cfg->fname,
	       cfg->linenr,
	       cfg->name,
	       cfg->value);
    }
  }
  else if (ack = getConfigParameter_d ("osc.eq.p1y", cfg, &eqP1y))
    ;
  else if (ack = getConfigParameter_d ("osc.eq.r1y", cfg, &eqR1y))
    ;
  else if (ack = getConfigParameter_d ("osc.eq.p4y", cfg, &eqP4y))
    ;
  else if (ack = getConfigParameter_d ("osc.eq.r4y", cfg, &eqR4y))
    ;
  else if (ack = getConfigParameter_d ("osc.eqv.ceiling", cfg, &eqvCeiling))
    ;
  else if (!strncasecmp (cfg->name, "osc.eqv.", 8)) {
    int n;
    double v;
    ack++;
    if (sscanf (cfg->name, "osc.eqv.%d", &n) == 1) {
      if ((0 <= n) && (n <= 127)) {
	if (sscanf (cfg->value, "%lf", &v) == 1) {
	  if ((0.0 <= v) && (v <= eqvCeiling)) {
	    eqvAtt[n] = v / eqvCeiling;
	    eqvSet[n] = 1;
	  }
	}
	else {
	  configDoubleUnparsable (cfg);
	}
      }
      else {
	configIntOutOfRange (cfg, 0, 127);
      }
    }
  }
  else if (!strncasecmp (cfg->name, "osc.harmonic.", 13)) {
    int n;
    double v;
    ack++;
    if (sscanf (cfg->name, "osc.harmonic.%d", &n) == 1) {
      if ((1 <= n) && (n <= MAX_PARTIALS)) {
	if (sscanf (cfg->value, "%lf", &v) == 1) {
	  wheelHarmonics[n - 1] = v;
	}
	else {
	  configDoubleUnparsable (cfg);
	}
      }
      else {
	configIntOutOfRange (cfg, 1, MAX_PARTIALS);
      }
    }
  }
  else if ((ack = getConfigParameter_dr ("osc.attack.click.level",
					 cfg, &d, 0.0, 100.0)) == 1) {
    setEnvAttackClickLevel (d);
  }
  else if ((ack = getConfigParameter_dr ("osc.attack.click.maxlength",
					 cfg, &d, 0.0, 100.0)) == 1) {
    setEnvAtkClkMaxLength (d);
  }
  else if ((ack = getConfigParameter_dr ("osc.attack.click.minlength",
					 cfg, &d, 0.0, 100.0)) == 1) {
    setEnvAtkClkMinLength (d);
  }
  else if ((ack = getConfigParameter_dr ("osc.release.click.level",
					 cfg, &d, 0.0, 100.0)) == 1) {
    setEnvReleaseClickLevel (d);
  }
  else if (!strcasecmp (cfg->name, "osc.release.model")) {
    ack++;
    if (!strcasecmp (getConfigValue (cfg), "click")) {
      setEnvReleaseModel (ENV_CLICK);
    }
    else if (!strcasecmp (getConfigValue (cfg), "cosine")) {
      setEnvReleaseModel (ENV_COSINE);
    }
    else if (!strcasecmp (getConfigValue (cfg), "linear")) {
      setEnvReleaseModel (ENV_LINEAR);
    }
    else if (!strcasecmp (getConfigValue (cfg), "shelf")) {
      setEnvReleaseModel (ENV_SHELF);
    }
  }
  else if (!strcasecmp (cfg->name, "osc.attack.model")) {
    ack++;
    if (!strcasecmp (getConfigValue (cfg), "click")) {
      setEnvAttackModel (ENV_CLICK);
    }
    else if (!strcasecmp (getConfigValue (cfg), "cosine")) {
      setEnvAttackModel (ENV_COSINE);
    }
    else if (!strcasecmp (getConfigValue (cfg), "linear")) {
      setEnvAttackModel (ENV_LINEAR);
    }
    else if (!strcasecmp (getConfigValue (cfg), "shelf")) {
      setEnvAttackModel (ENV_SHELF);
    }
  }
  return ack;
}


/*
 * This routine initialises the envelope shape tables.
 */
static void initEnvelopes () {
  int bss = BUFFER_SIZE_SAMPLES;
  int bs2 = BUFFER_SIZE_SAMPLES / 2;
  int pattern = 2;
  int b;
  int i;			/* 0 -- 127 */
  int j;			/* 127 -- 0 */
  int burst;			/* Samples in noist burst */
  int bound;
  int start;			/* Sample where burst starts */
  double T = (double) (BUFFER_SIZE_SAMPLES - 1); /* 127.0 */

  for (b = 0; b < 9; b++) {

    if (envAttackModel == ENV_CLICK) {
      /* Select a random length of this burst. */
      bound = envAtkClkMaxLength - envAtkClkMinLength;
      if (bound < 1) {
	bound = 1;
      }
      burst = envAtkClkMinLength + (rand () % bound);
      if (bss <= burst) {
	burst = bss - 1;
      }
      /* Select a random start position of the burst. */
      start = (rand () % (bss - burst));
      /* From sample 0 to start the amplification is zero. */
      for (i = 0; i < start; i++) attackEnv[b][i] = 0;
      /* In the burst area the amplification is random. */
      for (; i < (start + burst); i++) {
	attackEnv[b][i] = ENV_BASE - (rand () % envAttackClickLevel);
      }
      /* From the end of the burst to the end of the envelope the
	 amplification is unity. */
      for (; i < bss; i++) attackEnv[b][i] = ENV_BASE;

#if 1
      /* 2002-08-31/FK EXPERIMENTAL */
      /* Low-pass filter in the time domain. */
      {
	attackEnv[b][0] /= 2;
	for (i = 1; i < bss; i++) {
	  attackEnv[b][i] = (attackEnv[b][i-1] + attackEnv[b][i]) / 2;
	}
      }
#endif

    }

    if (envAttackModel == ENV_SHELF) {
      bound = envAtkClkMaxLength - envAtkClkMinLength;
      if (bound < 1) bound = 1;
      start = rand () % bound;
      if ((bss - 2) <= start) start = bss - 2;
      for (i = 0; i < start; i++) attackEnv[b][i] = 0;
      attackEnv[b][i + 0] = ENV_BASE / 3;
      attackEnv[b][i + 1] = (ENV_BASE * 2) / 3;
      for (i = i + 2; i < bss; i++) {
	attackEnv[b][i] = ENV_BASE;
      }
    }

    if (envReleaseModel == ENV_SHELF) {
      bound = envAtkClkMaxLength - envAtkClkMinLength;
      if (bound < 1) bound = 1;
      start = rand () % bound;
      if ((bss - 2) <= start) start = bss - 2;
      for (i = 0; i < start; i++) releaseEnv[b][i] = ENV_BASE;
      releaseEnv[b][i + 0] = (ENV_BASE * 2) / 3;
      releaseEnv[b][i + 1] = ENV_BASE / 3;
      for (i = i + 2; i < bss; i++) {
	releaseEnv[b][i] = 0;
      }      
    }

    if (envReleaseModel == ENV_CLICK) {
      burst = 8 + (rand () % 32);
      start = (rand () % (bss - burst));

      for (i = 0; i < start; i++) releaseEnv[b][i] = ENV_BASE;
      for (; i < (start + burst); i++) {
	releaseEnv[b][i] = ENV_BASE - (rand () % envReleaseClickLevel);
      }
      for (; i < bss; i++) releaseEnv[b][i] = 0;
      /* Filter the envelope */
      releaseEnv[b][0] /= 2;
      for (i = 1; i < bss; i++) {
	releaseEnv[b][i] = (releaseEnv[b][i-1] + releaseEnv[b][i]) / 2;
      }
    }

    /* cos(0)=1.0, cos(PI/2)=0, cos(PI)=-1.0 */

    if (envAttackModel == ENV_COSINE) {	/* Sigmoid decay */
      for (i = 0; i < BUFFER_SIZE_SAMPLES; i++) {
	int d = BUFFER_SIZE_SAMPLES - (i + 1);
	double a = (M_PI * (double) d) / T;	/* PI < a <= 0 */
	attackEnv [b][i] =
	  (short) ((0.5 + (0.5 * cos (a))) * (double) ENV_BASE);
      }
    }

    if (envReleaseModel == ENV_COSINE) {
      for (i = 0; i < BUFFER_SIZE_SAMPLES; i++) {
	double a = (M_PI * (double) i) / T;	/* 0 < b <= PI */
	releaseEnv[b][i] =
	  (short) ((0.5 + (0.5 * cos (a))) * (double) ENV_BASE);
      }
    }

    if (envAttackModel == ENV_LINEAR) {	/* Linear decay */
      int k = BUFFER_SIZE_SAMPLES;			/* TEST SPECIAL */

      for (i = 0; i < BUFFER_SIZE_SAMPLES; i++) {
	if (i < k) {
	  attackEnv[b][i]  = (ENV_BASE * i) / k;
	} else {
	  attackEnv[b][i] = ENV_BASE;
	}
      }
    }

    if (envReleaseModel == ENV_LINEAR) {
      int k = BUFFER_SIZE_SAMPLES;			/* TEST SPECIAL */

      for (i = 0; i < BUFFER_SIZE_SAMPLES; i++) {
	if (i < k) {
	  releaseEnv[b][i] = (ENV_BASE * (k - i)) / k;
	} else {
	  releaseEnv[b][i] = 0;
	}
      }
    }

  } /* for each envelope buffer */
}


/* MIDI controller callbacks */

static void setDrawbar0 (unsigned char v) {
  drawBars[0] = (8 * ((unsigned int) (127 - v))) / 127;
}
static void setDrawbar1 (unsigned char v) {
  drawBars[1] = (8 * ((unsigned int) (127 - v))) / 127;
}
static void setDrawbar2 (unsigned char v) {
  drawBars[2] = (8 * ((unsigned int) (127 - v))) / 127;
}
static void setDrawbar3 (unsigned char v) {
  drawBars[3] = (8 * ((unsigned int) (127 - v))) / 127;
}
static void setDrawbar4 (unsigned char v) {
  drawBars[4] = (8 * ((unsigned int) (127 - v))) / 127;
}
static void setDrawbar5 (unsigned char v) {
  drawBars[5] = (8 * ((unsigned int) (127 - v))) / 127;
}
static void setDrawbar6 (unsigned char v) {
  drawBars[6] = (8 * ((unsigned int) (127 - v))) / 127;
}
static void setDrawbar7 (unsigned char v) {
  drawBars[7] = (8 * ((unsigned int) (127 - v))) / 127;
}
static void setDrawbar8 (unsigned char v) {
  drawBars[8] = (8 * ((unsigned int) (127 - v))) / 127;
}

/*
 * This routine controls percussion from a MIDI controller.
 * It turns percussion on and off and switches between normal and soft.
 */
static void setPercVolumeFromMIDI (unsigned char u) {
  if (u < 64) {
    if (u < 16) {
      setPercussionEnabled (FALSE); /* off */
    }
    else {
      setPercussionEnabled (TRUE); /* on */
      setPercussionVolume (FALSE); /* normal volume */
    }
  }
  else if (u < 112) {
    setPercussionEnabled (TRUE); /* on */
    setPercussionVolume (TRUE);	/* soft volume */
  }
  else {
    setPercussionEnabled (FALSE); /* off */
  }
}

/*
 * This routine controls percussion from a MIDI controller.
 * It sets fast or slow decay.
 */
static void setPercDecayFromMIDI (unsigned char u) {
  if (u < 64) {
    setPercussionFast (TRUE);
  }
  else {
    setPercussionFast (FALSE);
  }
}

/*
 * This routine controls percussion from a MIDI controller.
 * It sets the third or second harmonic.
 */
static void setPercHarmonicFromMIDI (unsigned char u) {
  if (u < 64) {
    setPercussionFirst (FALSE);
  }
  else {
    setPercussionFirst (TRUE);
  }
}

/*
 * This routine initialises this module. When we come here during startup,
 * configuration files have already been read, so parameters should already
 * be set.
 */
void initToneGenerator () {
/*    initOscillators (tgVariant, tgPrecision); */
/*    initKeyMap (tgVariant); */

  /* Load TMBAS matrix with defaults where needed. */
  initTMBAssembly (tgVariant);
  /* Analyze TMBAS to find taper requirements */
  initFromTMBMatrix ();
  /* Allocate taper buffers, initialize oscillator structs, build keyOsc. */
  initOscillators (tgVariant, tgPrecision);

#ifdef KEYCOMPRESSION

  initKeyCompTable ();

#endif /* KEYCOMPRESSION */

  initEnvelopes ();

/*   setFastPercussionDecay (1.0); */
/*   setSlowPercussionDecay (4.0); */

  setPercussionFirst (FALSE);
  setPercussionVolume (FALSE);
  setPercussionFast (TRUE);
  setPercussionEnabled (FALSE);

  if (tgDumpFname != NULL) {
    dumpOscPS (tgDumpFname, "Tone Generator Attenuation");
  }

  setMIDIControllerFunction (RD_P1_S01, setDrawbar0);
  setMIDIControllerFunction (RD_P1_S02, setDrawbar1);
  setMIDIControllerFunction (RD_P1_S03, setDrawbar2);
  setMIDIControllerFunction (RD_P1_S04, setDrawbar3);
  setMIDIControllerFunction (RD_P1_S05, setDrawbar4);
  setMIDIControllerFunction (RD_P1_S06, setDrawbar5);
  setMIDIControllerFunction (RD_P1_S07, setDrawbar6);
  setMIDIControllerFunction (RD_P1_S08, setDrawbar7);
  setMIDIControllerFunction (RD_P1_K09, setDrawbar8);

  setMIDIControllerFunction (RD_P1_K10, setPercVolumeFromMIDI);
  setMIDIControllerFunction (RD_P1_K11, setPercDecayFromMIDI);
  setMIDIControllerFunction (RD_P1_K12, setPercHarmonicFromMIDI);

#if 0
  dumpKeyOsc ("keyOsc.txt", FALSE);
  dumpKeyOsc ("keyOscMap.txt", TRUE);
  dumpOscToText ("osc.txt");
#endif
}

/*
 * This routine turns off the supplied MIDI note number, if it was active.
 * It does this by writing disconnection events for all the buses and
 * oscillators connected by the key.
 * @param midiNote  The MIDI note to turn off.
 */
void oscNoteOff (unsigned char midiNote) {
  int i;

  //  fprintf (stderr, "OFF(%d)\n", (int) midiNote);

  if ((firstMIDINote <= midiNote) && (midiNote <= lastMIDINote)) {

    /* We only do this if we have previously marked the key as on. */
    if (activeKeys[midiNote] != 0) {

      /* Flag the key as inactive. */
      activeKeys[midiNote] = 0;

#ifdef KEYCOMPRESSION

      keyDownCount--;
      assert (0 <= keyDownCount);

#endif /* KEYCOMPRESSION */

      /* Each bus,oscillator pair referenced by the key generates a
	 disconnect/open event. */

      for (i = 0; i < 9; i++) {
	*cnxqWriter++ = CX_SETEVT(CX_OPEN, keyOsc[midiNote][i]);
	if (cnxqWriter == cnxqEnd) {
	  cnxqWriter = cnxQueue;
	}
      }

    } /* if note was active */
  } /* if note is within accepted range */
}

/*
 * This routine turns on (starts to play) the supplied MIDI note number.
 * It does this by writing connection events for all the buses and
 * oscillators connected by the key.
 * @param midiNote  The MIDI note number to turn on.
 */
void oscNoteOn (unsigned char midiNote) {
  int i;

  //  fprintf (stderr, "ON (%d)\n", (int) midiNote);

  if ((firstMIDINote <= midiNote) && (midiNote <= lastMIDINote)) {

    if (activeKeys[midiNote] != 0) {
      /* We were apparently were not informed of the key release. */
      /* Drop the key first so trigger effects work. */
      oscNoteOff (midiNote);
    }

    activeKeys[midiNote] = 1;	/* Flag the key as active. */

    /* keyOnClick = keyOnClickReset; /* Arm the key click */

#ifdef KEYCOMPRESSION

    keyDownCount++;

#endif /* KEYCOMPRESSION */

    /* Generate a connect event for each bus, oscillator referenced. */

    for (i = 0; i < 9; i++) {
/*        *cnxqWriter++ = CX_SETMSG(i, keyOsc[midiNote][i]); */
      *cnxqWriter++ = CX_SETEVT(CX_CLOSE, keyOsc[midiNote][i]);
      if (cnxqWriter == cnxqEnd) {
	cnxqWriter = cnxQueue;
      }
    }
  }
}

/*
 * This routine installs the drawbar setting for the tone generator.
 * The argument is an array of 9 integers where index 0 corresponds
 * to the setting of 16', index 1 to 5 1/3', index 2 to 8' etc up to
 * index 8 which corresponds to 1'. The values in each element are
 * expected to be 0 (off), 1 (lowest), ... , 8 (loudest).
 * The values are copied from the argument.
 * @param setting  Array of 9 integers.
 */
void setDrawBars (unsigned int setting []) {
  int i;
  for (i = 0; i < B_size; i++) {
    drawBars[i] = setting[i];
  }

}

/*
 * This routine generates the next fragment of output from the tonegenerator.
 * The buffer is overwritten with signed 16-bit samples.
 * --------------------------------
 * 29-May-2003/FK Revised these comments - they were referring to an
 *                earlier version.
 * 
 * (1) Read the connection event queue, compile a core program and
 *     update the state of the model. This is further decomposed into:
 *     (1.a) Connection events: Core instructions are compiled to modulate
 *           the signal with an attack envelope. The connection event is
 *           placed on the added-connection list.
 *     (1.b) Disconnect events. 
 *           The connection is removed from the added- or active-connection
 *           list.
 *     (1.c) Active connections.
 *           The list of active connections is read, and appropriate core
 *           instructions are produced. The contents of the addded-connection
 *           list is transferred to the active-connection list.
 * (2) Execute the core instruction program which writes the bus buffers.
 * (3) Mix the bus buffers according to the drawbar setting into the output
 *     buffer.
 * The rationale behind the multi-section approach is that each step can
 * be optimised to make the most of CPU memory caching and minimal
 * loop administation overhead.
 * Additional features:
 *   - A bitmask is used to keep track of the first write instruction to
 *     each bus buffer. That way the first core instruction is a copy, while
 *     subsequent instructions add, eliminating the need for an initial
 *     zero-filling pass.
 *   - A bitmask is used to keep track of the bus buffers that need or
 *     need not be written. The logic is a simple combination of the
 *     settings for drawbar and percussion.
 * Features not implemented:
 *   - Complete or partial unrolling of the loops where samples are
 *     shuffled around. There IS a benefit in a tight loop in that
 *     the code itself takes very little space and therefore almost
 *     certainly fits in the CPU memory cache.
 *   - Background noise (added from a separate 'noise track').
 * --------------------------------
 * @param buf            Pointer to target buffer.
 * @param lengthSamples  The number of 16-bit samples in the buffer.
 */

void oscGenerateFragment (unsigned char * buf,
			  size_t lengthSamples)
{

  int i;
  short * yp = (short *) buf;
  struct _oscillator * osp;
  int Y;
  unsigned int copyDone = 0;	/* Copy done on bus? */
  unsigned int busClip = 0;	/* Must this bus be rendered? */
  unsigned int oscRef[4];	/* Bitmap set of referenced oscillators */

#ifdef KEYCOMPRESSION
  
  static int keyCompLevel = KEYCOMP_ZERO_LEVEL;
  int keyComp = keyCompTable[keyDownCount];

#define KEYCOMPCHASE() {if (keyCompLevel < keyComp) keyCompLevel++; \
                        else if (keyComp < keyCompLevel) keyCompLevel--;}

#else

#define KEYCOMPCHASE()

#endif /* KEYCOMPRESSION else */

  /*
   * We have a current set of connections to maintain. Connection events
   * update that set. Additions and deletes are reflected in core instructions
   * that employ envelope. Steady state connections use plain instructions
   * except when processing percussion which also uses envelope.
   * The events in the connection queue are processed in arrival order.
   * Contact close events are copied to a temporary list. Contact open
   * events delete from the active set. After the active set has been
   * processed the temporary list of close events are added to the active set.
   *
   * Although we must always model all connections between oscillators and
   * buses, there is no point in generating core instructions that target
   * muted buses. There is a small glitch in there, however, since a bus
   * drawbar changes its state asynchronously. It may therefore happen that
   * a drawbar that goes from 0 to 8 between compilation and mixdown will
   * be included in the mix although the contents of the bus accumulation
   * buffer is undefined (old content, most likely). There will be a 5 ms
   * odd sound. The Proper Way is of course to have another event queue
   * where we listen for drawbar changes. I'll try the glitch first.
   */

/*    assert (lengthSamples == BUFFER_SIZE_SAMPLES); */

  /* Clear the added connections list. */

  addedCnxEnd = 0;

  /* Clear the oscillators reference set. */

  oscRef[0] = oscRef[1] = oscRef[2] = oscRef[3] = 0;

  /* Reset the core program pointers. */

  coreWriter = coreReader = corePgm;

  /* Determine the bus accumulation buffers that must be written. */

  if (0 < drawBars[0]) busClip |= oscBusFlags[0];
  if (0 < drawBars[1]) busClip |= oscBusFlags[1];
  if (0 < drawBars[2]) busClip |= oscBusFlags[2];
  if ((0 < drawBars[3]) || (percEnabled & percFirst)) {
    busClip |= oscBusFlags[3];
  }
  if ((0 < drawBars[4]) || (percEnabled & !percFirst)) {
    busClip |= oscBusFlags[4];
  }
  if (0 < drawBars[5]) busClip |= oscBusFlags[5];
  if (0 < drawBars[6]) busClip |= oscBusFlags[6];
  if (0 < drawBars[7]) busClip |= oscBusFlags[7];
  if (0 < drawBars[8]) busClip |= oscBusFlags[8];

  /* ================================================================
		   C O N N E C T I O N   Q U E U E
     ================================================================ */

  while (cnxqReader != cnxqWriter) {
    short * envelope;		/* Attack or decay envelope */
    unsigned short msg = *cnxqReader++;	/* Get the next event. */
    int cnxIdx = CX_GETCNX(msg); /* Get the connection index */
    ConnectionPtr CXP = &(connections[cnxIdx]);
    int busNumber = CXP->bus;
    int oscNumber = CXP->osc;
    int tapNumber = CXP->tpx;
    int m;
    int n;

/*      assert (isBus(busNumber)); */

    if (cnxqReader == cnxqEnd) { /* Check for wrap on queue */
      cnxqReader = cnxQueue;
    }

    if (CX_GETEVT(msg)) {	/* CLOSE/ATTACK EVENT */
      /* Append this connection to the added connections list. */

      CXP->acx = CNX_ADDED | addedCnxEnd; /* 07-may-2003 */

      addedCnx[addedCnxEnd++] = cnxIdx;
      envelope = attackEnv[busNumber];
    }
    else {			/* OPEN/RELEASE EVENT */

      if (CXP->acx & CNX_ADDED) {
	/* Drop connection from added connections */

	/* The last element's acx is assigned the victim's acx */
	connections[addedCnx[--addedCnxEnd]].acx = CXP->acx;

	/* The victim's element is overwritten with the last element */
	addedCnx[CXP->acx & CNX_ADDED_MASK] = addedCnx[addedCnxEnd];
      }
      else {
	/* Drop connection from active connections */

	/* The last element of the the list is given position of the victim. */
	connections[activeCnx[--actCnxEnd]].acx = CXP->acx;

	/* The victim is overwritten by the previous last element. */
	activeCnx[CXP->acx] = activeCnx[actCnxEnd];
      }

      envelope = releaseEnv[busNumber];
    }

    /* Look up the oscillator structure. */

    osp = &(oscillators[oscNumber]);

    /* If the drawbar is completely closed there is no point in writing
       to the bus accumulation buffer. */  

    if (busClip & oscBusFlags[busNumber]) {

      /* Check the reference bit for this oscillator. */

      if ((oscRef[osp->bitI] & osp->bitM) == 0) {

	/* The oscillator has not been referenced. Set the reference bit. */

	oscRef[osp->bitI] |= osp->bitM;

	/* Set up sources and counts. */

/*  	osp->srcStart0 = &(osp->wave[osp->pos]); */
	osp->srcPosition = osp->pos;

	if (osp->lengthSamples < (osp->pos + BUFFER_SIZE_SAMPLES)) {

	  /* We need two core instructions: the first one reads the number
	     of samples that remain from the current position to the end
	     of the sample. */

	  osp->srcCount0 = osp->lengthSamples - osp->pos;

/*  	  assert (0 < osp->srcCount0); */

	  /* The second instruction reads from the beginning of the sample,
	     the number of samples not consumed by the first instruction.
	     Conveniently enough, that number is exactly where want to leave
	     the read pointer. */

	  osp->pos = BUFFER_SIZE_SAMPLES - osp->srcCount0;

	}
	else {
	  /* One instruction is sufficient. */
	  osp->srcCount0 = BUFFER_SIZE_SAMPLES;

	  /* When we update the read pointer, we must check so that this
	     single instruction will not precisely read all that is left
	     in the sample. If that is the case we wrap the read pointer
	     back to the beginning of the sample buffer. */

	  if ((osp->pos + BUFFER_SIZE_SAMPLES) == osp->lengthSamples) {
	    osp->pos = 0;
	  }
	  else {
	    osp->pos += BUFFER_SIZE_SAMPLES;
	  }
	}
      } /* If oscillator not previously referenced. */

      /* Compile a core instruction for connection release. */

/*        assert (coreWriter < coreEnd); */

      if (copyDone & oscBusFlags[busNumber]) {
	coreWriter->opr = CR_ADDENV;
      } else {
	coreWriter->opr = CR_CPYENV;
	copyDone |= oscBusFlags[busNumber];
      }

      /* We know the target bus accumulation buffer */
      coreWriter->tgt = &(busSum[busNumber][0]);
/*        coreWriter->src = osp->srcStart0; */
      /* The source address is in a taper buffer. */
      coreWriter->src = osp->waves[tapNumber] + osp->srcPosition;
      /* The number of samples to process. */
      coreWriter->cnt = osp->srcCount0;

/*        assert (0 < coreWriter->cnt); */

/*        coreWriter->env = releaseEnv[busNumber]; */
      /* Point to the correct envelope. */
      coreWriter->env = envelope;

      /*
       * If the source sample counter is less than a full fragment we must
       * compile a second instruction.
       */

      if (osp->srcCount0 < BUFFER_SIZE_SAMPLES) {
	CoreIns * prev = coreWriter; /* Remember current writing position. */
	coreWriter += 1;	/* Advance to new instruction */
	coreWriter->opr = prev->opr; /* Copy operation */
	coreWriter->tgt = prev->tgt + prev->cnt; /* Tgt at end of previous */
	coreWriter->src = osp->waves[tapNumber]; /* Src at beginning */
	coreWriter->cnt = osp->pos; /* What remains to copy */

/*  	assert (0 < coreWriter->cnt); */
/*  	assert ((prev->cnt + coreWriter->cnt) == BUFFER_SIZE_SAMPLES); */

	coreWriter->env = prev->env + prev->cnt;

      }
      coreWriter += 1;
    } /* if not muted by drawbar at zero */



  } /* while there are (dis)connection events */

  /* ================================================================
		 A C T I V E   C O N N E C T I O N S
     ================================================================ */

  for (i = 0; i < actCnxEnd; i++) {
    ConnectionPtr CXP = &(connections[activeCnx[i]]);
    int busNumber = CXP->bus;
    int tapNumber = CXP->tpx;

    osp = &(oscillators[CXP->osc]); /* Point to oscillator */

    /* See if the indicated bus is active. */

    if (busClip & oscBusFlags[busNumber]) {

      /* Check the reference bit for this oscillator. */

      if ((oscRef[osp->bitI] & osp->bitM) == 0) {

	/* This oscillator has not been referenced. Set the reference bit. */

	oscRef[osp->bitI] |= osp->bitM;

	/* Note the starting position for this reference. */

	osp->srcPosition = osp->pos;

	/* See if the remainder of the sample is sufficient for fragment. */

	if (osp->lengthSamples < (osp->pos + BUFFER_SIZE_SAMPLES)) {

	  /* It is not, so we need two instructions. The first one copies
	     from the current position to the end of the sample. */

	  osp->srcCount0 = osp->lengthSamples - osp->pos;

	  /* Update the position. */

	  osp->pos = BUFFER_SIZE_SAMPLES - osp->srcCount0;

	}
	else {			/* We need only one instruction. */

	  /* Read one full fragment buffer from oscillator */

	  osp->srcCount0 = BUFFER_SIZE_SAMPLES;

	  /* Check overrun at end of position update. */

	  if ((osp->pos + BUFFER_SIZE_SAMPLES) == osp->lengthSamples) {
	    osp->pos = 0;
	  }
	  else {
	    osp->pos += BUFFER_SIZE_SAMPLES;
	  }

	}
      } /* if oscillator not previously referenced. */

/*  	assert(coreWriter < coreEnd); */

      /* At this point the oscillator is referenced and the srcPosition
         and srcCount0 fields initialized. */

      if (copyDone & oscBusFlags[busNumber]) {
	coreWriter->opr = CR_ADD;
      } else {
	coreWriter->opr = CR_CPY;
	copyDone |= oscBusFlags[busNumber];
      }

      /* Compile the first instruction. */

      coreWriter->tgt = &(busSum[busNumber][0]);
/*        coreWriter->src = osp->srcStart0; */
      coreWriter->src = osp->waves[tapNumber] + osp->srcPosition;
      coreWriter->cnt = osp->srcCount0;

      /* If the source count is less than the fragment buffer we
	 need another instruction. */

      if (osp->srcCount0 < BUFFER_SIZE_SAMPLES) {
	CoreIns * prev = coreWriter;
	coreWriter += 1;
	coreWriter->opr = prev->opr;
	coreWriter->tgt = prev->tgt + prev->cnt;
	coreWriter->src = osp->waves[tapNumber];
	coreWriter->cnt = osp->pos;

/*  	  assert (0 < coreWriter->cnt); */
/*  	  assert ((prev->cnt + coreWriter->cnt) == BUFFER_SIZE_SAMPLES); */

      }
/*  	else { */
/*  	  assert (osp->srcCount0 == BUFFER_SIZE_SAMPLES); */
/*  	} */

      coreWriter += 1;
    }	/* if not muted by drawbar at zero */

  } /* for each active connection */

/*    assert (coreWriter < coreEnd); */

  /* ================================================================
	  U P D A T E   A C T I V E   C O N N E C T I O N S
     ================================================================ */

  for (i = 0; i < addedCnxEnd; i++) {
    connections[addedCnx[i]].acx = actCnxEnd;
    activeCnx[actCnxEnd++] = addedCnx[i];
  }

  /* ================================================================
		     C O R E   P R E - C H E C K
     ================================================================ */

  /* If there are no core instructions, just zero the output buffer. */

  if (coreReader == coreWriter) {
    for (i = 0; i < lengthSamples; i++) {
      *yp++ = 0;
      KEYCOMPCHASE();
    }

    /* Reset the percussion envelope. */

    percEnvGain = percEnvGainReset; /* Alternate 26-May-2003 */

    return;
  }

  /* ================================================================
		   C O R E   I N T E R P R E T E R
     ================================================================ */

  while (coreReader < coreWriter) {
    short opr = coreReader->opr;
    int   * y = coreReader->tgt;
    short * x = coreReader->src;
    int     n = coreReader->cnt;


    short * e = coreReader->env;


/*      if (!inRng (1, n, 1+BUFFER_SIZE_SAMPLES)) { */
/*        printf ("corePgm=0x%04x coreReader=0x%04x coreWriter=0x%04x\n", */
/*  	      corePgm, coreReader, coreWriter); */
/*        printf ("opr=%d, tgt=0x%04x src=0x%04x cnt=%d\n", */
/*  	      coreReader->opr, coreReader->tgt, coreReader->src, */
/*  	      coreReader->cnt); */
/*        printf ("n = %d\n", n); */
/*        assert (inRng (1, n, 1+BUFFER_SIZE_SAMPLES)); */
/*      } */

    if (opr & 1) {		/* ADD group */
      if (opr & 2) {		/* ADDENV */

	for (; 0 < n; n--) {
	  *y++ += (((int) (*x++)) * (*e++)) >> ENV_NORM;
	}

/*  	assert (opr == CR_ADDENV); */

      } else {			/* ADD */

	assert (opr == CR_ADD);

	for (; 0 < n; n--) {
	  *y++ += (int) (*x++);
	}
      }
    } else {			/* CPY group */
      if (opr & 2) {		/* CPYENV */

	for (; 0 < n; n--) {
	  *y++ = (((int) (*x++)) * (*e++)) >> ENV_NORM;
	}

/*  	assert (opr == CR_CPYENV); */

      } else {			/* CPY */

/*  	assert (opr == CR_CPY); */

	for (; 0 < n; n--) {
	  *y++ = (int) (*x++);
	}
      }

    }
    
    coreReader += 1;
  } /* while reading core instructions */



  /* ================================================================
			   M I X - D O W N
     ================================================================ */

  /*
   * The MIX_AT_UNITY symbols postpones the drawbar normalization
   * until the summed 32-bit sample is written to the buffer.
   */

#define MIX_AT_UNITY

#ifdef MIX_AT_UNITY

#define YNORM 6

#else

  /* Normalization shift no perc. Was 3 for VERY long time. */
#define NRM 2

#endif /* NOT MIX_AT_UNITY */

  /* Normalization shift with perc. Was 4 for VERY long time. */
#define NRP 3

  {
#define PERCUPDATE {}

#ifdef KEYCOMPRESSION

#define WRITESAMPLE(S) *yp++ = (short) (((S) * keyCompLevel) >> 11)

#else

#ifdef MIX_AT_UNITY
#define WRITESAMPLE(S) *yp++ = (short) ((S) >> YNORM)
#else
#define WRITESAMPLE(S) *yp++ = (short) ((S) >> 4)
#endif /* NOT MIX_AT_UNITY */

#endif /* NOT KEYCOMPRESSION */

    /* We declare these as static to keep them off the stack where they
       must be accessed relative the stack pointer. */

    static int * b0;
    static int * b1;
    static int * b2;
    static int * b3;
    static int * b4;
    static int * b5;
    static int * b6;
    static int * b7;
    static int * b8;

    b0 = busSum[0];
    b1 = busSum[1];
    b2 = busSum[2];
    b3 = busSum[3];
    b4 = busSum[4];
    b5 = busSum[5];
    b6 = busSum[6];
    b7 = busSum[7];
    b8 = busSum[8];

    if (percEnabled) {
      if (percFirst) {
	for (i = 0; i < lengthSamples; i++) {

#ifdef MIX_AT_UNITY

	  Y = (*b0++ * drawBars[0])
	    + (*b1++ * drawBars[1])
	    + (*b2++ * drawBars[2])
	    + (*b3   * drawBars[3])
	    + (*b4++ * drawBars[4])
	    + (*b5++ * drawBars[5])
	    + (*b6++ * drawBars[6])
	    + (*b7++ * drawBars[7]);

	  Y += (int) (percEnvGain * ((float) (*b3++)));

	  percEnvGain *= percEnvGainDecay;

#else

	  Y = ((*b0++ * drawBars[0]) >> NRP)
	    + ((*b1++ * drawBars[1]) >> NRP)
	    + ((*b2++ * drawBars[2]) >> NRP)
	    + ((*b3++ * percLevel  ) >> NRP)
	    + ((*b4++ * drawBars[4]) >> NRP)
	    + ((*b5++ * drawBars[5]) >> NRP)
	    + ((*b6++ * drawBars[6]) >> NRP)
	    + ((*b7++ * drawBars[7]) >> NRP);

#endif /* MIX_AT_UNITY else */

	  WRITESAMPLE(Y);
	  PERCUPDATE;
	  KEYCOMPCHASE();
	}
      }
      else {
	for (i = 0; i < lengthSamples; i++) {

#ifdef MIX_AT_UNITY

	  Y = (*b0++ * drawBars[0])
	    + (*b1++ * drawBars[1])
	    + (*b2++ * drawBars[2])
	    + (*b3++ * drawBars[3])
	    + (*b4   * drawBars[4])
	    + (*b5++ * drawBars[5])
	    + (*b6++ * drawBars[6])
	    + (*b7++ * drawBars[7]);

	  Y += (int) (percEnvGain * ((float) (*b4++)));

	  percEnvGain *= percEnvGainDecay;

#else

	  Y = ((*b0++ * drawBars[0]) >> NRP)
	    + ((*b1++ * drawBars[1]) >> NRP)
	    + ((*b2++ * drawBars[2]) >> NRP)
	    + ((*b3++ * drawBars[3]) >> NRP)
	    + ((*b4++ * percLevel  ) >> NRP)
	    + ((*b5++ * drawBars[5]) >> NRP)
	    + ((*b6++ * drawBars[6]) >> NRP)
	    + ((*b7++ * drawBars[7]) >> NRP);

#endif /* MIX_AT_UNITY else */

	  WRITESAMPLE(Y);
	  PERCUPDATE;
	  KEYCOMPCHASE();
	}
      }
    }
    else {			/* Percussion not enabled */
      for (i = 0; i < lengthSamples; i++) {

#ifdef MIX_AT_UNITY

	Y = (*b0++ * drawBars[0]) /* 16' */
	  + (*b1++ * drawBars[1]) /* 5 1/3' */
	  + (*b2++ * drawBars[2]) /* 8' */
	  + (*b3++ * drawBars[3]) /* 4' */
	  + (*b4++ * drawBars[4]) /* 2 2/3' */
	  + (*b5++ * drawBars[5]) /* 2' */
	  + (*b6++ * drawBars[6]) /* 1 3/5' */
	  + (*b7++ * drawBars[7]) /* 1 1/3' */
	  + (*b8++ * drawBars[8]); /* 1' */

#else

	Y = ((*b0++ * drawBars[0]) >> NRM) /* 16' */
	  + ((*b1++ * drawBars[1]) >> NRM) /* 5 1/3' */
	  + ((*b2++ * drawBars[2]) >> NRM) /* 8' */
	  + ((*b3++ * drawBars[3]) >> NRM) /* 4' */
	  + ((*b4++ * drawBars[4]) >> NRM) /* 2 2/3' */
	  + ((*b5++ * drawBars[5]) >> NRM) /* 2' */
	  + ((*b6++ * drawBars[6]) >> NRM) /* 1 3/5' */
	  + ((*b7++ * drawBars[7]) >> NRM) /* 1 1/3' */
	  + ((*b8++ * drawBars[8]) >> NRM); /* 1' */

#endif /* NOT MIX_AT_UNITY */

	WRITESAMPLE(Y);
	KEYCOMPCHASE();
      }	/* end of for lengthSamples */
    } /* else no percussion */
  } /* end of anonymous block */
  
  return;

} /* oscGenerateFragment */

